;*=====================================================================*/
;*    serrano/prgm/project/hiphop/examples/gas/gas.scm                 */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Apr  5 10:50:45 2012                          */
;*    Last change :  Thu Apr  5 16:18:34 2012 (serrano)                */
;*    Copyright   :  2012 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    GAS particles demonstration.                                     */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module gas_client

   (library hiphop)

   (static (class particle
	      (init-particle!)
	      (generation::long (default generation))
	      color::bstring
	      x::long
	      y::long
	      (dx::long (default 0))
	      (dy::long (default 0))))

   (export (gas& count size canvas)
	   (gas-contaminate!)
	   (gas-set-animation-speed! v)))

;*---------------------------------------------------------------------*/
;*    gas-machine ...                                                  */
;*---------------------------------------------------------------------*/
(define gas-machine #f)

;*---------------------------------------------------------------------*/
;*    animation-speed ...                                              */
;*---------------------------------------------------------------------*/
(define animation-speed 50)

;*---------------------------------------------------------------------*/
;*    generation ...                                                   */
;*---------------------------------------------------------------------*/
(define generation 1)

;*---------------------------------------------------------------------*/
;*    gas-contaminate! ...                                             */
;*---------------------------------------------------------------------*/
(define (gas-contaminate!)
   (hiphop-input! gas-machine "contaminate"))

;*---------------------------------------------------------------------*/
;*    gas-set-animation-speed! ...                                     */
;*---------------------------------------------------------------------*/
(define (gas-set-animation-speed! v)
   (set! animation-speed v))

;*---------------------------------------------------------------------*/
;*    gas& ...                                                         */
;*---------------------------------------------------------------------*/
(define (gas& count size canvas)
   (let* ((ctx (canvas.getContext "2d"))
	  (prog (hiphop&
		   :input "particle"
		   :input (event& "contaminate" :multi? #f :default 0)
		   (par&
		      (loop&
			 (atom& (ctx.clearRect 0 0 canvas.width canvas.height))
			 (pause&))
		      (map (lambda (i)
			      (run& (particle& i size canvas)))
			 (iota count))))))
      (set! gas-machine (make-hiphop-machine prog))
      (let loop ()
	 (hiphop-react! gas-machine)
	 (after animation-speed loop))))

;*---------------------------------------------------------------------*/
;*    init-particle! ...                                               */
;*---------------------------------------------------------------------*/
(define (init-particle! o::particle)
   (let loop ()
      (set! o.dx (- (random 10) 5))
      (set! o.dy (- (random 10) 5))
      (when (and (=fx o.dx 0) (=fx o.dy 0))
	 (loop))))

;*---------------------------------------------------------------------*/
;*    particle& ...                                                    */
;*---------------------------------------------------------------------*/
(define (particle& i size canvas)
   (let ((p::particle (instantiate::particle
			 (color "#00ff00")
			 (x (random canvas.width))
			 (y (random canvas.height)))))
      (hiphop& :output "particle"
	 :input (event& "contaminate" :multi? #f :default 0)
	 (par&
	    (if (= i 1)
		(loop&
		   (await& (now& "contaminate"))
		   (atom& (begin
			     (set! p.color
				(format "#~2,0x~2,0x~2,0x"
				   (random 255) (random 255) (random 255)))
			     (set! generation (+ 1 generation))
			     (set! p.generation generation))))
		(atom& #f))
	    (loop&
	       (atom& (move-particle p size canvas))
	       (emit& "particle" p)
	       (atom& (collision! p size (val& "particle")))
	       (pause&))))))

;*---------------------------------------------------------------------*/
;*    collision! ...                                                   */
;*---------------------------------------------------------------------*/
(define (collision! p::particle size lst)
   (for-each (lambda (p2::particle)
		(unless (eq? p p2)
		   (when (> p2.generation p.generation)
		      (when (close-enough? p2 p size)
			 (unless (eq? p2.color p.color)
			    (set! p.generation p2.generation)
			    (set! p.color p2.color))))))
      lst))

;*---------------------------------------------------------------------*/
;*    close-enough? ...                                                */
;*---------------------------------------------------------------------*/
(define (close-enough? p1::particle p2::particle size)
   (and (< (abs (- p1.x p2.x)) size) (< (abs (- p1.y p2.y)) size)))

;*---------------------------------------------------------------------*/
;*    move-particle ...                                                */
;*---------------------------------------------------------------------*/
(define (move-particle p::particle size canvas)

   ;; compute the new particle coordinates
   (when (or (<= p.x 0) (>= p.x canvas.width))
      (set! p.dx (- p.dx)))
   (when (or (<= p.y 0) (>= p.y canvas.height))
      (set! p.dy (- p.dy)))
   (set! p.x (+ p.x p.dx))
   (set! p.y (+ p.y p.dy))

   ;; draw it on the canvas
   (let ((ctx (canvas.getContext "2d")))
      (ctx.beginPath)
      (ctx.arc p.x p.y (/ size 2) 0 (* 2 Math.PI) #t)
      (canvas-properties-set! ctx :fill-style p.color)
      (ctx.fill)))
