(module hiphop_musicplayer 
   (library hiphop)

   (import hiphop_xhr)
   
   (export (urlplayer& urls send-to-player audio)))

(cond-expand
   (scheme2js
;; These bigloo/hop procedures does not exist on client side

; This is actually NOT a sort...
(define (sort proc list) list)

; This is NOT a natural comparison, just a plain string-compare3
(define (string-natural-compare3 s1 s2)
   (cond
      ((string-ci=? s1 s2) 0)
      ((string-ci>? s1 s2) 1)
      (else -1)))
))

(define (rename prefix urls)
   (map (lambda (u) (string-append prefix u)) urls))

(define (my-select arg)
   (tprint arg)
   (let ((selection
	    (if (pair? arg)
		(let ((sorted (sort (lambda (s1 s2)
				       (<fx (string-natural-compare3 s1 s2) 0))
				 arg)))
		   (if (pair? (cdr sorted))
		       (cadr sorted)
		       (car sorted)))
		arg)))
      (tprint "Selecting: " selection)
      selection))

(define (urlplayer& urls send-to-player audio)
   (let ((loads (rename "load-" urls))
	 (errors (rename "error-" urls)))
      (hiphop& input: loads errors "ended"
	 (local& ("who" "ready" "go" "play" "play-done"
		    "reqs-done" "play-done")
	    (par&
	       (seq&
		  (all-reqs& urls loads errors)
		  (sustain& "reqs-done"))
	       (run& orch&)
	       (run& (player& send-to-player audio)))))))

(define (all-reqs& urls loads errors)
   (tprint urls)
   (par&
      (map (lambda (u l e)
	      (run& (req& u)
		 bind: "load" l
		 bind: "error" e))
	 urls loads errors)))

(define orch&
   (hiphop&
      input: "ready" "play-done"
      output: "who" "go"
      (loop&
	 (seq& :id 'orch&
	    (until& (now& "ready")
	       (loop&
		  (emit& "who")
		  (pause&)))
	    (emit& "go" (my-select (val& "ready")))
	    (await& (now& "play-done"))
	    (atom& (tprint "orch&-loop"))))))

(define (player& send-to-player audio)
   (hiphop&
      input: "play" "ended"
      output: "play-done"
      initialize: (lambda (the-machine full-name)
		     (let ((m (the-machine))
			   (e (full-name "ended")))
			(set! audio.onended
			   (lambda (audio-event)
			      (tprint "ENDED")
			      (hiphop-input-and-react! m e audio-event)))))
      (seq& :id 'player&
	 (loop&
	    (seq&
	       (await& immediate: #t (now& "play"))
	       (atom& (send-to-player (val& "play")))
	       (await& (now& "ended"))
	       (emit& "play-done")
	       (pause&)
	       (atom& (tprint "player&-loop")))))))

(define (req& url)
   (tprint url)
   (hiphop&
      input: "who" "go" "load" "error"
      output: "ready" "play"
      (local& ("complete" "failed")
	 (run& (with-url& url))
	 (if& (now& "complete")
	    (trap& done
	       (loop&
		  (seq&
		     (await& (now& "who"))
		     (until& :immediate #t (now& "go")
			(sustain& "ready" url))
		     (if& (string=? (val& "go") url)
			(seq&
			   (emit& "play" url)
			   (atom& (tprint "Exiting: " url))
			   (exit& done)))
		     (pause&)
		     (atom& (tprint "req&-loop: " url)))))))))
