(module phidget_spacial_action
   (library hiphop phidget pthread)

   (export (phidget-action& ::spatial-sensor))

   (export (class spatial-sensor
	      (spatial-sensor-init!)
	      (serial-number read-only (default -1))
	      (datarate read-only (default 400))
	      (%spatial::obj (default #unspecified)))
	   
	   (generic spatial-sensor-init! ::spatial-sensor))

   (main test))

(define-macro (dump-now& event)
   `(every& (now& ,event)
       (atom& (print "now: " ,event))
       (pause&)))
(define-macro (dump-pre& event)
   `(every& (pre& ,event)
       (atom& (print "pre: " ,event))
       (pause&)))
(define-macro (dump-val& event)
   `(every& (now& ,event)
       (atom& (print ,event ": " (val& ,event)))
       (pause&)))

(define (phidget-action& sensor::spatial-sensor)
   (define (time-threshold? pair-sec-msec ev::phidget-spatialdata-event)
      (when (and (pair? pair-sec-msec) (pair? (cdr pair-sec-msec)))
	 (let ((action-seconds (car pair-sec-msec))
	       (action-ms (cadr pair-sec-msec))
	       (action-time-threshold (* 10 1000)))
	    (let ((ds (-fx ev.seconds action-seconds)))
	       (cond
		  ((=fx ds 0)
		   (>fx (-fx ev.microseconds action-ms) action-time-threshold))
		  ((>fx ds 1)
		   #t)
		  (else
		   (or (>=fx ev.microseconds action-ms)
		       (>fx (+fx (-fx (*fx 1000 1000) action-ms) ev.microseconds)
			  action-time-threshold))))))))
   (hiphop&
      :input (event& "sensor" :multi? #f :default (class-nil phidget-spatialdata-event))
      :output (event& "next" :multi? #f) (event& "prev" :multi? #f)
      :initialize (lambda (the-machine resolve)
		     (let ((machine (the-machine))
			   (sensor-event (resolve "sensor")))
			(phidget-add-event-listener! sensor.%spatial "spatialdata"
			   (lambda (event::phidget-spatialdata-event)
			      (hiphop-input-and-react! machine sensor-event event)))))
      (local& ((event& "void" :multi? #f)
	       (event& "next-event" :multi? #f :default '())
	       (event& "prev-event" :multi? #f :default '())
	       (event& "next-prepare" :multi? #f :default '())
	       (event& "prev-prepare" :multi? #f :default '()))
	 (every& (now& "sensor")
	    (seq&
	       (par&
		  (if& (let ((ev::phidget-spatialdata-event (val& "sensor"))) (<fl ev.acy -0.5))
		     (emit& "next-event" (let ((ev::phidget-spatialdata-event (val& "sensor")))
					    (list ev.seconds ev.microseconds))))
		  (if& (let ((ev::phidget-spatialdata-event (val& "sensor"))) (>fl ev.acy  0.5))
		     (emit& "prev-event" (let ((ev::phidget-spatialdata-event (val& "sensor")))
					    (list ev.seconds ev.microseconds))))

		  (if& (now& "next-event")
		     (if& (pre& "next-prepare")
			(if& (time-threshold? (preval& "next-prepare") (val& "sensor"))
			   (emit& "next"))
			(if& (or (pre& "void") (pre& "prev-prepare"))
			   (emit& "next-prepare" (val& "next-event"))))
		     
		     (if& (now& "prev-event")
			(if& (pre& "prev-prepare")
			   (if& (time-threshold? (preval& "prev-prepare") (val& "sensor"))
			      (emit& "prev"))
			   (if& (or (pre& "void") (pre& "next-prepare"))
			      (emit& "prev-prepare" (val& "prev-event"))))
			
			(emit& "void")))

		  ;; debug
		  ; (every& (now& "sensor")
		  ;    (atom& (let ((ev::phidget-spatialdata-event (val& "sensor")))
		  ; 	       (print "sensor: acy=" ev.acy "; seconds=" ev.seconds)))
		  ;    (pause&))
		  ; (dump-now& "void")
		  ; (dump-now& "next")
		  ; (dump-val& "next-event")
		  ; (dump-val& "next-prepare")
		  ; (dump-now& "prev")
		  ; (dump-val& "prev-event")
		  ; (dump-val& "prev-prepare")
		  )

	       (halt&))))))

(define (test args)
   (let* ((sensor (instantiate::spatial-sensor (serial-number 166036)))
	  (machine (make-hiphop-machine (phidget-action& sensor))))
      (hiphop-add-event-listener! machine "next" (lambda (e) (print "NEXT!")))
      (hiphop-add-event-listener! machine "prev" (lambda (e) (print "PREV!")))
      (let loop ()
	 (sleep (* 1000 400))
	 (loop))
      (exit 0)))


;*---------------------------------------------------------------------*/
;*    spatial-sensor-init! ::spatial-sensor ...                        */
;*---------------------------------------------------------------------*/
(define-generic (spatial-sensor-init! o::spatial-sensor)
   (with-access::spatial-sensor o (serial-number datarate %spatial)
      (set! %spatial (instantiate::phidget-spatial
			(serial-number serial-number)))
      (phidget-open %spatial)
      (phidget-wait-for-attachment %spatial 2000)
      (with-access::phidget-spatial %spatial ((dt datarate))
         (set! dt datarate))))
