(module phidget_spacial_fall
   (library hiphop phidget pthread)

   (export (phidget-fall& ::spatial-sensor))

   (export (class spatial-sensor
	      (spatial-sensor-init!)
	      (serial-number read-only (default -1))
	      (datarate read-only (default 400))
	      (%spatial::obj (default #unspecified)))
	   
	   (generic spatial-sensor-init! ::spatial-sensor))

   (main test))

(define (phidget-fall& sensor::spatial-sensor)
   (hiphop&
      :input (event& "sensor" :multi? #f :default (class-nil phidget-spatialdata-event))
      :output "down" "up"
      :initialize (lambda (the-machine resolve)
		     (let ((machine (the-machine))
			   (sensor-event (resolve "sensor")))
			(phidget-add-event-listener! sensor.%spatial "spatialdata"
			   (lambda (event::phidget-spatialdata-event)
			      (hiphop-input-and-react! machine sensor-event event)))))
      (local& ((event& "fall-event" :multi? #f)
	       (event& "falling" :multi? #f :default 0) (event& "raising" :multi? #f))
	 (par&
	    (suspend& (or (now& "falling") (pre& "down"))
	       (sustain& "up"))

	    (suspend& (or (now& "up") (now& "raising"))
	       (sustain& "down"))
	    
	    (every& (now& "sensor")
	       (par&
		  ; translation of fall-event? procedure
		  (if& (let ((ev::phidget-spatialdata-event (val& "sensor")))
			  (or (<fl ev.acy -0.5) (>fl ev.acy 0.5)))
		     (emit& "fall-event"))

		  (if& (now& "fall-event")
		     (if& (pre& "falling")
			(if& (>fx (-fx (let ((ev::phidget-spatialdata-event (val& "sensor"))) ev.seconds) (preval& "falling")) 1)
			   (emit& "down")
			   (emit& "falling" (preval& "falling")))
			(if& (pre& "up")
			   (emit& "falling" (let ((ev::phidget-spatialdata-event (val& "sensor"))) ev.seconds))))
		     (if& (pre& "down")
			(emit& "raising")))))
	    
	    ;; debug
	    ; (every& (now& "sensor")
	    ;    (atom& (let ((ev::phidget-spatialdata-event (val& "sensor")))
	    ; 		 (print "sensor: acy=" ev.acy "; seconds=" ev.seconds)))
	    ;    (pause&))
	    ; (every& (now& "fall-event")
	    ;    (atom& (print "fall-event"))
	    ;    (pause&))
	    ; (every& (now& "up")
	    ;    (atom& (print "up"))
	    ;    (pause&))
	    ; (every& (now& "down")
	    ;    (atom& (print "down"))
	    ;    (pause&))
	    ; (every& (now& "falling")
	    ;    (atom& (print "falling: "  (val& "falling")))
	    ;    (pause&))
	    ))))

(define (test args)
   (let* ((sensor (instantiate::spatial-sensor (serial-number 165842)))
	  (machine (make-hiphop-machine (phidget-fall& sensor))))
      (hiphop-add-event-listener! machine "up" (lambda (e) (print "UP!")))
      (hiphop-add-event-listener! machine "down" (lambda (e) (print "DOWN!")))
      (let loop ()
	 (sleep (* 1000 400))
	 (loop))
      (exit 0)))


;*---------------------------------------------------------------------*/
;*    spatial-sensor-init! ::spatial-sensor ...                        */
;*---------------------------------------------------------------------*/
(define-generic (spatial-sensor-init! o::spatial-sensor)
   (with-access::spatial-sensor o (serial-number datarate %spatial)
      (set! %spatial (instantiate::phidget-spatial
			(serial-number serial-number)))
      (phidget-open %spatial)
      (phidget-wait-for-attachment %spatial 2000)
      (with-access::phidget-spatial %spatial ((dt datarate))
         (set! dt datarate))))
