(module prey_predator_datatypes

   (export
      (class item
	 (item-builder!)
	 (id::bint (default -1))
	 (pos-x::bint (default 0))
	 (pos-y::bint (default 0))
	 (vx::bint (default 0))
	 (vy::bint (default 0))
	 (color::bstring (default ""))
	 (radius::bint (default 50)))

      (class predator::item
	 (has-prey?::bool (default #f))
	 (prey::prey (default (class-nil prey))))

      (class prey::item)

      (escape-predator! me::prey pred::pair)
      (find-and-follow-prey! pred::predator l::pair-nil)
      (prey-close-enough? pred::predator)))

(define (random-speed!)
   (-fx (random 11) 5))

(define next-item-id!
   (let ((i 0))
      (lambda ()
	 (set! i (+fx i 1))
	 i)))

(define-generic (item-builder! o::item)
   (set! o.id (next-item-id!))
   (let loop ()
      (let ((vx (random-speed!))
	    (vy (random-speed!)))
	 (if (or (zerofx? vx)
		 (zerofx? vy))
	     (loop)
	     (begin
		(set! o.vx vx)
		(set! o.vy vy)))))
   o)

(define-method (item-builder! o::predator)
   (set! o.color "#00ff00")
   (set! o.radius 20)
   (call-next-method))

(define-method (item-builder! o::prey)
   (set! o.color "#ff0000")
   (set! o.radius 15)
   (call-next-method))


(define (escape-predator! me::prey pred::pair)
   (unless (and (pair? pred) (pair? (cdr pred)))
      (error "escape-predator!" "Illegal argument" pred))
   (unless (=fx me.id (car pred))
      (error "escape-predator!" "Illegal argument" pred))
   (let ((predator::predator (cadr pred)))
      (let ((pvx predator.vx)
	    (pvy predator.vy))
	 (set! me.vx (+fx pvx (if (<fx pvx 0) 1 -1)))
	 (set! me.vy (+fx pvy (if (<fx pvy 0) 1 -1))))))


(define (find-and-follow-prey! pred::predator l::pair-nil)
   (error "find-and-follow-prey!" "NYI" #f))

(define (prey-close-enough? pred::predator)
   (error "prey-close-enough?" "NYI" #f))
