(module prey_predator_predator

   (library hiphop)

   (import prey_predator_datatypes)

   (export (predators&::hiphop-program #!key (init 3))))

(define (predators& #!key (init 3))
   (hiphop&
      ;; Interface
      :output (list
		 (event& "chased" :multi? #t) ; list of currently chased preys
		 (event& "eaten" :multi? #t)) ; list of currently eaten preys
      :output (list
		 (event& "preys-alive" :combine + :default 0)  ; counter of alive preys
		 (event& "entities" :multi? #t)) ; list of alive entities

      ;; Body
      (local& ((event& "newborn")) ; new borns pred

	 ;; The initial predators
	 (emit& "newborn" (iota init))


	 (par&
	    ;; Create the initial preys, we could have used map
	    (genpar& ((new-pred (val& "newborn")))
	       (run& (pred&)))
	    ;; Dynamic generation of futures preys
	    (dyngenpar& ((new-pred (val& "newborn"))) :when (now& "newborn")
	       (run& (pred&)))))))


;; A predator is managed by its starvation index
; - a starvation index, decremented at each reactions. When it falls to zero, the predator die
; - Each time it eats a prey, it gains some count
; - When the index is big enough, it can spend some points to reproduce
(define (pred&)
   (let ((me::predator (instantiate::predator)))
      (hiphop&
	 ;; Interface
	 :output (list
		    (event& "chased" :multi? #t) ; list of currently chased preys
		    (event& "eaten" :multi? #t)) ; list of currently eaten preys
	 :output (list
		    (event& "preys-alive" :combine + :default 0)  ; counter of alive preys
		    (event& "entities" :multi? #t) ; list of alive entities
		    (event& "newborn" :multi? #t))

	 ;; Body
	 (trap& starved ; a predator may die of starvation
	    (local& ((event& "starvation" :multi? #f :default 30) ; number of reaction before the pred dies
		     (event& "eating")) ; local event

	      (emit& "starvation" 30)

	      (par&
		 ;; Starvation
		 (loop&
		    (pause&)
		    (if& (or (not (now& "newborn")) (not (now& "eating")))
		       (atom& (find-and-follow-prey! me (val& "entities")))
		       (emit& "starvation" (-fx (preval& "starvation") 1)))
		    (if& me.has-prey?
		       (emit& "chased" (list me.prey.id me))))

		 ;; Starved status
		 (seq&
		    (await& (<=fx (val& "starvation") 0))
		    (exit& starved))

		 ;; Reproduction
		 (loop&
		    (await& (>fx (val& "starvation") 50))
		    (emit& "newborn" 1)
		    (emit& "starvation" (-fx (preval& "starvation") 15)))

		 ;; Feeding
		 (loop&
		    (await& (prey-close-enough? me))
		    (emit& "eating")
		    (emit& "eaten" (list me.prey.id me))
		    (emit& "starvation" (+fx (preval& "starvation") 12)))))))))

