(module prey_predator_prey

   (library hiphop)

   (import prey_predator_datatypes)

   (export (preys&::hiphop-program #!key (init 10))))

;;; Main preys behavior
;; Parameters
; - init: initial number of preys (integer)
(define (preys&::hiphop-program #!key (init 10))
   (hiphop&
      ;; Interface
      :input (list
		(event& "chased" :multi? #t) ; list of currently chased preys
		(event& "eaten" :multi? #t)) ; list of currently eaten preys
      :output (list
		 (event& "preys-alive" :combine + :default 0)  ; counter of alive preys
		 (event& "entities" :multi? #t)) ; list of alive entities
      ;; Body

      (local& ((event& "newborn" :combine append :default '())) ; new borns preys

	 ;; The initial preys born
	 (emit& "newborn" (iota init))

	 ;; Because dyngenpar& first pauses, we start a plain genpar&
	 ;; statement in parallel to handle the initial prey creation
	 (par&
	    ;; Create the initial preys, we could have used map
	    (genpar& ((new-prey (val& "newborn")))
	       (run& (prey&)))
	    ;; Dynamic generation of futures preys
	    (dyngenpar& ((new-prey (val& "newborn"))) :when (now& "newborn")
	       (run& (prey&)))))))

;;; A prey have two states
;; - chased
; While chased, a prey tries to escape its predator until either the
; prey is reached and eaten by the predator, or the predator dies due
; to starvation.
;; - in peace
; When in peace, a prey try to find another prey in order to reproduce
; the specie. It takes several instants for the reproduction
; to happen.
(define (prey&)
   (let ((me::prey (instantiate::prey)))
      (hiphop&
	 ;; Interface (same as preys&)
	 :input (list
		   (event& "chased" :multi? #t) ; list of currently chased preys
		   (event& "eaten" :multi? #t))  ; list of currently eaten preys
	 :output (list
		    (event& "preys-alive" :combine + :default 0)  ; counter of alive preys
		    (event& "entities" :multi? #t) ; list of alive entities
		    (event& "newborn" :combine append :default '()))

	 ;; Body
	 (trap& eaten ; a prey dies when it is eaten
	    (local& ((event& "escaping")
		     (event& "reproducing" :multi? #f :default 0))

	       (par&
		  ; Eaten or Alive status
		  (loop&
		     (pause&)
		     (if& (and (now& "eaten") (assoc me.id (val& "eaten")))
			(exit& eaten)
			(seq&
			   (emit& "preys-alive" 1)
			   (emit& "entities" me))))

		  ; Chased status
		  (loop&
		     (await& (and (now& "chased") (assoc me.id (val& "chased"))))
		     (atom& (escape-predator! me (assoc me.id (val& "chased"))))
		     (emit& "escaping"))

		  ; Check if the prey can reproduce itself
		  (loop&
		     (if& (not (now& "escaping"))
			(abort& (now& "escaping")
			   (emit& "reproducing" 0)
			   (pause&)
			   (sustain& "reproducing" (+fx 1 (preval& "reproducing"))))
			(pause&)))

		  ; Reproduction
		  (loop&
		     (await& (>fx (val& "reproducing") 5))
		     (emit& "newborn" 1))))))))
