(module prims_fish
   (library hiphop)

   (export (class %fish
	      value
	      pos-x
	      pos-y
	      (dx (default 0))
	      (dy (default 0))
	      (eaten (default 0))
	      (pray::%fish (default (class-nil %fish)))
	      (alive? (default #t))
	      (chased? (default #f))
	      (bbox (default #f))) ; vector of two fixnums height and width

	   (make-fish-program::hiphop-program ::int)))

(cond-expand
   (scheme2js
      (define maxfx max)
))

(define *width* 800)
(define *height* 600)

; (make-fish-program::hiphop-program ::int)
; (create-fish::%fish ::int)
; (fish-direction-reset! ::%fish)
; (fish-move! ::%fish)
; (fish-detect-and-follow! ::%fish ::%fish)

(define (make-fish-program value)
   (let ((fish::%fish (create-fish value)))
      (hiphop& :output "here"
	 (trap& eaten
	    (loop&
	       (atom& (fish-move! fish))
	       (emit& "here" fish)
	       (atom& (bind-exit (return) ; for early aborting the detection
			 (for-each (lambda (other) (when (fish-detect-and-follow! fish other) (return #t))) (val& "here"))))
	       (if& (not fish.alive?) (exit& eaten))
	       (pause&))))))

;; Instantiation
(define (create-fish value)
   (let ((fish (instantiate::%fish
		  (value value)
		  (pos-x (random *width*))
		  (pos-y (random *height*)))))
      (fish-direction-reset! fish)
      fish))

;; Reset de la direction
(define (fish-direction-reset! f::%fish)
   (let loop ()
      (let ((new-dx (-fx (random 6) 3))
	    (new-dy (-fx (random 6) 3)))
	 (if (and (zerofx? new-dx) (zerofx? new-dy))
	     (loop)
	     (begin
		(set! f.dx new-dx)
		(set! f.dy new-dy))))))

;; Le déplacement, copié-collé depuis fthreads/examples/prims.scm
(define (fish-move! fish::%fish)
   (if (nil? fish.pray)
       (let ((fish-h (if fish.bbox (vector-ref fish.bbox 0) 0))
	     (fish-w (if fish.bbox (vector-ref fish.bbox 1) 0)))
	  (if (<fx fish.pos-x 0)                    (set! fish.dx (abs fish.dx)))
	  (if (>fx fish.pos-x (-fx *width* fish-w)) (set! fish.dx (- (abs fish.dx))))
	  (if (<fx fish.pos-y fish-h)               (set! fish.dy (abs fish.dy)))
	  (if (>fx fish.pos-y *height*)             (set! fish.dy (- (abs fish.dy))))
	  (set! fish.pos-x (+fx fish.pos-x fish.dx))
	  (set! fish.pos-y (+fx fish.pos-y fish.dy)))
       (unless fish.pray.alive?
	  (set! fish.pray (class-nil %fish)))))

(define (fish-detect-and-follow! fish1::%fish fish2::%fish)
   (when (and (not (eq? fish1 fish2))         ; different fishes
	      (and fish1.alive? fish2.alive?) ; they are both alive
	      (or (nil? fish1.pray)           ; fish1 does not have a pray
		  (eq? fish1.pray fish2))     ; or the pray is fish2
	      (>fx fish2.value fish1.value)     ; fish2 is bigger than fish1
	      (zerofx? (modulo fish2.value fish1.value))) ; fish1 divides fish2
      (set! fish1.pray fish2)
      (set! fish2.chased? #t)
      (let ((px fish2.pos-x)
	    (py fish2.pos-y))
	 (cond
	    ((<fx px fish1.pos-x)
	     (let ((d (-fx px fish1.pos-x)))
		(set! fish1.dx (if (>fx d -4) d -3))))
	    ((>fx px fish1.pos-x)
	     (let ((d (-fx px fish1.pos-x)))
		(set! fish1.dx (if (<fx d 4) d 3))))
	    (else
	     (set! fish1.dx 0)))
	 (cond
	    ((<fx py fish1.pos-y)
	     (let ((d (-fx py fish1.pos-y)))
		(set! fish1.dy (if (>fx d -4) d -3))))
	    ((>fx py fish1.pos-y)
	     (let ((d (-fx py fish1.pos-y)))
		(set! fish1.dy (if (<fx d 4) d 3))))
	    (else
	     (set! fish1.dy 0)))
	 (set! fish1.pos-x (+fx fish1.pos-x fish1.dx))
	 (set! fish1.pos-y (+fx fish1.pos-y fish1.dy))
	 
	 (when (and
		(<=fx (abs (-fx (+fx fish1.dx fish1.pos-x) px)) (maxfx 3 fish1.dx))
		(<=fx (abs (-fx (+fx fish1.dy fish1.pos-y) py)) (maxfx 3 fish1.dy)))
	    (fish-direction-reset! fish1)
	    (tprint "Fish " fish2.value " has been eat by " fish1.value)
	    (set! fish1.eaten (+fx fish1.eaten 1))
	    (set! fish2.alive? #f))
	 #t)))
