;*=====================================================================*/
;*    serrano/prgm/project/hiphop/examples/sticker/sticker.scm         */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Apr  5 10:50:45 2012                          */
;*    Last change :  Mon Apr 16 11:40:17 2012 (serrano)                */
;*    Copyright   :  2012 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    STICKER demonstration.                                           */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module sticker&

   (library hiphop)

   (static (class particle
	      x::long y::long))

   (export (sticker& count size canvas)
           (sticker-mouse-move e)
	   (sticker-set-animation-speed! v)))

;*---------------------------------------------------------------------*/
;*    sticker-machine ...                                              */
;*---------------------------------------------------------------------*/
(define sticker-machine #f)

;*---------------------------------------------------------------------*/
;*    animation-speed ...                                              */
;*---------------------------------------------------------------------*/
(define animation-speed 50)

;*---------------------------------------------------------------------*/
;*    sticker-set-animation-speed! ...                                 */
;*---------------------------------------------------------------------*/
(define (sticker-set-animation-speed! v)
   (set! animation-speed v))

;*---------------------------------------------------------------------*/
;*    sticker-mouse-move ...                                           */
;*---------------------------------------------------------------------*/
(define (sticker-mouse-move e)
   (hiphop-input! sticker-machine "mouse" e))

;*---------------------------------------------------------------------*/
;*    sticker& ...                                                     */
;*---------------------------------------------------------------------*/
(define (sticker& count size canvas)
   (let* ((ctx (canvas.getContext "2d"))
	  (env (make-matrix (/fx canvas.width size) (/fx canvas.height size)))
	  (prog (hiphop&
		   :input (event& "mouse" :default 0)
		   (par&
		      (loop&
			 (atom& (ctx.clearRect 0 0 canvas.width canvas.height))
			 (pause&))
		      (map (lambda (i)
			      (run& (particle& i size canvas env)))
			 (iota count))))))
      (set! sticker-machine (make-hiphop-machine prog))
      (let loop ()
	 (hiphop-react! sticker-machine)
	 (after animation-speed loop))))

;*---------------------------------------------------------------------*/
;*    make-matrix ...                                                  */
;*---------------------------------------------------------------------*/
(define (make-matrix w h)
   (cons w (make-vector (* w h) #f)))

;*---------------------------------------------------------------------*/
;*    matrix-ref ...                                                   */
;*---------------------------------------------------------------------*/
(define (matrix-ref m x y)
   (vector-ref (cdr m) (+ (* (car m) y) x)))

;*---------------------------------------------------------------------*/
;*    matrix-set! ...                                                  */
;*---------------------------------------------------------------------*/
(define (matrix-set! m x y v)
   (vector-set! (cdr m) (+ (* (car m) y) x) v))

;*---------------------------------------------------------------------*/
;*    particle& ...                                                    */
;*---------------------------------------------------------------------*/
(define (particle& i size canvas env)
   (let* ((x (quotient i canvas.width))
	  (y (remainder i canvas.width))
	  (p (instantiate::particle
		(x x)
		(y y))))
      (hiphop& :input (event& "mouse" :default 0)
	 (loop&
	    (atom& (move-particle p size canvas (val& "mouse") env))
	    (pause&)))))

;*---------------------------------------------------------------------*/
;*    move-particle ...                                                */
;*---------------------------------------------------------------------*/
(define (move-particle p::particle size canvas e env)

   (define (delta target pos)
      (if (> target pos)
	  (min (round (/ (- target pos) 10)) 4)
	  (max (round (/ (- target pos) 10)) -4)))

   (when (pair? e)
      (let ((tx (event-mouse-x (car e)))
	    (ty (event-mouse-y (car e))))
	 (let loop ((dx (delta tx p.x))
		    (dy (delta ty p.y)))
	    (unless (and (= dx 0) (= dy 0))
	       (let* ((nx (min (max (+ p.x dx) 0) canvas.width))
		      (ny (min (max (+ p.y dy) 0) canvas.height))
		      (oref (matrix-ref env (/fx nx size) (/fx ny size))))
		  (cond
		     ((or (not oref) (eq? oref p))
		      ;; move
		      (matrix-set! env (/fx p.x size) (/fx p.y size) #f)
		      (matrix-set! env (/fx nx size) (/fx ny size) p)
		      (set! p.x nx)
		      (set! p.y ny))
		     ((= dy 0)
		      ;; try to move horizontally
		      (loop (+ dx (if (< dx 0) 1 -1)) dy))
		     ((= dx 0)
		      ;; try to move vertically
		      (loop dx (+ dy (if (< dy 0) 1 -1))))
		     (else
		      ;; try a random move
		      (if (= (random 2) 0)
			  (loop dx (+ dy (if (< dy 0) 1 -1)))
			  (loop (+ dx (if (< dx 0) 1 -1)) dy)))))))
	 (let* ((dist (inexact->exact
			 (sqrt (+ (expt (abs (- tx p.x)) 2)
				  (expt (abs (- ty p.y)) 2)))))
		(color (cond
			  ((>= dist 200) "#00ff00")
			  ((>= dist 100) "#ffec18")
			  ((>= dist 50) "#ff7f17")
			  (else "#ff0000"))))
	    (let ((ctx (canvas.getContext "2d")))
	       (ctx.beginPath)
	       (ctx.arc p.x p.y (/ size 2) 0 (* 2 Math.PI) #t)
	       (canvas-properties-set! ctx :fill-style color)
	       (ctx.fill))))))



