(module wristwatch_alarm_datatype

   (import wristwatch_beep wristwatch_watch_datatype)

   (export
      ;; classes
      (class alarm-time
	 hours::bint
	 minutes::bint
	 mode12h::bool)

      ;; functions
      (initial-alarm-time-position::bint)
      (alarm-duration::bint)
      (alarm-beep-value::beep)
      (initial-alarm-time::alarm-time)
      (toggle-24h-mode-in-alarm-time!::alarm-time at::alarm-time)
      (next-alarm-time-position::bint atp::bint)
      (set-alarm-time!::alarm-time at::alarm-time atp::bint)
      (compare-alarm-time-to-watch-time::bool at::alarm-time wt::watch-time)))

(define (initial-alarm-time-position::bint) 1)

(define (alarm-duration::bint) 30)

(define (alarm-beep-value::beep) four-beeps)


;; Internal
(define minutes-per-hour::bint 60)
(define minutes-per-10mn::bint 10)
(define hours-per-day::bint 24)
(define pm-value::bint 12)

;; Alarm API
(define (initial-alarm-time::alarm-time)
   (instantiate::alarm-time (hours 0) (minutes 0) (mode12h #f)))

(define (toggle-24h-mode-in-alarm-time!::alarm-time at::alarm-time)
   (set! at.mode12h (not at.mode12h))
   at)

(define (next-alarm-time-position::bint atp::bint)
   (let ((next (remainder (+fx atp 1) 3)))
      (if (zerofx? next)
	  3
	  next)))

(define (set-alarm-time!::alarm-time at::alarm-time atp::bint)
   (case atp
      ((0) ; hours
       (set! at.hours (remainder (+fx at.hours 1) hours-per-day)))
      ((1) ; 10 minutes
       (set! at.minutes
	  (if (>=fx at.minutes (-fx minutes-per-hour minutes-per-10mn))
	      (-fx at.minutes (-fx minutes-per-hour minutes-per-10mn))
	      (+fx at.minutes minutes-per-10mn))))
      ((2) ; minutes
       (set! at.minutes
	  (if (=fx (remainder at.minutes 10) (-fx minutes-per-10mn 1))
	      (-fx at.minutes (-fx minutes-per-10mn 1))
	      (+fx at.minutes 1)))))
   at)

(define (compare-alarm-time-to-watch-time::bool at::alarm-time wt::watch-time)
   (and (zerofx? wt.seconds)
	(=fx at.hours wt.hours)
	(=fx at.minutes wt.minutes)))

