(module wristwatch_alarm

   (library hiphop)

   (import wristwatch_alarm_datatype wristwatch_beep wristwatch_watch_datatype)

   (export alarm&::hiphop-program))

(define alarm&::hiphop-program
   (hiphop&
      name: "alarm-module"

      ;; Interface

      input:
      (list
	 (event2& "toggle-24h-mode-command"          'pure))
      output:
      (list
	 (event2& "alarm-time"                       'single :default (class-nil alarm-time)))
      input:
      (list
	 (event2& "enter-set-alarm-mode-command"     'pure)
	 (event2& "next-alarm-time-position-command" 'pure)
	 (event2& "exit-set-alarm-mode-command"      'pure)
	 (event2& "set-alarm-command"                'pure))
      output:
      (list
	 (event2& "start-enhancing"                  'single :default 0)
	 (event2& "stop-enhancing"                   'single :default 0))
      input:
      (list
	 (event2& "watch-time"                       'single :default (class-nil watch-time))
	 (event2& "watch-being-set"                  'pure))
      input:
      (list
	 (event2& "S"                                'pure))
      input:
      (list
	 (event2& "toggle-alarm-command"             'pure)
	 (event2& "stop-alarm-beep-command"          'pure))
      output:
      (list
	 (event2& "alarm-status"                     'single :default #f)
	 (event2& "beep"                             'combined :combine combine-beeps :default no-beep-value))

      ;; Body

      (local& ((event2& "start-beeping" 'pure))

	 ;; initializations

	 (emit& "alarm-time" (initial-alarm-time))
	 (emit& "alarm-status" #f)

	 ;; main loop

	 (par&
	    (loop&
	       ;; normal mode
	       (abort& (now& "enter-set-alarm-mode-command")
		  (loop&
		     (trap& await-case
			(par&
			   (seq&
			      (await& (now& "toggle-24h-mode-command"))
			      (emit& "alarm-time" (toggle-24h-mode-in-alarm-time! (preval& "alarm-time")))
			      (exit& await-case))
			   (seq&
			      (await& (now& "toggle-alarm-command"))
			      (emit& "alarm-status" (not (preval& "alarm-status")))
			      (exit& await-case))
			   (seq&
			      (await& (now& "watch-time"))
			      (if& (now& "watch-being-set")
				 (if& (and (compare-alarm-time-to-watch-time (val& "alarm-time") (val& "watch-time"))
					   (val& "alarm-status"))
				    (emit& "start-beeping")))
			      (exit& await-case))))))

	       ;; set-alarm mode

	       ;; the currently set position is the value of START_ENHANCING
	       ;; (one might also accept TOGGLE_24H_MODE_COMMAND
	       ;; and TOGGLE_ALARM_COMMAND; for this one just has to
	       ;; copy the corresponding cases above into the next await).
	       ;; Notice that the alarm does not ring in set mode

	       (abort& (now& "exit-set-alarm-mode-command")
		  (emit& "start-enhancing" (initial-alarm-time-position))
		  (loop&
		     (trap& await-case
			(par&
			   (seq&
			      (await& (now& "set-alarm-command"))
			      (emit& "alarm-time" (set-alarm-time! (preval& "alarm-time") (val& "start-enhancing")))
			      (exit& await-case))
			   (seq&
			      (await& (now& "next-alarm-time-position-command"))
			      (emit& "stop-enhancing" (preval& "start-enhancing"))
			      (emit& "start-enhancing" (next-alarm-time-position (preval& "start-enhancing")))
			      (exit& await-case))))))
	       (emit& "stop-enhancing" (val& "start-enhancing"))
	       (emit& "alarm-status" #t))

	    ;; beeping sequence
	    (every& (now& "start-beeping")
	       (abort& (now& "stop-alarm-beep-command")
		  (emit& "beep" (alarm-beep-value))
		  (every& (now& "S")
		     (emit& "beep" (alarm-beep-value)))))))))
