(module wristwatch_button

   (library hiphop)

   (export button&::hiphop-program))

(define button&::hiphop-program
   (hiphop&
      name: "button-module"

      ;; Interface

      input: (list ; buttons
		(event2& "UL" 'pure)  ; upper-left button
		(event2& "UR" 'pure)  ; upper-right button
		(event2& "LL" 'pure)  ; lower-left button
		(event2& "LR" 'pure)) ; lower-right button
      output: (list
		 (event2& "watch-mode-command"                'pure)

		 (event2& "enter-set-watch-mode-command"      'pure)
		 (event2& "set-watch-command"                 'pure)
		 (event2& "next-watch-time-position-command"  'pure)
		 (event2& "exit-set-watch-mode-command"       'pure)

		 (event2& "toggle-24h-mode-command"           'pure)
		 (event2& "toggle-chime-command"              'pure)

		 (event2& "stopwatch-mode-command"            'pure)

		 (event2& "start-stop-command"                'pure)
		 (event2& "lap-command"                       'pure)

		 (event2& "alarm-mode-command"                'pure)

		 (event2& "enter-set-alarm-mode-command"      'pure)
		 (event2& "set-alarm-command"                 'pure)
		 (event2& "next-alarm-time-position-command"  'pure)
		 (event2& "exit-set-alarm-mode-command"       'pure)

		 (event2& "toggle-alarm-command"              'pure)
		 (event2& "stop-alarm-beep-command"           'pure))

      ;; Body

      (par&
	 (loop&

	    ;; Watch and set-watch mode (exit by LL in watch mode only,
	    ;; not in set-watch mode)

	    (emit& "watch-mode-command")
	    (trap& watch-and-set-watch-mode
	       (loop&
		  ;; watch mode
		  (abort& (now& "UL") ; that turns to set-watch mode
		     (par&
			(seq&
			   (await& (now& "LL"))
			   (exit& watch-and-set-watch-mode))
			(every& (now& "LR")
			   (emit& "toggle-24h-mode-command"))))
		  ;; set-watch mode
		  (emit& "enter-set-watch-mode-command")
		  (abort& (now& "UL") ;; that brings back to watch mode
		     (par&
			(every& (now& "LL")
			   (emit& "next-watch-time-position-command"))
			(every& (now& "LR")
			   (emit& "set-watch-command"))))
		  (emit& "exit-set-watch-mode-command")))

	    ;; Stopwatch mode (exit by LL); LR is START/STOP, UR is LAP

	    (emit& "stopwatch-mode-command")
	    (abort& (now& "LL")
	       (par&
		  (every& (now& "LR")
		     (emit& "start-stop-command"))
		  (every& (now& "UR")
		     (emit& "lap-command"))))

	    ;; Alarm and set-alarm mode (exit by LL in alarm mode only,
	    ;; not in set-alarm mode)

	    (trap& alarm-and-set-alarm-mode
	       (emit& "alarm-mode-command")
	       (loop&
		  ;; alarm mode
		  (abort& (now& "UL") ; that turns to set-alarm mode
		     (par&
			(seq&
			   (await& (now& "LL"))
			   (exit& alarm-and-set-alarm-mode))
			(every& (now& "LR")
			   (emit& "toggle-chime-command"))
			(every& (now& "UR")
			   (emit& "toggle-alarm-command"))))
		  ;; set-alarm mode
		  (emit& "enter-set-alarm-mode-command")
		  (abort& (now& "UL") ; that returns to alarm mode
		     (par&
			(every& (now& "LL")
			   (emit& "next-alarm-time-position-command"))
			(every& (now& "LR")
			   (emit& "set-alarm-command"))))
		  (emit& "exit-set-alarm-mode-command"))))

	 ;; transforms permanently UR into stop-alarm-bee-command
	 (every& (now& "UR")
	    (emit& "stop-alarm-beep-command")))))
