(module wristwatch_display_datatype

   (import
      wristwatch_alarm_datatype
      wristwatch_beep
      wristwatch_stopwatch_datatype
      wristwatch_watch_datatype)

   (export
      ;; classes
      (class main-display
	 hours::string
	 minutes::string
	 seconds::string
	 timemode::string)
      (class mini-display
	 fst::string
	 sep::string
	 snd::string)

      ;; functions
      watch-time-to-main-display
      watch-time-to-mini-display
      watch-date-to-mini-display
      (watch-day-to-alphabetic-display::string wt::watch-time)
      (watch-display-position::bint wtp::bint)
      stopwatch-time-to-main-display
      alarm-time-to-main-display
      (alarm-display-position::bint atp::bint)))

;; Internals
(define pm-value 12)

;; Utils
; Convert a fixnum into a string, and add pad if its length is less
; than 2
(define (fx->string i::bint pad::char)
   (let ((s (integer->string i)))
      (if (<fx (string-length s) 2)
	  (format "~a~a" pad s)
	  s)))


;; Exported
(define watch-time-to-main-display
   (let ((maindt::main-display (instantiate::main-display (hours "0") (minutes "00") (seconds "00") (timemode "24H")))
	 (hours::bint 0))
      (lambda (wt::watch-time)
	 (if wt.mode12h
	     (begin
		(set! hours (remainder wt.hours pm-value))
		(when (zerofx? hours)
		   (set! hours pm-value)))
	     (set! hours wt.hours))
	 (set! maindt.hours (fx->string hours #\space))
	 (set! maindt.minutes (fx->string wt.minutes #\0))
	 (set! maindt.seconds (fx->string wt.seconds #\0))
	 (if wt.mode12h
	     (if (>fx wt.hours (-fx pm-value 1))
		 (set! maindt.timemode "PM")
		 (set! maindt.timemode ""))
	     (set! maindt.timemode "24H"))
	 maindt)))

(define watch-time-to-mini-display
   (let ((minidt::mini-display (instantiate::mini-display (fst "00") (sep ":") (snd "00"))))
      (lambda (wt::watch-time)
	 (set! minidt.fst (fx->string wt.hours #\space))
	 (set! minidt.sep ":")
	 (set! minidt.snd (fx->string wt.minutes #\0))
	 minidt)))

(define watch-date-to-mini-display
   (let ((minidt::mini-display (instantiate::mini-display (fst "NA") (sep "-") (snd "NA"))))
      (lambda (wt::watch-time)
	 (set! minidt.fst (fx->string (+fx wt.month 1) #\space))
	 (set! minidt.sep "-")
	 (set! minidt.snd (fx->string wt.day #\space))
	 minidt)))

(define day-names (vector "SU" "MO" "TU" "WE" "TH" "FR" "SA"))

(define (watch-day-to-alphabetic-display::string wt::watch-time)
   (vector-ref day-names wt.day-in-week))

(define (watch-display-position::bint wtp::bint)
   wtp)

(define stopwatch-time-to-main-display
   (let ((maindt::main-display (instantiate::main-display (hours "0") (minutes "00") (seconds "00") (timemode ""))))
      (lambda (st::stopwatch-time)
	 (set! maindt.hours (fx->string st.minutes #\space))
	 (set! maindt.minutes (fx->string st.seconds #\0))
	 (set! maindt.seconds (fx->string st.hundredths #\0))
	 (set! maindt.timemode "")
	 maindt)))

(define alarm-time-to-main-display
   (let ((maindt::main-display (instantiate::main-display (hours "0") (minutes "00") (seconds "00") (timemode "")))
	 (hours::bint 0))
      (lambda (at::alarm-time)
	 (if at.mode12h
	     (begin
		(set! hours (remainder at.hours pm-value))
		(when (zerofx? hours) (set! hours pm-value)))
	     (set! hours at.hours))
	 (set! maindt.hours (fx->string hours #\space))
	 (set! maindt.minutes (fx->string at.minutes #\0))
	 (set! maindt.seconds "")
	 (if at.mode12h
	     (if (>fx at.hours (-fx pm-value 1))
		 (set! maindt.timemode "PM")
		 (set! maindt.timemode ""))
	     (set! maindt.timemode "24H"))
	 maindt)))

(define (alarm-display-position::bint atp::bint)
   (+fx atp 1))
