(module wristwatch_display

   (library hiphop)

   (import
      wristwatch_alarm_datatype
      wristwatch_beep
      wristwatch_display_datatype
      wristwatch_stopwatch_datatype
      wristwatch_watch_datatype)

   (export display&::hiphop-program))

(define display&::hiphop-program
   (hiphop&
      :name "display-module"

      ;; Interface

      :output
      (list
	 ; For the main display
	 (event2& "main-display"           'single :default (class-nil main-display))
	 ; For the mini display
	 (event2& "mini-display"           'single :default (class-nil mini-display))
	 ; For the alphabetic display
	 (event2& "alphabetic-display"     'single :default "--")
	 ; For display positions
	 (event2& "start-enhancing"        'single :default 0)
	 (event2& "stop-enhancing"         'single :default 0))
      :input
      (list
	 ; To handle the watch
	 (event2& "watch-mode-command"     'pure)
	 (event2& "watch-time"             'single  :default (class-nil watch-time))
	 (event2& "watch-start-enhancing"  'single :default (initial-watch-time-position))
	 (event2& "watch-stop-enhancing"   'single :default (initial-watch-time-position))
	 ; To handle the stopwatch
	 (event2& "stopwatch-mode-command" 'pure)
	 (event2& "stopwatch-time"         'single  :default (class-nil stopwatch-time))
	 ; To handle the alarm
	 (event2& "alarm-mode-command"     'pure)
	 (event2& "alarm-time"             'single :default (class-nil alarm-time))
	 (event2& "alarm-start-enhancing"  'single :default (initial-alarm-time-position))
	 (event2& "alarm-stop-enhancing"   'single :default (initial-alarm-time-position)))

      ;; Body

      (loop&

	 ;; In watch mode, the main display shows the regular time and
	 ;; the mini display shows the date

	 (abort& (now& "stopwatch-mode-command")
	    (par&
	       (seq&
		  (emit& "main-display" (watch-time-to-main-display (val& "watch-time")))
		  (emit& "mini-display" (watch-date-to-mini-display (val& "watch-time")))
		  (emit& "alphabetic-display" (watch-day-to-alphabetic-display (val& "watch-time")))
		  (every& (now& "watch-time")
		     (emit& "main-display" (watch-time-to-main-display (val& "watch-time")))
		     (emit& "mini-display" (watch-date-to-mini-display (val& "watch-time")))
		     (emit& "alphabetic-display" (watch-day-to-alphabetic-display (val& "watch-time")))))
	       (every& (now& "watch-start-enhancing")
		  (emit& "start-enhancing" (watch-display-position (val& "watch-start-enhancing"))))
	       (every& (now& "watch-stop-enhancing")
		  (emit& "stop-enhancing" (watch-display-position (val& "watch-stop-enhancing"))))))

	 ;; Stopwatch and alarm modes

	 (abort& (now& "watch-mode-command")

	    (par&

	       ;; The mini display contains the regular watch time
	       (seq&
		  (emit& "mini-display" (watch-time-to-mini-display (val& "watch-time")))
		  (every& (now& "watch-time")
		     (emit& "mini-display" (watch-time-to-mini-display (val& "watch-time")))))

	       ;; stopwatch mode
	       (seq&
		  (abort& (now& "alarm-mode-command")
		     (emit& "alphabetic-display" "ST")
		     (emit& "main-display" (stopwatch-time-to-main-display (val& "stopwatch-time")))
		     (every& (now& "stopwatch-time")
			(emit& "main-display" (stopwatch-time-to-main-display (val& "stopwatch-time")))))

		  ;; alarm mode
		  (abort& (now& "watch-mode-command")
		     (par&
			(seq&
			   (emit& "alphabetic-display" "AL")
			   (emit& "main-display" (alarm-time-to-main-display (val& "alarm-time")))
			   (every& (now& "alarm-time")
			      (emit& "main-display" (alarm-time-to-main-display (val& "alarm-time")))))
			(every& (now& "alarm-start-enhancing")
			   (emit& "start-enhancing" (alarm-display-position (val& "alarm-start-enhancing"))))
			(every& (now& "alarm-stop-enhancing")
			   (emit& "stop-enhancing" (alarm-display-position (val& "alarm-stop-enhancing"))))))))))))
