(module wristwatch_watch_datatype

   (import wristwatch_beep)

   (export ;; data-types
      (class watch-time
	 day-in-week::bint
	 month::bint
	 day::bint
	 hours::bint
	 minutes::bint
	 seconds::bint
	 mode12h::bool))

   (export ;; functions
      (initial-watch-time-position)
      (get-initial-watch-time::watch-time)
      (increment-watch-time!::watch-time wt::watch-time)
      (toggle-24h-mode-in-watch-time!::watch-time wt::watch-time)
      (next-watch-time-position::bint wtp::bint)
      (set-watch-time::watch-time wt::watch-time wtp::bint)
      (increment-watch-time-in-set-mode wt::watch-time wtp::bint)
      (watch-beep::beep wt::watch-time b::bool))

   (export ;; constants
      initial-watch-time))

(define (initial-watch-time-position) 0)
(define watch-beep-value::beep two-beeps)

;; Time handling constants
(define seconds-per-minute 60)
(define minutes-per-10mn 10)
(define minutes-per-hour 60)
(define hours-per-day 24)
(define pm-value 12)
(define days-per-week 7)
(define months-per-year 12)

(define month-lengths (vector 31 28 31 30 31 30 31 31 30 31 30 31))


(define initial-watch-time #f)

(define (get-initial-watch-time::watch-time)
   (let ((cd (current-date)))
      (instantiate::watch-time
	 (day-in-week (-fx (date-wday cd) 1)) ; date-wday in [1..7]
	 (month (-fx (date-month cd) 1)) ; date-month in [1..12]
	 (day (date-day cd))
	 (hours (date-hour cd))
	 (minutes (date-minute cd))
	 (seconds (date-second cd))
	 (mode12h #f))))

(define (increment-watch-time!::watch-time wt::watch-time)
   (set! wt.seconds (+fx wt.seconds 1))
   (when (=fx wt.seconds seconds-per-minute)
      (set! wt.seconds 0)
      (set! wt.minutes (+fx wt.minutes 1))
      (when (=fx wt.minutes minutes-per-hour)
	 (set! wt.minutes 0)
	 (set! wt.hours (+fx wt.hours 1))
	 (when (=fx wt.hours hours-per-day)
	    (set! wt.hours 0)
	    (set! wt.day (+fx wt.day 1))
	    (set! wt.day-in-week (modulo (+fx wt.day-in-week 1) days-per-week))
	    (when (>fx wt.day (vector-ref month-lengths wt.month))
	       (set! wt.day 1)
	       (set! wt.month (modulo (+fx wt.month 1) months-per-year))))))
   wt)

(define (toggle-24h-mode-in-watch-time!::watch-time wt::watch-time)
   (set! wt.mode12h (not wt.mode12h))
   wt)

(define (next-watch-time-position::bint wtp::bint)
   (modulo (+fx wtp 1) 7))

(define (set-watch-time::watch-time wt::watch-time wtp::bint)
   (case wtp
      ((0) ; seconds
       (set! wt.seconds 0))
      ((1) ; hours
       (set! wt.hours 0))
      ((2) ; 10 minutes
       (set! wt.minutes
	  (if (>=fx wt.minutes (-fx minutes-per-hour minutes-per-10mn))
	      (-fx wt.minutes (-fx minutes-per-hour minutes-per-10mn))
	      (+fx wt.minutes minutes-per-10mn))))
      ((3) ; minutes
       (set! wt.minutes
	  (if (=fx (remainder wt.minutes minutes-per-10mn) minutes-per-10mn)
	      (-fx wt.minutes (-fx minutes-per-10mn 1))
	      (+fx wt.minutes 1))))
      ((4) ; month
       (set! wt.month (remainder (+fx wt.month 1) months-per-year)))
      ((5) ; day
       (set! wt.day (if (=fx wt.day (vector-ref month-lengths wt.month))
			1
			(+fx wt.day 1))))
      ((6) ; day in week
       (set! wt.day-in-week (remainder (+fx wt.day-in-week 1) days-per-week))))
   wt)

(define (increment-watch-time-in-set-mode wt::watch-time wtp::bint)
   (bind-exit (return)
      (set! wt.seconds (+fx wt.seconds 1))
      (when (=fx wt.seconds seconds-per-minute)
	 (set! wt.seconds 0)
	 (when (=fx wtp 3) ; in set minutes mode
	    (return wt))
	 (when (=fx wtp 2) ; set 10 minutes mode
	    (if (=fx (remainder wt.minutes 10) 9)
		(set! wt.minutes (-fx wt.minutes 9))
		(set! wt.minutes (+fx wt.minutes 1)))
	    (return wt))
	 (set! wt.minutes (+fx wt.minutes 1))
	 (when (=fx wt.minutes minutes-per-hour)
	    (set! wt.minutes 0)
	    (when (=fx wtp 1) ; set hours mode
	       (return wt))
	    (set! wt.hours (+fx wt.hours 1))
	    (when (=fx wt.hours hours-per-day)
	       (set! wt.hours 0)
	       (when (or (=fx wtp 4) (=fx wtp 6)) ; set day mode
		  (return wt))
	       (set! wt.day (+fx wt.day 1))
	       (set! wt.day-in-week (remainder (+fx wt.day-in-week 1) days-per-week))
	       (when (>fx wt.day (vector-ref month-lengths wt.month))
		  (set! wt.day 1)
		  (when (=fx wtp 5) ; sets month mode
		     (return wt))
		  (set! wt.month (remainder (+fx wt.month 1) months-per-year))))))
      (return wt)))

(define (watch-beep::beep wt::watch-time b::bool)
   (if (and b (zerofx? wt.seconds) (zerofx? wt.minutes))
       watch-beep-value
       no-beep-value))
