(module wristwatch_watch

   (library hiphop)

   (import wristwatch_beep wristwatch_watch_datatype)

   (export watch&::hiphop-program))

(define watch&::hiphop-program
   (hiphop&
      name: "watch-module"

      ;; Interface

      input:
      (event2& "S"                                'pure)
      (event2& "toggle-24h-mode-command"          'pure)

      output:
      (event2& "watch-time"                       'single :default (class-nil watch-time))

      input:
      (event2& "enter-set-watch-mode-command"     'pure)
      (event2& "set-watch-command"                'pure)
      (event2& "next-watch-time-position-command" 'pure)
      (event2& "exit-set-watch-mode-command"      'pure)

      output:
      (event2& "watch-being-set"                  'pure)
      (event2& "start-enhancing"                  'single :default (initial-watch-time-position))
      (event2& "stop-enhancing"                   'single :default (initial-watch-time-position))

      input:
      (event2& "toggle-chime-command"             'pure)

      output:
      (event2& "chime-status"                     'single :default #f)
      (event2& "beep"                             'combined :combine combine-beeps :default no-beep-value)

      ;; Body

      (seq&

	 ;; initializations

	 (emit& "watch-time" (get-initial-watch-time))
	 (emit& "chime-status" #f)

	 ;; main loop

	 (loop&
	    ;; normal mode
	    (abort& (now& "enter-set-watch-mode-command")
	       (loop&
		  (trap& await-case
		     (par&
			(seq&
			   (await& (now& "S"))
			   (emit& "watch-time" (increment-watch-time! (preval& "watch-time")))
			   (emit& "beep" (watch-beep (val& "watch-time") (val& "chime-status")))
			   (exit& await-case))
			(seq&
			   (await& (now& "toggle-24h-mode-command"))
			   (emit& "watch-time" (toggle-24h-mode-in-watch-time! (preval& "watch-time")))
			   (exit& await-case))
			(seq&
			   (await& (now& "toggle-chime-command"))
			   (emit& "chime-status" (not (preval& "chime-status")))
			   (exit& await-case))))))

	    ;; set-watch mode
	    ;; (in set-watch mode one might as well accept the
	    ;; commands TOGGLE_24H_MODE_COMMAND and
	    ;; TOGGLE_CHIME_COMMAND; for this one just could copy the
	    ;; corresponding cases above into the select!)

	    (abort& (now& "exit-set-watch-mode-command")
	       (emit& "start-enhancing" (initial-watch-time-position))
	       (loop&
		  (trap& await-case
		     (par&
			(seq&
			   (await& (now& "S"))
			   (emit& "watch-time" (increment-watch-time-in-set-mode (preval& "watch-time") (val& "start-enhancing")))
			   (exit& await-case))
			(seq&
			   (await& (now& "set-watch-command"))
			   (emit& "watch-time" (set-watch-time (preval& "watch-time") (val& "start-enhancing")))
			   (emit& "watch-being-set")
			   (exit& await-case))
			(seq&
			   (await& (now& "next-watch-time-position-command"))
			   (emit& "stop-enhancing" (preval& "start-enhancing"))
			   (emit& "start-enhancing" (next-watch-time-position (preval& "start-enhancing")))
			   (exit& await-case))))))
	    (emit& "stop-enhancing" (preval& "start-enhancing"))))))
