(module wristwatch_main

   (library hiphop)

   (import
      ;; The Beep data-type, constants and functions
      wristwatch_beep
      ;; The Button module
      wristwatch_button
      ;; The Watch module
      wristwatch_watch
      wristwatch_watch_datatype
      ;; The Stopwatch module
      wristwatch_stopwatch
      wristwatch_stopwatch_datatype
      ;; The Alarm module
      wristwatch_alarm
      wristwatch_alarm_datatype
      ;; The Display module
      wristwatch_display
      wristwatch_display_datatype)

   (export wristwatch&::hiphop-program)

   (main main))

(define wristwatch&::hiphop-program
   (hiphop&
      name: "wristwatch-module"

      ;; Interface

      input: (list ; buttons
		(event2& "UL" 'pure)  ; upper-left button
		(event2& "UR" 'pure)  ; upper-right button
		(event2& "LL" 'pure)  ; lower-left button
		(event2& "LR" 'pure)) ; lower-right button
      input: (list ; time events
		(event2& "HS" 'pure)  ; quartz - 1/100 second
		(event2& "S"  'pure)) ; quartz - second
      output: (list ; displays
		 (event2& "main-display"         'single :default (class-nil main-display))
		 (event2& "mini-display"         'single :default (class-nil mini-display))
		 (event2& "alphabetic-display"   'single :default "--")
		 ; Display positions
		 (event2& "start-enhancing"      'single :default 0)
		 (event2& "stop-enhancing"       'single :default 0)
		 ; Boolean indicators
		 (event2& "alarm-status"         'single :default #f)
		 (event2& "chime-status"         'single :default #f)
		 (event2& "stopwatch-run-status" 'single :default #f)
		 (event2& "stopwatch-lap-status" 'single :default #f)
		 ; Beep combination
		 (event2& "beep"                 'combined :combine combine-beeps :default no-beep-value))

      ;; Local signals

      (local&
	 (
	  (event2& "watch-mode-command"               'pure)
	  (event2& "stopwatch-mode-command"           'pure)
	  (event2& "alarm-mode-command"               'pure)

	  (event2& "toggle-24h-mode-command"          'pure)
	  (event2& "toggle-chime-command"             'pure)

	  (event2& "enter-set-watch-mode-command"     'pure)
	  (event2& "set-watch-command"                'pure)
	  (event2& "next-watch-time-position-command" 'pure)
	  (event2& "exit-set-watch-mode-command"      'pure)

	  (event2& "watch-time"                       'single :default (class-nil watch-time))
	  (event2& "watch-being-set"                  'pure)

	  (event2& "watch-start-enhancing"            'single :default (initial-watch-time-position))
	  (event2& "watch-stop-enhancing"             'single :default (initial-watch-time-position))

	  (event2& "start-stop-command"               'pure)
	  (event2& "lap-command"                      'pure)

	  (event2& "stopwatch-time"                   'single :default (class-nil stopwatch-time))

	  (event2& "toggle-alarm-command"             'pure)

	  (event2& "enter-set-alarm-mode-command"     'pure)
	  (event2& "next-alarm-time-position-command" 'pure)
	  (event2& "exit-set-alarm-mode-command"      'pure)
	  (event2& "set-alarm-command"                'pure)
	  (event2& "stop-alarm-beep-command"          'pure)

	  (event2& "alarm-time"                       'single :default (class-nil alarm-time))

	  (event2& "alarm-start-enhancing"            'single :default (initial-alarm-time-position))
	  (event2& "alarm-stop-enhancing"             'single :default (initial-alarm-time-position)))

	 ;; Body

	 (par&
	    (run& button&)
	    (run&
	       bind: "watch-start-enhancing" "start-enhancing"
	       bind: "watch-stop-enhancing" "stop-enhancing"
	       watch&)
	    (run& stopwatch&)
	    (run&
	       bind: "alarm-start-enhancing" "start-enhancing"
	       bind: "alarm-stop-enhancing" "stop-enhancing"
	       alarm&)
	    (run& display&)
	    ))))

(cond-expand
   ((not scheme2js)
(define (main args)
   (when (and (pair? args) (pair? (cdr args)))
      (bigloo-debug-set! (string->number (cadr args))))
   (let ((M (make-hiphop-machine wristwatch&)))
      (hiphop-add-event-listener! M "main-display"
	 (lambda (e) (print "Main: " e)))
      (hiphop-add-event-listener! M "mini-display"
	 (lambda (e) (print "Mini: " e)))
      (hiphop-add-event-listener! M "alphabetic-display"
	 (lambda (e) (print "Alph: " e)))
      (let loop ((i 0))
	 (when (zerofx? (remainder i 4))
	    (print "Reaction: " i)
	    (hiphop-input! M "S"))
	 (hiphop-input-and-react! M "HS")
	 (sleep (* 1000 250)) ; one HS every 25/100th second
	 (loop (+fx i 1)))))
))
