;;; Based on Stephen A. Edwards ECC: Esterel Columbia Compiler
(module esterel_to_hiphop
   (main main))


; * Esterel Grammar for ANTLR based on the grammar in the Esterel Primer, v91
; * Also includes deprecated syntax: see commentary below.
; *
; * Author: Stephen A. Edwards
; * sedwards@cs.columbia.edu
; *
; * This generates a generic AST using ANTLR&#039;s built-in facilities for AST
; * synthesis.
; *
; * Rewritten for Bigloo/Hop by Cyprien Nicolas on Nov 23 2011

(define main-module #f)

(define esterel-grammar
   (lalr-grammar
      
      (EOF COLON PERIOD ID PARALLEL COMMA (right: SEMICOLON) COLEQUALS IMPLIES POUND
	 LPAREN RPAREN LBRACKET RBRACKET QUESTION DQUESTION
	 ;(right: Tmodule)
	 Tend Tcombine Twith Tpre Ttrue Tfalse Timmediate Tnothing Tpause Thalt Tmodule
	 Temit Tsustain Tcall Tpresent Tif Tloop Teach Tpositive Trepeat Ttimes Tabort
	 Twhen Tweak Tawait Tevery Tdo Tsuspend Ttrap Tin Thandle Texit Texec Treturn
	 Tvar Tsignal Trun Tcopymodule Twatching Ttimeout Tupto Tthen Tcase Telse Telseif
	 Ttype Tconstant Tfunction Tprocedure Ttask
	 Tinput Toutput Tinputoutput Tsensor Trelation
	 Tand Tor Tnot Tmod Tmain
	 Integer FloatConst DoubleConst StringConstant
	 PLUS DASH STAR SLASH LESSTHAN GREATERTHAN LEQUAL GEQUAL NEQUAL EQUALS)
      
      (file
	 ((module-list)
	  `((module ,(string->symbol main-module)
	       (library hiphop)
	       (export main))
	    (define main ,(string->symbol main-module))
	    ,@module-list)))
      (module-list
	 ((module) (list module))
	 ((module module-list) (cons module module-list)))
      (module
	 ((Tmain Tmodule module-ident COLON declarations stmt module-end)
	  (set! main-module module-ident)
	  `(define ,(string->symbol module-ident) (hiphop& ,@declarations ,stmt)))
	 ((Tmodule module-ident COLON declarations stmt module-end)
	  `(define ,(string->symbol module-ident) (hiphop& ,@declarations ,stmt))))
      (module-end
	 ((Tend Tmodule))
	 ((PERIOD)))
      (module-ident
	 ((ID) ID))
      (declarations
	 (() '())
	 ((interface-decls declarations) (append interface-decls declarations)))
      (stmt
	 ((sequence) (cons 'seq& sequence))
	 ((sequence@p PARALLEL sequence@q) `(par& (seq& ,@p) (seq& ,@q))))
      (interface-decls
	 ((type-decls))
	 ((constant-decls))
	 ((function-decls))
	 ((procedure-decls))
	 ((task-decls))
	 ((interface-signal-decls) interface-signal-decls)
	 ((sensor-decls))
	 ((relation-decls)))
      (type-decls
	 ((Ttype type-ident-list SEMICOLON)))
      (type-ident-list
	 ((type-ident))
	 ((type-ident-list COMMA type-ident)))
      (constant-decls
	 ((Tconstant constant-decl-list SEMICOLON)))
      (constant-decl-list
	 ((constant-decl))
	 ((constant-decl-list COMMA constant-decl)))
      (function-decls
	 ((Tfunction function-decl-list SEMICOLON)))
      (function-decl-list
	 ((function-decl))
	 ((function-decl-list COMMA function-decl)))
      (procedure-decls
	 ((Tprocedure procedure-decl-list SEMICOLON)))
      (procedure-decl-list
	 ((procedure-decl))
	 ((procedure-decl-list COMMA procedure-decl)))
      (task-decls
	 ((Ttask task-decl-list SEMICOLON)))
      (task-decl-list
	 ((task-decl))
	 ((task-decl-list COMMA task-decl)))
      (interface-signal-decls
	 ((Tinput signal-decl-list SEMICOLON)       (cons :input  signal-decl-list))
	 ((Toutput signal-decl-list SEMICOLON)      (cons :output signal-decl-list))
	 ((Tinputoutput signal-decl-list SEMICOLON) (cons :inout  signal-decl-list))
	 ((Treturn signal-decl-list SEMICOLON)      (cons :return signal-decl-list)))
      (signal-decl-list
	 ((signal-decl) (list signal-decl))
	 ((signal-decl COMMA signal-decl-list) (cons signal-decl signal-decl-list)))
      (sensor-decls
	 ((Tsensor sensor-decl-list SEMICOLON)))
      (sensor-decl-list
	 ((sensor-decl))
	 ((sensor-decl-list COMMA sensor-decl)))
      (relation-decls
	 ((Trelation relation-decl-list SEMICOLON)))
      (relation-decl-list
	 ((relation-decl))
	 ((relation-decl-list COMMA relation-decl)))
      (type-ident
	 ((ID)))
      (constant-decl
	 ((constant-ident constant-init? COLON type-ident))
	 ((identifier-list COLON type-ident)))
      (constant-init?
	 (())
	 ((constant-initializer)))
      (constant-ident
	 ((ID) ID))
      (constant-initializer
	 ((EQUALS constant-atom)))
      (identifier-list
	 ((ID COMMA ID))
	 ((identifier-list COMMA ID)))
      (constant-atom
	 ((constant-literal))
	 ((signed-number)))
      (function-decl
	 ((function-ident opt-type-ident-list COLON type-ident)))
      (function-ident
	 ((ID)))
      (opt-type-ident-list
	 ((LPAREN type-ident-list RPAREN))
	 ((LPAREN RPAREN)))
      (procedure-decl
	 ((procedure-ident opt-type-ident-list type-ident)))
      (procedure-ident
	 ((ID)))
      (task-decl
	 ((task-ident opt-type-ident-list opt-type-ident-list)))
      (task-ident
	 ((ID)))
      (signal-decl
	 ((signal-ident) signal-ident)
	 ((signal-ident signal-init? COLON channel-type))
	 ((signal-ident LPAREN channel-type RPAREN)))
      (signal-init?
	 (())
	 ((signal-initializer)))
      (signal-ident
	 ((ID) ID))
      (signal-initializer
	 ((EQUALS expression)))
      (channel-type
	 ((type-ident))
	 ((Tcombine type-ident Twith combine-function)))
      (combine-function
	 ((function-ident))
	 ((predefined-combine-function)))

      ;;; Expressions
      (expression
	 ((or-expr) or-expr))
      (predefined-combine-function
	 ((PLUS) '+)
	 ((STAR) '*)
	 ((Tor)  'or)
	 ((Tand) 'and))
      (sensor-decl
	 ((sensor-ident COLON type-ident))
	 ((sensor-ident LPAREN type-ident RPAREN)))
      (sensor-ident
	 ((ID)))
      (relation-decl
	 ((implication-decl))
	 ((exclusion-decl)))
      (implication-decl
	 ((signal-ident IMPLIES signal-ident)))
      (exclusion-decl
	 ((signal-ident POUND signal-ident))
	 ((exclusion-decl POUND signal-ident)))
      (or-expr
	 ((and-expr) and-expr)
	 ((and-expr Tor or-expr) `(or ,and-expr ,or-expr)))
      (and-expr
	  ((not-expr) not-expr)
	  ((not-expr Tand and-expr) `(and ,not-expr ,and-expr)))
      (not-expr
	 ((Tnot cmp-expr) `(not& ,cmp-expr))
	 ((cmp-expr) cmp-expr))
      (cmp-expr
	 ((add-expr) add-expr)
	 ((add-expr EQUALS cmp-expr) `(= ,add-expr ,cmp-expr))
	 ((add-expr NEQUAL cmp-expr) `(not (= ,add-expr ,cmp-expr)))
	 ((add-expr LESSTHAN cmp-expr))
	 ((add-expr GREATERTHAN cmp-expr))
	 ((add-expr LEQUAL cmp-expr))
	 ((add-expr GEQUAL cmp-expr)))
      (add-expr
	 ((mul-expr) mul-expr)
	 ((mul-expr PLUS add-expr) `(+ ,mul-expr ,add-expr))
	 ((mul-expr DASH add-expr) `(- ,mul-expr ,add-expr)))
      (mul-expr
	 ((unary-expr) unary-expr)
	 ((unary-expr STAR  mul-expr) `(* ,unary-expr ,mul-expr))
	 ((unary-expr SLASH mul-expr) `(/ ,unary-expr ,mul-expr))
	 ((unary-expr Tmod  mul-expr) `(modulo ,unary-expr ,mul-expr)))
      (unary-expr
	 ((DASH unary-expr))
	 ((LPAREN expression RPAREN))
	 ((QUESTION signal-ident))
	 ((Tpre LPAREN QUESTION signal-ident RPAREN))
	 ((DQUESTION trap-ident))
	 ((function-call))
	 ((constant) constant))
      (trap-ident
	 ((ID) ID))
      (function-call
	 ((function-ident LPAREN expression-list RPAREN))
	 ((function-ident LPAREN RPAREN)))
      (expression-list
	 ((expression))
	 ((expression-list COMMA expression)))
      (constant
	 ((constant-literal) constant-literal)
	 ((unsigned-number)))
      (constant-literal
	 ((constant-ident) constant-ident)
	 ((Ttrue))
	 ((Tfalse))
	 ((string-constant) string-constant))
      (unsigned-number
	 ((Integer))
	 ((FloatConst))
	 ((DoubleConst)))
      (string-constant
	 ((StringConstant) StringConstant))
      (signed-number
	 ((unsigned-number))
	 ((DASH unsigned-number)))
      (signal-expr
	 ((s-and-expr))
	 ((signal-expr Tor s-and-expr)))
      (s-and-expr
	 ((s-unary-expr))
	 ((s-and-expr Tand s-unary-expr)))
      (s-unary-expr
	 ((signal-ident))
	 ((Tpre LPAREN signal-ident RPAREN))
	 ((Tnot s-unary-expr))
	 ((LPAREN signal-expr RPAREN)))
      (delay-expr
	 ((delay-pair) delay-pair)
	 ((bracketed-signal-expr) bracketed-signal-expr)
	 ((Timmediate bracketed-signal-expr) bracketed-signal-expr))
      (bracketed-signal-expr
	 ((signal-ident) signal-ident)
	 ((LBRACKET signal-expr RBRACKET) signal-expr))
      (delay-pair
	 ((expression bracketed-signal-expr) (list expression bracketed-signal-expr)))

      (sequence
	 ((atomic-stmt) (list atomic-stmt))
	 ((atomic-stmt SEMICOLON) (list atomic-stmt))
	 ((atomic-stmt SEMICOLON sequence) (cons atomic-stmt sequence)))
      (atomic-stmt
	 ((Tnothing) (list 'nothing&))
	 ((Tpause)   (list 'pause&))
	 ((Thalt)    (list 'halt&))
	 ((emit)     emit)
	 ((sustain) sustain)
	 ((assignment))
	 ((procedure-call))
	 ((present) present)
	 ((if-stmt) if-stmt)
	 ((loop) loop)
	 ((repeat) repeat)
	 ((abort) abort)
	 ((await) await)
	 ((every) every)
	 ((suspend) suspend)
	 ((trap) trap)
	 ((exit) exit)
	 ((exec) exec)
	 ((local-variable-decl))
	 ((local-signal-decl))
	 ((run-module))
	 ((LBRACKET stmt RBRACKET) stmt)
	 ((do-stmt)))

      (emit
	 ((Temit signal-ident LPAREN expression RPAREN))
	 ((Temit signal-ident) `(emit& ,signal-ident)))
      (sustain
	 ((Tsustain signal-ident LPAREN expression RPAREN))
	 ((Tsustain signal-ident)))

      (assignment
	 ((variable-ident COLEQUALS expression)))
      (procedure-call
	 ((Tcall procedure-ident ref-args value-args)))

      (present
	 ((Tpresent present-then else-part? present-end)))
      (present-then
	 ((present-then-part))
	 ((present-case-list)))
      (present-case-list
	 ((present-case))
	 ((present-case-list present-case)))
      (else-part?
	 (() `(nothing&))
	 ((else-part) else-part))
      (present-end
	 ((Tend Tpresent))
	 ((Tend)))

      (if-stmt
	 ((Tif expression then-part? else-if* else-part? if-end)
	  `(if& ,expression ,then-part? ,@else-if* ,else-part?)))
      (then-part?
	 (() `(nothing&))
	 ((then-part) then-part))
      (if-end
	 ((Tend Tif))
	 ((Tend)))

      (loop
	 ((Tloop stmt loop-end) `(loop& ,stmt))
	 ((Tloop stmt Teach delay-expr)
	  ;`(loop& :each ,delay-expr ,stmt)
	  `(loop&
	      (until& ,delay-expr ,stmt (halt&)))))
      (loop-end
	 ((Tend Tloop))
	 ((Tend)))

      (repeat
	 ((positive? Trepeat expression Ttimes stmt repeat-end)))
      (positive?
	 (())
	 ((Tpositive)))
      (repeat-end
	 ((Tend Trepeat))
	 ((Tend)))

      (abort
	 ((Tabort stmt Twhen abort-one-case-strong)
	  `(abort& ,abort-one-case-strong ,stmt))
	 ((Tabort stmt Twhen acase-list abort-end))
	 ((Tweak Tabort stmt Twhen abort-one-case-weak))
	 ((Tweak Tabort stmt Twhen acase-list abort-end-weak)))
      (abort-end
	 ((Tend Tabort))
	 ((Tend)))
      (abort-end-weak
	 ((Tend Tweak Tabort))
	 ((Tend Tabort))
	 ((Tend)))

      (await
	 ((Tawait await-one-case) `(await& ,await-one-case))
	 ((Tawait acase-list await-end) `(await& ,acase-list)))
      (await-end
	 ((Tend Tawait))
	 ((Tend)))

      (every
	 ((Tevery delay-expr Tdo expression every-end)))
      (every-end
	 ((Tend Tevery))
	 ((Tend)))

      (suspend
	 ((Tweak Tsuspend stmt Twhen delay-expr)
	  `(suspend& :weak #t ,delay-expr ,stmt))
	 ((Tsuspend stmt Twhen delay-expr)
	  `(suspend& ,delay-expr ,stmt)))

      (trap
	 ((Ttrap trap-decl-list Tin stmt trap-handler+ trap-end)
	  `(trap& ,trap-decl-list (handler& ,@trap-handler+) ,stmt))
	 ((Ttrap trap-decl-list Tin stmt trap-end)
	  `(trap& ,trap-decl-list ,stmt)))
      (trap-handler+
	 ((trap-handler) (list trap-handler))
	 ((trap-handler trap-handler+) (cons trap-handler trap-handler+)))
      (trap-end
	 ((Tend Ttrap))
	 ((Tend)))

      (exit
	 ((Texit trap-ident) `(exit& ,trap-ident))
	 ((Texit trap-ident LPAREN expression RPAREN)))

      (exec
	 ((Texec exec-one-case))
	 ((Texec exec-case-list exec-end)))
      (exec-case-list
	 ((exec-case))
	 ((exec-case-list exec-case)))
      (exec-end
	 ((Tend Texec))
	 ((Tend)))

      (local-variable-decl
	 ((Tvar variable-decl-list Tin stmt Tend Tvar))
	 ((Tvar variable-decl-list Tin stmt Tend)))

      (local-signal-decl
	 ((Tsignal signal-decl-list Tin stmt Tend Tsignal))
	 ((Tsignal signal-decl-list Tin stmt Tend)))

      (run-module
	 ((Trun copy-module? module-ident slash-module-ident? run-renaming?)))
      (copy-module?
	 (())
	 ((Tcopymodule)))
      (slash-module-ident?
	 (())
	 ((SLASH module-ident)))
      (run-renaming?
	 (())
	 ((LBRACKET renaming-list RBRACKET)))
      (renaming-list
	 ((renaming))
	 ((renaming-list SEMICOLON renaming)))

      (do-stmt
	 ((Tdo stmt Twatching delay-expr timeout-expr?))
	 ((Tdo stmt Tupto delay-expr)))
      (timeout-expr?
	 (())
	 ((Ttimeout stmt Tend Ttimeout))
	 ((Ttimeout stmt Tend)))

      (variable-ident
	 ((ID)))

      (ref-args
	 ((LPAREN variable-ident-list RPAREN))
	 ((LPAREN RPAREN)))
      (variable-ident-list
	 ((variable-ident))
	 ((variable-ident-list COMMA variable-ident)))

      (value-args
	 ((LPAREN expression-list RPAREN))
	 ((LPAREN RPAREN)))

      (present-then-part
	 ((present-event))
	 ((present-event Tthen stmt)))

      (present-case
	 ((Tcase present-event))
	 ((Tcase present-event Tdo stmt)))

      (else-part
	 ((Telse stmt) stmt))

      (present-event
	 ((signal-expr))
	 ((LBRACKET signal-expr RBRACKET)))

      (then-part
	 ((Tthen stmt) stmt))

      (else-if*
	 (() '())
	 ((else-if else-if*) (cons else-if else-if*)))
      (else-if
	 ((Telseif expression Tthen stmt)))

      (abort-one-case-strong
	 ((delay-expr) delay-expr)
	 ((delay-expr Tdo stmt abort-end)))

      (acase-list
	 ((acase))
	 ((acase-list acase)))
      (acase
	 ((Tcase delay-expr))
	 ((Tcase delay-expr Tdo stmt)))

      (abort-one-case-weak
	 ((delay-expr))
	 ((delay-expr Tdo stmt abort-end-weak)))

      (await-one-case
	 ((delay-expr) delay-expr)
	 ((delay-expr Tdo stmt await-end)))

      (trap-decl-list
	 ((trap-decl) (list trap-decl))
	 ((trap-decl COMMA trap-decl-list) (cons trap-decl trap-decl-list)))

      (trap-handler
	 ((Thandle trap-event Tdo stmt)))

      (trap-decl
	 ((trap-ident trap-init? COLON channel-type))
	 ((trap-ident trap-init?) trap-ident))
      (trap-init?
	 (())
	 ((trap-initializer)))
      (trap-initializer
	 ((COLEQUALS expression)))

      (trap-event
	 ((e-and))
	 ((trap-event Tor e-and)))
      (e-and
	 ((e-unary))
	 ((e-and Tand e-unary)))
      (e-unary
	 ((trap-ident))
	 ((LPAREN trap-event RPAREN))
	 ((Tnot e-unary)))

      (exec-one-case
	 ((task-ident ref-args value-args Treturn signal-ident))
	 ((task-ident ref-args value-args Treturn signal-ident Tdo stmt exec-end)))

      (exec-case
	 ((Tcase task-ident ref-args value-args Treturn signal-ident))
	 ((Tcase task-ident ref-args value-args Treturn signal-ident Tdo stmt)))

      (variable-decl-list
	 ((variable-decls))
	 ((variable-decl-list COMMA variable-decls)))
      (variable-decls
	 ((variable-decl-list-2 COLON type-ident)))
      (variable-decl-list-2
	 ((variable-decl))
	 ((variable-decl-list-2 COMMA variable-decl)))
      (variable-decl
	 ((variable-ident))
	 ((variable-ident variable-initializer)))
      (variable-initializer
	 ((COLEQUALS expression)))

      (renaming
	 ((Ttype type-renaming-list))
	 ((Tconstant constant-renaming-list))
	 ((Tfunction function-renaming-list))
	 ((Tprocedure procedure-renaming-list))
	 ((Ttask task-renaming-list))
	 ((Tsignal signal-renaming-list)))

      (type-renaming-list
	 ((type-renaming))
	 ((type-renaming-list COMMA type-renaming)))
      (type-renaming
	 ((type-ident SLASH type-ident)))

      (constant-renaming-list
	 ((constant-renaming))
	 ((constant-renaming-list COMMA constant-renaming)))
      (constant-renaming
	 ((constant-ident SLASH constant-ident)))

      (function-renaming-list
	 ((function-renaming))
	 ((function-renaming-list COMMA function-renaming)))
      (function-renaming
	 ((function-ident SLASH function-ident))
	 ((predefined-function SLASH function-ident)))

      (procedure-renaming-list
	 ((procedure-renaming))
	 ((procedure-renaming-list COMMA procedure-renaming)))
      (procedure-renaming
	 ((procedure-ident SLASH procedure-ident)))

      (task-renaming-list
	 ((task-renaming))
	 ((task-renaming-list COMMA task-renaming)))
      (task-renaming
	 ((task-ident SLASH task-ident)))

      (signal-renaming-list
	 ((signal-renaming))
	 ((signal-renaming-list COMMA signal-renaming)))
      (signal-renaming
	 ((signal-ident SLASH signal-ident)))

      (predefined-function
	 ((Tand))
	 ((Tor))
	 ((Tnot))
	 ((PLUS))
	 ((DASH))
	 ((STAR))
	 ((SLASH))
	 ((Tmod))
	 ((LESSTHAN))
	 ((GREATERTHAN))
	 ((LEQUAL))
	 ((GEQUAL))
	 ((NEQUAL))
	 ((EQUALS)))))

(define esterel-lexer
   (regular-grammar
	 ((ident (: alpha (* (or alpha digit #\_))))
	  (blank (+ (in #" \t\n\r"))))
      
      ("end" 'Tend)
      ("combine" 'Tcombine)
      ("with" 'Twith)
      ("pre" 'Tpre)
      ("true" 'Ttrue)
      ("false" 'Tfalse)
      ("immediate" 'Timmediate)
      ("nothing" 'Tnothing)
      ("pause" 'Tpause)
      ("halt" 'Thalt)
      ("main" 'Tmain)
      ("module" 'Tmodule)
      ("emit" 'Temit)
      ("sustain" 'Tsustain)
      ("call" 'Tcall)
      ("present" 'Tpresent)
      ("if" 'Tif)
      ("loop" 'Tloop)
      ("each" 'Teach)
      ("positive" 'Tpositive)
      ("repeat" 'Trepeat)
      ("times" 'Ttimes)
      ("abort" 'Tabort)
      ("when" 'Twhen)
      ("weak" 'Tweak)
      ("await" 'Tawait)
      ("every" 'Tevery)
      ("do" 'Tdo)
      ("suspend" 'Tsuspend)
      ("trap" 'Ttrap)
      ("in" 'Tin)
      ("handle" 'Thandle)
      ("exit" 'Texit)
      ("exec" 'Texec)
      ("return" 'Treturn)
      ("var" 'Tvar)
      ("signal" 'Tsignal)
      ("run" 'Trun)
      ("copymodule" 'Tcopymodule)
      ("watching" 'Twatching)
      ("timeout" 'Ttimeout)
      ("upto" 'Tupto)
      ("then" 'Tthen)
      ("case" 'Tcase)
      ("else" 'Telse)
      ("elseif" 'Telseif)
      ("type" 'Ttype)
      ("constant" 'Tconstant)
      ("function" 'Tfunction)
      ("procedure" 'Tprocedure)
      ("task" 'Ttask)
      ("input" 'Tinput)
      ("output" 'Toutput)
      ("inputoutput" 'Tinputoutput)
      ("sensor" 'Tsensor)
      ("relation" 'Trelation)
      ("and" 'Tand)
      ("or" 'Tor)
      ("not" 'Tnot)
      ("mod" 'Tmod)
      (ident (cons 'ID (the-string)))
      (blank (ignore))
      (#\. 'PERIOD)
      ((: #\| #\|) 'PARALLEL)
      (#\, 'COMMA)
      (#\; 'SEMICOLON)
      ((: #\: #\=) 'COLEQUALS)
      (#\: 'COLON)
      ((: #\= #\>) 'IMPLIES)
      (#\# 'POUND)
      (#\( 'LPAREN)
      (#\) 'RPAREN)
      (#\{ 'LBRACKET)
      (#\} 'RBRACKET)
      ((: #\? #\?) 'DQUESTION)
      (#\? 'QUESTION)
      (#\+ 'PLUS)
      (#\- 'DASH)
      (#\* 'STAR)
      (#\/ 'SLASH)
      (#\< 'LESSTHAN)
      (#\> 'GREATERTHAN)
      ((: #\< #\=) 'LEQUAL)
      ((: #\> #\=) 'GEQUAL)
      ((: #\< #\>) 'NEQUAL)
      (#\= 'EQUALS)
      (else
       (let ((c (the-failure)))
	  (if (eof-object? c)
	      c
	      (error 'rgc "Illegal character" c))))))

(define (main args)
   ;(bigloo-debug-set! 101)
   (let ((files (list "ex1" "ex2" "ex3" "ex_2_1_2" "abro")))
      (for-each
	 (lambda (file)
	    (let ((src (make-file-path "." "Esterel" "PureV5" (string-append file ".strl"))))

	       (print "Reading " src)
	       (let ((ip (open-input-file src)))
		  (let ((transform (read/lalrp esterel-grammar esterel-lexer ip)))
		     (map pp transform)
		     (with-output-to-file "/tmp/test.hip"
			(lambda ()
			   (map pp transform)))))))
	 files)))
