(module hiphop_tests
   (library hiphop)

   (export *tests*
	   (run-tests #!key ignore-failures? port))

(cond-expand
   (scheme2js
   (js
      console
      (scheme2js-debug "__sc_traceLevel")
      (*stdout* "SC_DEFAULT_OUT")
      (*stderr* "SC_ERROR_OUT"))
))

   (main main))

(define add-event-listener! hiphop-add-event-listener!)

(define *tests* '())

(define (dprint msg) (display msg (current-error-port)))

(cond-expand
   (scheme2js
(define (current-output-port) *stdout*)
(define (current-error-port) *stderr*)
(define (flush-output-port p) #f)
(define (fprint port msg . msgs)
   (display msg port)
   (for-each (lambda (m) (display m port)) msgs)
   (newline port))
(define (fprintf port fmt . args)
   (fprint port (apply format fmt args)))
(define (bigloo-debug) scheme2js-debug)
(define (bigloo-debug-set! v) (set! scheme2js-debug v))
(define (bigloo-trace-color-set! bool) bool)

(define (with-error-to-port port thunk)
   (with-output-to-port port thunk))

(define (sort precedes? ls) ; merge sort
   (define (split ls)
      (let loop ((rest ls)
		 (left '())
		 (right '()))
	 (cond ((null? rest)
		(cons left right))
	       ((null? (cdr rest))
		(cons (cons (car rest) left) right))
	       (else
		(loop (cddr rest)
                   (cons (car rest) left)
                   (cons (cadr rest) right))))))
   (define (merge list-1 list-2)
      (let loop ((source-1 list-1)
		 (source-2 list-2)
		 (so-far '()))
	 (cond ((null? source-1)
		(append (reverse so-far) source-2))
	       ((null? source-2)
		(append (reverse so-far) source-1))
	       (else
		(let ((car-1 (car source-1))
		      (car-2 (car source-2)))
		   (if (precedes? car-1 car-2)
		       (loop (cdr source-1)
			  source-2
			  (cons car-1 so-far))
		       (loop source-1
			  (cdr source-2)
			  (cons car-2 so-far))))))))
   (if (and (pair? ls) (pair? (cdr ls)))
       (let ((parts (split ls)))
	  (merge
	     (sort precedes? (car parts))
	     (sort precedes? (cdr parts))))
       ls))
(define pp print)
))

(define (compare-alist list1 list2)
   (define (compare-alist-presence list1 list2)
      (let loop ((l1 list1) (l2 list2))
	 (cond
	    ((and (null? l1) (null? l2)) #f) ; success
	    ((or  (null? l1) (null? l2))
	     (format "Different length\n  Expected: ~s\n  Got:      ~s" list1 list2))
	    ((and (string? (car l1)) (string? (car l2)) (string=? (car l1) (car l2)))
	     (loop (cdr l1) (cdr l2)))
	    ((equal? (car l1) (car l2))
	     (loop (cdr l1) (cdr l2)))
	    (else
	     (format "\n  Expected: ~s\n  Got:      ~s" list1 list2)))))

   (define (compare-alist-values list1 list2)
      (let loop ((l1 list1) (l2 list2))
	 (cond
	    ((and (null? l1) (null? l2)) #f) ; success
	    ((or (null? l1) (null? l2))
	     (format "Different length\n  Expected: ~s\n  Got:      ~s" list1 list2))
	    (else
	     (let ((f1 (car l1)) (f2 (car l2)))
		(cond
		   ((or (not (pair? f1)) (not (pair? f2)))
		    (format "Missing value\n  Expected: ~s\n  Got:      ~s" list1 list2))
		   ((string=? (car f1) (car f2))
		    (let ((val1 (cdr f1)) (val2 (cdr f2)))
		       (let leep ((v2 val2))
			  (if (pair? v2)
			      (if (member (car v2) val1)
				  (leep (cdr v2))
				  (format "Missing value: ~s\n  Expected: ~s\n  Got:      ~s" (car v2) list1 list2))
			      (if (= (length val1) (length val2))
				  (loop (cdr l1) (cdr l2))
				  (format "Missing values\n  Expected: ~s\n  Got:      ~s"
				     list1 list2))))))
		   (else (error "compare-alist-values" "Dunno what to do with input"
			    (vector list1 list2)))))))))

   (let ((sorted-list1 (if (pair? list1)
			   (if (pair? (car list1))
			       (sort (lambda (e1 e2) (string<? (car e1) (car e2))) list1)
			       (sort string<? list1))
			   list1))
	 (sorted-list2 (if (pair? list2)
			   (if (pair? (car list2))
			       (sort (lambda (e1 e2) (string<? (car e1) (car e2))) list2)
			       (sort string<? list2))
			   list2)))

      (if (and (pair? list1) (pair? (car list1))) ; valued alists
	  (compare-alist-values sorted-list1 sorted-list2)
	  (compare-alist-presence sorted-list1 sorted-list2))))

(define (make-tests name #!key program events)
   (let ((len (vector-length events)))
      (let loop ((i 0))
	 (when (< i len)
	    (let ((args (vector-ref events i))
		  (name (format "~a-t~a" name (+fx i 1))))
	       (apply make-test name :program program args))
	    (loop (+fx i 1))))))

(define (make-test name #!key
	   (reactions 1) ; number of reactions to run
	   invec         ; vector of alists, vector[i] is used at instant i
	   outvec        ; vector of alists of expected outputs
	   compile-error ; The error message this test shall raise at compile-time
	   causality-error ; The error message this test shall raise at run-time
	   interface-error ; The error message this test shall raise at run-time
	   (abort-on-error #t) ; shall the test abort if an error occurs?
	   (expected-output-string "") ; expected output string at the end of all reactions
	   (dynamic? #f) ; Is the machine dynamic?
	   program) ; the program to run
   (when (=fx reactions 1)
      (set! reactions (max 1
			 (if invec (vector-length invec) 0)
			 (if outvec (vector-length outvec) 0))))
   (unless invec  (set! invec  (make-vector reactions '())))
   (unless outvec (set! outvec (make-vector reactions '())))
   (unless (=fx (vector-length invec) (vector-length outvec))
      (error "make-test" "Malformed test" name))

   (let* ((prgm::hiphop-program
	    (cond
	       ((isa? program hiphop-program) program)
	       ;((procedure? program) (program))
	       (else (error "make-test" "Illegal program" program))))
	  (test
	     (lambda args
		(bind-exit (early-abort)
		   (let* ((debug? (pair? args))
			  (output-string (open-output-string))
			  (M::hiphop-machine%
			     (with-handler
				(lambda (e::&error)
				   (when (and compile-error
					      (isa? e &hiphop-compile-error))
				      (when (string=? e.msg compile-error)
					 (early-abort #t)) ; Expected message
				      (print "*** ERROR in test: " name
					 "\nExpected error message -- " compile-error
					 "\nGot message -- " e.msg))
				   (raise e))
				;(dprint prgm)
				(if debug?
				    (make-hiphop-machine prgm :raw-errors? #t :dynamic? dynamic?)
				    (let ((err-str (open-output-string)))
				       (with-error-to-port err-str
					  (lambda ()
					     (let ((m (make-hiphop-machine prgm :raw-errors? #t :dynamic? dynamic?))
						   (err (get-output-string err-str)))
						(unless (zerofx? (string-length err))
						   ; We only display this message if the compilation succeeded
						   (print "*** WARNING: Some data has been output on the error port"))
						m)))))))
			 (outs '())
			 (on-error #f))

			; Set event listeners
		      (for-each
			 (lambda (output)
			    (let ((output-event::bstring "")
				  (multi? #t))
			       (cond
				  ((isa? output hiphop-user-event)
				   (let ((uevent::hiphop-user-event output))
				      (set! output-event uevent.name)
				      (set! multi? uevent.multi?)))
				  ((isa? output hiphop-user-event2)
				   (let ((uevent::hiphop-user-event2 output))
				      (set! output-event uevent.name)
				      (set! multi? (isa? uevent hiphop-user-event-multi))))
				  (else
				   (set! output-event output)))
			       (add-event-listener! M output-event
				  (lambda (value)
				     (set! outs
					(cond
					   (multi?
					      (cons (cons output-event value) outs))
					   (value
					      (cons (list output-event value) outs))
					   (else
					    (cons output-event outs)))))
				  )))
			 prgm.formal-outputs)

		      (bind-exit (abort)
			 (let loop ((instant 0))
			    (if (< instant reactions)
				(let ((vin  (vector-ref invec instant))
				      (vout (vector-ref outvec instant)))
				   ; Sets invec
				   (map (lambda (e)
					   (if (pair? e)
					       (apply hiphop-input! M (car e) (cdr e))
					       (hiphop-input! M e)))
				      vin)
				   ; Run the reaction
				   (when debug?
				      (print "Reaction " instant ". Inputs: " vin)
				      (print "  Expected outputs: " vout))
				   (with-handler
				      (lambda (e::&error)
					 (cond
					    ((and causality-error (isa? e &hiphop-causality-error))
					     (if (string=? e.msg causality-error)
						 (early-abort #t) ; Expected message
						 (print "*** ERROR in test: " name
						    "\nExpected error message -- "
						    causality-error "\nGot message -- " e.msg)))
					    ((and interface-error (isa? e &hiphop-interface-error))
					     (if (string=? e.msg interface-error)
						 (early-abort #t) ; Expected message
						 (print "*** ERROR in test: " name
						    "\nExpected error message -- "
						    interface-error "\nGot message -- " e.msg))))
					 (raise e))
				      (with-output-to-port output-string
					 (lambda () (hiphop-react! M))))
				   ; check outvec
				   (let ((diff (compare-alist vout (reverse outs))))
				      (when diff
					 (set! on-error #t)
					 (print "Unexpected result: " diff)

					 (when debug?
					    (print "Dumping environment")
					    (pp M.environment.events))

					 (when abort-on-error
					    (print "Aborting at instant " instant)
					    (abort #f)))

				      ; Prepare next reaction
				      (set! outs '())
				      (loop (+ instant 1))))
				(let ((outstr (close-output-port output-string)))
				   (unless (string=? outstr expected-output-string)
				      (print  "*** ERROR in test: " name
					 (format "\nExpected output -- ~s\nGot -- ~s"
					    expected-output-string outstr))
				      (error "run-test" "Unexpected output" outstr))
				   (when debug?
				      (print "Test terminated at instant " instant))))))
		      (not on-error))))))
      (if (pair? *tests*)
	  (append! *tests* (list (cons name test)))
	  (set! *tests* (list (cons name test))))))

(make-test "rename-expanded" :program
   (hiphop& :output (event2& "O1" 'multi) (event2& "O2" 'combined :combine * :default 1)
      (seq&
	 (emit& "O1" 1)
	 (emit& "O2" 3)
	 (local& ((event2& "L1" 'single :default 0) (event2& "L2" 'single :default 0))
	    (par&
	       (seq&
		  (emit& "L1" 2)
		  (emit& "L2" 4))
	       (if& (now& "L1")
		  (emit& "O1" (val& "L1")))
	       (if& (now& "L2")
		  (emit& "O2" (val& "L2")))))))
   :outvec '#((("O1" 1 2) ("O2" 12))))

(make-test "rename-expand" :program
   (hiphop& :output  (event2& "O1" 'multi) (event2& "O2" 'multi)
      (seq&
	 (emit& "O1" 1)
	 (emit& "O2" 3)
	 (run& bind: "L1" "O1" bind: "L2" "O2"
	    (hiphop& :output (event2& "L1" 'multi) (event2& "L2" 'multi)
	       (seq&
		  (emit& "L1" 2)
		  (emit& "L2" 4))))))
   :outvec '#((("O1" 1 2) ("O2" 3 4))))

(make-test "Gustave" :program
   (hiphop&
      :output (event2& "S" 'single :default 0)
              (event2& "T" 'single :default 0)
              (event2& "U" 'single :default 0)
      (par&
	 (emit& "S" 1)
	 (if& (> (val& "S") 0)
	    (emit& "T" (val& "S")))
	 (if& (now& "S")
	    (nothing&)
	    (seq&
	       (emit& "U" (> (val& "T") 0))
	       (emit& "S" 2)))))
   :outvec '#((("S" 1) ("T" 1))))

(make-test "emit-undeclared"
   :program (hiphop& (emit& "O"))
   :compile-error "Event is not declared")

;;; Unit test
(make-test "emit-no-value" :program
   (hiphop& :output (event2& "O" 'pure)
      (emit& "O"))
   :outvec '#(("O") ()))

(make-test "emit-value-const" :program
   (hiphop& :output (event2& "O" 'single :default 0)
      (emit& "O" 3))
   :outvec '#((("O" 3)) ()))

(make-test "emit-value-proc" :program
   (hiphop& :output (event2& "O" 'single :default 0)
      (emit& "O" (+ 2 3)))
   :outvec '#((("O" 5)) ()))

(make-tests "emit-value-val&" :program
   (hiphop&
      :input (event2& "I" 'multi)
      :output (event2& "O" 'single :default '())
      (emit& "O" (val& "I")))
   :events (vector
	      (list :invec '#(())          :outvec '#((("O" ()))))
	      (list :invec '#((("I" 2)))   :outvec '#((("O" (2)))))
	      (list :invec '#((("I" 1 2))) :outvec '#((("O" (2 1)))))
	      ))

(make-test "emit-value-val&+" :program
   (hiphop&
      :input (event2& "I" 'single :default 0)
      :output (event2& "O" 'single :default 0)
      (emit& "O" (+ (val& "I") 1)))
   :invec '#((("I" 2)))
   :outvec '#((("O" 3))))

(make-test "sequence/emit" :program
   (hiphop& :output (event2& "O1" 'pure) (event2& "O2" 'pure)
      (seq&
	 (emit& "O1")
	 (emit& "O2")))
   :outvec '#(("O1" "O2") ()))

(make-test "nothing/emit" :program
   (hiphop& :output (event2& "O" 'pure)
      (seq&
	 (nothing&)
	 (emit& "O")))
   :outvec '#(("O") ()))

(make-test "pause/emit" :program
   (hiphop& :output (event2& "O" 'pure)
      (seq&
	 (pause&)
	 (emit& "O")))
   :outvec '#(() ("O") ()))

(define IfEmit
   (hiphop& :input (event2& "I" 'pure) :output (event2& "O1" 'pure) (event2& "O2" 'pure)
      (if& (now& "I")
	 (emit& "O1")
	 (emit& "O2"))))

(make-tests "if-emit"
   :program IfEmit
   :events (vector ; the way to specify different combination of inputs and output vectors
	      (list :invec '#(("I")) :outvec '#(("O1")))
	      (list :invec '#(())    :outvec '#(("O2")))
	      (list :invec '#(("I") ()) :outvec '#(("O1") ()))
	      (list :invec '#(() ("I")) :outvec '#(("O2") ()))
	      ))

(make-test "if-value-true" :program
   (hiphop& :output (event2& "OK" 'pure) (event2& "NOK" 'pure)
      (if& #t
	 (emit& "OK")
	 (emit& "NOK")))
   :outvec '#(("OK")))

(make-test "if-value-false" :program
   (hiphop& :output (event2& "OK" 'pure) (event2& "NOK" 'pure)
      (if& #f
	 (emit& "OK")
	 (emit& "NOK")))
   :outvec '#(("NOK")))

(let ((IfVal (hiphop&
		:input (event2& "I" 'combined :combine (lambda (a b) (or a b)) :default #f)
		:output (event2& "true" 'pure) (event2& "false" 'pure)
		(if& (val& "I")
		   (emit& "true")
		   (emit& "false")))))
   (make-tests "if-val"
      :program IfVal
      :events (vector
		 (list :invec '#((("I" #t))) :outvec '#(("true")))
		 (list :invec '#((("I" #f))) :outvec '#(("false")))
		 )))

(make-tests "suspend" :program
   (hiphop&
      :input  (event2& "I" 'pure)
      :output (event2& "O" 'pure)
      (suspend& (now& "I")
	 (sustain& "O")))
   :events (vector
	      (list :invec '#(()) :outvec '#(("O")))    ;#1
	      (list :invec '#(("I")) :outvec '#(("O"))) ;#2
	      ; 2 reactions
	      (list :invec '#(() ()) :outvec '#(("O") ("O")))    ;#3
	      (list :invec '#(() ("I")) :outvec '#(("O") ()))    ;#4
	      (list :invec '#(("I") ()) :outvec '#(("O") ("O"))) ;#5
	      (list :invec '#(("I") ("I")) :outvec '#(("O") ())) ;#6
	      ; 3 reactions
	      (list :invec '#(()    ()    ())    :outvec '#(("O") ("O") ("O"))) ;# 7
	      (list :invec '#(()    ()    ("I")) :outvec '#(("O") ("O") ()))    ;# 8
	      (list :invec '#(()    ("I") ())    :outvec '#(("O") ()    ("O"))) ;# 9
	      (list :invec '#(()    ("I") ("I")) :outvec '#(("O") ()    ()))    ;#10
	      (list :invec '#(("I") ()    ())    :outvec '#(("O") ("O") ("O"))) ;#11
	      (list :invec '#(("I") ()    ("I")) :outvec '#(("O") ("O") ()))    ;#12
	      (list :invec '#(("I") ("I") ())    :outvec '#(("O") ()    ("O"))) ;#13
	      (list :invec '#(("I") ("I") ("I")) :outvec '#(("O") ()    ()))))  ;#14

(make-tests "suspend-imm" :program
   (hiphop&
      :input  (event2& "I" 'pure)
      :output (event2& "O" 'pure)
      (suspend& :immediate #t (now& "I")
	 (sustain& "O")))
   :events (vector
	      (list :invec '#(()) :outvec '#(("O")))	;# 1
	      (list :invec '#(("I")) :outvec '#(()))	;# 2
	      ; 2 reactions
	      (list :invec '#(() ()) :outvec '#(("O") ("O")))	;# 3
	      (list :invec '#(() ("I")) :outvec '#(("O") ()))	;# 4
	      (list :invec '#(("I") ()) :outvec '#(() ("O")))	;# 5
	      (list :invec '#(("I") ("I")) :outvec '#(() ()))	;# 6
	      ; 3 reactions
	      (list :invec '#(()    ()    ())    :outvec '#(("O") ("O") ("O")))	;# 7
	      (list :invec '#(()    ()    ("I")) :outvec '#(("O") ("O") ()))	;# 8
	      (list :invec '#(()    ("I") ())    :outvec '#(("O") ()    ("O")))	;# 9
	      (list :invec '#(()    ("I") ("I")) :outvec '#(("O") ()    ()))	;#10
	      (list :invec '#(("I") ()    ())    :outvec '#(() ("O") ("O")))	;#11
	      (list :invec '#(("I") ()    ("I")) :outvec '#(() ("O") ()))	;#12
	      (list :invec '#(("I") ("I") ())    :outvec '#(() ()    ("O")))	;#13
	      (list :invec '#(("I") ("I") ("I")) :outvec '#(() ()    ()))))	;#14

(make-tests "sustain" :program
   (hiphop&
      :output (event2& "O" 'pure)
      (sustain& "O"))
   :events (let ((output '("O")))
	      (list->vector
		 (map (lambda (i) (list :outvec (make-vector i output)))
		    (iota 9 1)))))


(define AwaitEmit
   (hiphop&
      :input  (event2& "I" 'pure)
      :output (event2& "O" 'pure)
      (seq&
	 (await& (now& "I"))
	 (emit& "O"))))

(make-tests "await-emit" :program AwaitEmit
   :events (vector
	      (list :invec '#(()) :outvec '#(()))
	      (list :invec '#(("I")) :outvec '#(()))
	      (list :invec '#(() ()) :outvec '#(() ()))
	      (list :invec '#(() ("I")) :outvec '#(() ("O")))
	      (list :invec '#(() ("I") ()) :outvec '#(() ("O") ()))))

(make-test "loop-instantaneous"
   :program (hiphop&
	       :output (event2& "O" 'pure)
	       (loop& (emit& "O")))
   :causality-error "Potentially unsafe loop")

(make-test "loop" :program (hiphop& (loop& (pause&))) :reactions 2)

(make-test "loop-atom" :program
   (let ((i 0))
      (hiphop& :output (event2& "O" 'single :default 0)
	 (seq&
	    (atom& (set! i 0))
	    (loop&
	       (atom& (display "ok"))
	       (emit& "O" (begin (set! i (+fx i 1)) i))
	       (pause&)))))
   :expected-output-string "okokokok"
   :outvec '#((("O" 1)) (("O" 2)) (("O" 3)) (("O" 4))))

(make-test "par" :program
   (hiphop& :output (event2& "O1" 'pure) (event2& "O2" 'pure)
      (par&
	 (emit& "O1")
	 (emit& "O2")))
   :outvec '#(("O1" "O2")))

(make-test "par2" :program
   (hiphop&
      :output (event2& "O" 'pure)
      :input  (event2& "a" 'pure) (event2& "b" 'pure)
      (seq&
	 (par&
	    (await& (now& "a"))
	    (await& (now& "b")))
	 (emit& "O")))
   :invec '#(() ("a") ("b")) :outvec '#(() () ("O")) :reactions 3)

(define AbortSustain
   (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
      (abort& (now& "I") (sustain& "O"))))

(make-tests "abort-sustain"
   :program AbortSustain
   :events (vector
	      (list :invec '#(()) :outvec '#(("O")))
	      (list :invec '#(("I")) :outvec '#(("O")))
	      (list :invec '#(() ()) :outvec '#(("O") ("O")))
	      (list :invec '#(() ("I")) :outvec '#(("O") ()))))

(define AbortImmSustain
   (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
      (abort& :immediate #t (now& "I") (sustain& "O"))))

(make-tests "abort-immediate-sustain"
   :program AbortImmSustain
   :events (vector
	      (list :invec '#(()) :outvec '#(("O")))
	      (list :invec '#(("I")) :outvec '#(()))
	      (list :invec '#(() ()) :outvec '#(("O") ("O")))
	      (list :invec '#(() ("I")) :outvec '#(("O") ()))))

(define AbortHandler
   (hiphop& :input (event2& "I" 'pure) :output (event2& "E" 'pure) (event2& "O" 'pure)
      (abort& (now& "I") (handler& (emit& "E")) (emit& "O"))))

(make-tests "abort-handler"
   :program AbortHandler
   :events (vector
	      (list :invec '#(()) :outvec '#(("O")))
	      (list :invec '#(("I")) :outvec '#(("O")))
	      (list :invec '#(() ("I")) :outvec '#(("O") ()))))

(define AbortImmHandler
   (hiphop&
      :input (event2& "I" 'pure)
      :output (event2& "E" 'pure) (event2& "O" 'pure)
      (abort& :immediate #t (now& "I")
	 (handler& (emit& "E"))
	 (emit& "O"))))

(make-tests "abort-immediate-handler"
   :program AbortImmHandler
   :events (vector
	      (list :invec '#(()) :outvec '#(("O")))
	      (list :invec '#(("I")) :outvec '#(("E")))
	      (list :invec '#(() ("I")) :outvec '#(("O") ()))))

(make-test "until" :program
   (hiphop& :input (event2& "R" 'pure)
      :output (event2& "O" 'pure)
      (until& :immediate #t (now& "R")
	 (seq&
	    (pause&)
	    (emit& "O"))))
   :invec '#(("R") () ()))

(make-test "loop/until" :program
   (hiphop& :input (event2& "break" 'pure) :output (event2& "O" 'pure)
      (seq&
	 (until& (now& "break")
	    (loop&
	       (pause&)))
	 (emit& "O")))
   :invec '#(() ("break"))
   :outvec '#(() ("O")))

(make-test "trap" :program
   (hiphop&
      (seq&
	 (trap& U
	    (par&
	       (trap& T
		  (par&
		     (nothing&)
		     (pause&)
		     (exit& T)
		     (exit& U)))
	       (exit& U)))
	 (pause&)))
   :reactions 3)

(make-tests "untilR-awaitI-emitO" :program
   (hiphop&
      :input (event2& "R" 'pure) (event2& "I" 'pure)
      :output (event2& "O" 'pure) (event2& "handle" 'multi)
	       (until& (now& "R")
		  (handler& (emit& "handle" 1))
		  (await& (now& "I"))
		  (handler& (emit& "handle" 2)) ; Awful syntax, but still works
		  (emit& "O")
		  (handler& (emit& "handle" 3))))
   :events (vector
	      (list :invec '#(() ("I") ("R")) :outvec '#(() ("O") ()))
	      (list :invec '#(() ("R") ("I")) :outvec '#(() (("handle" 1 2 3)) ()))
	      (list :invec '#(() ("R" "I") ()) :outvec '#(() ("O") ()))))

(make-test "par3" :program (hiphop& :output (event2& "O" 'pure) :input (event2& "a" 'pure) (event2& "b" 'pure) (event2& "R" 'pure)
			      (until& (now& "R")
				 (seq&
				    (par&
				       (await& (now& "a"))
				       (await& (now& "b")))
				    (emit& "O"))))
   :invec '#(() ("a") ("R") ("b") ("a")))

(define ABRO
   (hiphop&
      :input (event2& "A" 'pure) (event2& "B" 'pure) (event2& "R" 'pure)
      :output (event2& "O" 'pure)
      (loop&
	 (until& (now& "R")
	    (seq&
	       (par&
		  (await& (now& "A"))
		  (await& (now& "B")))
	       (emit& "O")
             (halt&))))))

(make-test "abro"
   :invec '#(()
	     ("A" "B")
	     ("A")
	     ("R")
	     ("A")
	     ("B")
	     ("R")
	     ("A" "B" "R"))
   :outvec '#(()
	      ("O")
	      ()
	      ()
	      ()
	      ("O")
	      ()
	      ("O"))
   :program
   ABRO)

(make-test "atom"
   :program (hiphop& (atom& (display "ok")))
   :expected-output-string "ok")

(make-test "atom-input" :program
   (hiphop& :input (event2& "A" 'pure) :output (event2& "O" 'pure)
      (par&
	 (atom& (hiphop-input! (current-machine) "A"))
	 (seq&
	    (await& (now& "A"))
	    (emit& "O"))))
   :outvec '#(() ("O")))

; Catched, raise an error
(make-test "atom3" :program
   (hiphop& (atom& (hiphop-react! the-machine)))
   :interface-error "Illegal machine state")

; Detected
(make-test "causality #1" :program
   (hiphop& input: (event2& "I" 'pure) output: (event2& "O" 'pure)
      (if& (now& "I") (emit& "O") (emit& "I")))
   :causality-error "Cannot react. Unresolved program")

; Detected
(make-test "causality #2" ;:invec '#(("A"))
   :program (hiphop& :input (event2& "A" 'pure) (event2& "B" 'pure)
	       (seq&
		  (emit& "B")
		  (par&
		     (if& (now& "A") (nothing&) (emit& "B"))
		     (if& (now& "B") (emit& "A")))))
   :causality-error "Cannot react. Unresolved program")

(make-test "causality #3"
   :program (hiphop& :output (event2& "A" 'pure)
	       (if& (now& "A")
		  (emit& "A")))
   :outvec '#(())
   :causality-error "Cannot react. Unresolved program")

; Error: unresolved test; par fredo
(make-test "causality4" :program
   (hiphop& :output (event2& "F" 'pure)
      (local& ((event2& "G" 'pure) (event2& "E" 'pure))
	 (emit& "G")
	 (until& :immediate #t (now& "G")
	    (seq&
	       (await& :immediate #t (now& "E"))
	       (emit& "F")))
	 (emit& "E")))
   :causality-error "Cannot react. Unresolved program"
   :reactions 3)

(make-test "run" :program
   (hiphop& :output (event2& "O" 'pure)
      (run&
	 (hiphop& :output (event2& "O" 'pure)
	    (emit& "O"))))
   :outvec '#(("O")))

(make-test "rename:in" :program
   (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
      (run& :bind "I1" "I"
	 (hiphop& :input (event2& "I1" 'pure) :output (event2& "O" 'pure)
	    (seq&
	       (await& (now& "I1"))
	       (emit& "O")))))
   :invec '#(() ("I")) :outvec '#(() ("O")))

(make-test "rename:out" :program
   (hiphop& :output (event2& "O" 'pure)
      (run& :bind "O1" "O"
	 (hiphop& :output (event2& "O1" 'pure)
	    (emit& "O1"))))
   :outvec '#(("O")))

(make-test "rename:both" :program
   (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
      (run&
	 (hiphop& :input (event2& "I1" 'pure) :output (event2& "O1" 'pure)
	    (seq&
	       (await& (now& "I1"))
	       (emit& "O1")))
	 :bind "I1" "I" :bind "O1" "O"))
   :invec '#(() ("I"))
   :outvec '#(() ("O")))

(make-test "run-external-input" :program
   (hiphop& :output (event2& "O" 'pure) :input (event2& "I" 'pure)
      (run&
	 (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
	    (seq&
	       (atom& (hiphop-input! (current-machine) "I"))
	       (await& (now& "I"))
	       (emit& "O")))))
   :outvec '#(() ("O")))

(make-test "run-external-input+renaming" :program
   (hiphop& :output (event2& "O" 'pure)
      :input (event2& "I" 'pure)
      (run& :bind "O1" "O" :bind "I1" "I"
	 (let ((external-name "I1"))
	    (hiphop& :input (event2& "I1" 'pure) :output (event2& "O1" 'pure)
	       :initialize (lambda (m deep-name)
			      (set! external-name (deep-name "I1")))
	       (seq&
		  (atom& (hiphop-input! (current-machine) external-name))
		  (await& (now& "I1"))
		  (emit& "O1"))))))
   :outvec '#(() ("O")))

(make-tests "until-R-R"
   :program
   (hiphop& :input (event2& "R" 'pure) (event2& "I" 'pure) :output (event2& "O" 'pure)
      (loop&
	 (until& (now& "R")
	    (await& (now& "I"))
	    (emit& "O")
	    (halt&))))
   :events (vector
	      (list :invec '#(()) :outvec '#(()))
	      (list :invec '#(() ("I")) :outvec '#(() ("O")))
	      (list :invec '#(() ("R")) :outvec '#(() ()))
	      (list :invec '#(() ("I" "R")) :outvec '#(() ("O")))
	      (list :invec '#(() ("I") ()) :outvec '#(() ("O") ()))
	      (list :invec '#(() ("I") ("I")) :outvec '#(() ("O") ()))
	      (list :invec '#(() ("I") ("R")) :outvec '#(() ("O") ()))
	      (list :invec '#(() ("R") ()) :outvec '#(() () ()))
	      (list :invec '#(() ("R") ("I")) :outvec '#(() () ("O")))
	      (list :invec '#(() ("R") ("R")) :outvec '#(() () ()))
	      (list :invec '#(() ("I" "R") () ("R") ("I")) :outvec '#(() ("O") () () ("O")))
	      ))

(make-test "emit-val-emit" :program
   (hiphop& :output (event2& "O1" 'combined :combine + :default 0) (event2& "O2" 'single :default 0)
      (seq&
	 (emit& "O1" 1)
	 (emit& "O2" (val& "O1"))
	 (emit& "O1" 2)))
   :outvec '#((("O1" 3) ("O2" 1)))
   :causality-error "Cannot react. Unresolved program")

(make-test "atom-val&" :program
   (hiphop&
      :input (event2& "I" 'single :default 0)
      :output (event2& "O" 'single :default 0)
      (let ((x 0))
	 (seq&
	    (atom& (set! x (val& "I")))
	    (emit& "O" x))))
   :invec '#((("I" 2)))
   :outvec '#((("O" 2))))

(define ABCRO
   (hiphop& :input (event2& "A" 'pure) (event2& "B" 'pure) (event2& "C" 'pure) (event2& "R" 'pure) :output (event2& "O" 'pure)
      (loop&
 	 (until& (now& "R")
 	    (local& ((event2& "O1" 'pure))
 	       (par&
 		  (run& ABRO :bind "O" "O1")
 		  (run& ABRO
 		     :bind "A" "C"
 		     :bind "B" "O1")))))))

(make-test "abcro" :program ABCRO
   :invec '#(()
	      ("A" "B" "C")
	      ("A" "C")
	      ("R")
	      ("A")
	      ("C")
	      ("B")
	      ("A" "B" "C" "R"))
   :outvec '#(()
	       ("O")
	       ()
	       ()
	       ()
	       ()
	       ("O")
	       ())
   :reactions 8)

(make-test "abcdro" :program
   (hiphop&
      :input (event2& "A" 'pure) (event2& "B" 'pure) (event2& "C" 'pure) (event2& "D" 'pure) (event2& "R" 'pure)
      :output (event2& "O" 'pure)
      (loop&
	 (until& (now& "R")
	    (local& ((event2& "AC" 'pure) (event2& "BD" 'pure))
	       (par&
		  (run& ABRO :bind "B" "C"  :bind "O" "AC")
		  (run& ABRO :bind "A" "D"  :bind "O" "BD")
		  (run& ABRO :bind "A" "AC" :bind "B" "BD"))))))
   :invec '#(()
	      ("A" "B" "C" "D")
	      ("A" "C" "B")
	      ("R")
	      ("D")
	      ("C")
	      ("A")
	      ("B")
	      ("A" "B" "C" "R"))
   :outvec '#(()
	       ("O")
	       ()
	       ()
	       ()
	       ()
	       ()
	       ("O")
	       ()))

(make-test "abcdro2" :program
   (hiphop&
      :input (event2& "A" 'pure) (event2& "B" 'pure) (event2& "C" 'pure) (event2& "D" 'pure) (event2& "R" 'pure)
      :output (event2& "O" 'pure)
      (loop&
	 (until& (now& "R")
	    (local& ((event2& "O1" 'pure))
	       (par&
		  (run& ABCRO :bind "O" "O1")
		  (run& ABRO  :bind "A" "D" :bind "B" "O1"))))))
   :invec '#(()
	     ("A" "B" "C" "D")
	     ("A" "C" "B")
	     ("R")
	     ("D")
	     ("C")
	     ("A")
	     ("B")
	     ("A" "B" "C" "R"))
   :outvec '#(()
	      ("O")
	      ()
	      ()
	      ()
	      ()
	      ()
	      ("O")
	      ()))

(make-tests "every"
   :program (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
	       (every& (now& "I")
		  (emit& "O")))
   :events (vector
	      ; 1 reaction
	      (list :invec '#(())    :outvec '#(()))
	      (list :invec '#(("I")) :outvec '#(()))
	      ; 2 reactions
	      (list :invec '#(()    ())    :outvec '#(() ()))
	      (list :invec '#(()    ("I")) :outvec '#(() ("O")))
	      (list :invec '#(("I") ())    :outvec '#(() ()))
	      (list :invec '#(("I") ("I")) :outvec '#(() ("O")))))

(make-tests "every-immediate"
   :program (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
	       (every& :immediate #t (now& "I")
		  (emit& "O")))
   :events (vector
	      ; 1 reaction
	      (list :invec '#(())    :outvec '#(()))
	      (list :invec '#(("I")) :outvec '#(("O")))
	      ; 2 reactions
	      (list :invec '#(()    ())    :outvec '#(()    ()))
	      (list :invec '#(()    ("I")) :outvec '#(()    ("O")))
	      (list :invec '#(("I") ())    :outvec '#(("O") ()))
	      (list :invec '#(("I") ("I")) :outvec '#(("O") ("O")))))

(make-test "let/val" :program
   (hiphop& :input (event2& "I" 'single :default 0) :output (event2& "O" 'multi)
      (seq&
	 (emit& "O" (val& "I"))
	 (emit& "O" (val& "I"))))
   :invec '#((("I" 5)))
   :outvec '#((("O" 5 5))))

(make-test "emit/val" :program
   (hiphop& :input (event2& "I" 'single :default 0) :output (event2& "O" 'multi)
      (seq&
	 (await& (now& "I"))
	 (emit& "O" (val& "I"))))
   :invec '#(() (("I" 3)))
   :outvec '#(() (("O" 3))))

(make-test "2emit-1val" :program
   (hiphop& :output (event2& "O" 'single :default 0)
      (local& ((event2& "S" 'single :default 0))
	 (seq&
	    (par&
	       (emit& "O" (val& "S"))
	       (emit& "S" 1))
	    (pause&)
	    (par&
	       (emit& "O" (val& "S"))
	       (emit& "S" 2)))))
   :outvec '#((("O" 1)) (("O" 2))))

(make-test "loop-emit/val" :program
   (hiphop&
      :input (event2& "I" 'single :default 0)
      :output (event2& "O" 'single :default 0)
      (loop&
	 (await& (now& "I"))
	 (emit& "O" (val& "I"))))
      :invec '#(() (("I" 1)) (("I" 2)) (("I" 3)) (("I" 4)) (("I" 5)))
      :outvec '#(() (("O" 1)) (("O" 2)) (("O" 3)) (("O" 4)) (("O" 5))))

(let ((events (map (lambda (i) (format "O~a" i)) (iota 9 1))))
   (make-test "par-list"
      :program (hiphop&
		  :output (map (lambda (event) (event2& event 'pure)) events)
		  (par& (nothing&)
		     (map (lambda (ev) (emit& ev)) events)))
      :outvec `#(,events)))

(let ((events (map (lambda (i) (format "O~a" i)) (iota 2 1))))
   (make-test "par-list-emit"
      :program (hiphop& :output (event2& "O1" 'multi) (event2& "O2" 'multi)
		  (par& (nothing&)
		     (map (lambda (ev) (emit& ev 1)) events)))
      :outvec `#(,(map (lambda (e) (list e 1)) events))))

(make-tests "P1"
   :program (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
	       (local& ((event2& "S1" 'pure) (event2& "S2" 'pure))
		  (par&
		     (if& (now& "I") (emit& "S1"))
		     (if& (now& "S1") (nothing&) (emit& "S2"))
		     (if& (now& "S2") (emit& "O")))))
   :events (vector
	      (list :invec '#(()) :outvec '#(("O")))
	      (list :invec '#(("I")) :outvec '#(()))
	      (list :invec '#(("I") ()) :outvec '#(() ()))
	      (list :invec '#(() ("I")) :outvec '#(("O") ()))))

(make-test "parallel-emit-multiple"
   :program (hiphop& :output (event2& "O" 'single :default 0) (event2& "L" 'multi)
	       (par&
		  (loop&
		     (emit& "L" 1)
		     (emit& "O" (apply + (val& "L")))
		     (pause&))
		  (loop&
		     (emit& "L" 2)
		     (pause& :id 'bogus))))
   :outvec '#((("O" 3) ("L" 1 2))))

(make-test "P18" :program
   (hiphop&
      :output (event2& "S1_and_S2" 'pure)
      :output (event2& "S1_and_not_S2" 'pure)
      :output (event2& "not_S1_and_S2" 'pure)
      :output (event2& "not_S1_and_not_S2" 'pure)
      (loop&
	 (trap& T1
	    (local& ("S1")
	       (par&
		  (seq&
		     (pause&)
		     (emit& "S1")
		     (exit& T1))
		  (loop&
		     (trap& T2
			(local& ("S2")
			   (par&
			      (seq&
				 (pause&)
				 (emit& "S2")
				 (exit& T2))
			      (loop&
				 (if& (now& "S1")
				    (if& (now& "S2")
				       (emit& "S1_and_S2")
				       (emit& "S1_and_not_S2"))
				    (if& (now& "S2")
				       (emit& "not_S1_and_S2")
				       (emit& "not_S1_and_not_S2")))
				 (pause&)))))))))))
   :outvec '#(("not_S1_and_not_S2")
	      ("S1_and_S2" "S1_and_not_S2" "not_S1_and_not_S2")))

(make-test "local-scope" :program
   (hiphop& :output (event2& "O" 'multi)
      (seq&
	 (emit& "O" 1)
	 (local& ((event2& "O" 'multi))
	    (emit& "O" 2))))
   :outvec '#((("O" 1)) ()))

(make-test "run-rename-val1" :program
   (hiphop&
      :output (event2& "O" 'single :default 0)
              (event2& "nil" 'single :default 0)
	      (event2& "not_nil" 'single :default 0)
      (par&
	 (if& (zero? (val& "O"))
	    (emit& "nil" 0)
	    (emit& "not_nil" (val& "O")))
	 (local& ((event2& "O1" 'single :default 0))
	    (par&
	       (run&
		  (hiphop& :output (event2& "O1" 'single :default 0)
		     (emit& "O1" 1)))
	       (if& (val& "O1")
		  (emit& "O" (val& "O1"))
		  (if& (now& "O1") (emit& "O")))))))
   :outvec '#((("O" 1) ("not_nil" 1))))

(make-test "run-multiple" :program
   (let ((part (lambda (x)
		  (let ((i "I"))
		     (hiphop&
			:input (event2& "I" 'single :default 0)
			:output (event2& "O" 'single :default 0)
			:initialize (lambda (m n) (set! i (n "I")))
			(par&
			   (sustain& "O" x)
			   (atom& (hiphop-input! (current-machine) i (* x 3)))))))))
      (hiphop&
	 :output (list
		    (event2& "O" 'multi)
		    (event2& "O1" 'single :default 0)
		    (event2& "O2" 'single :default 0)
		    (event2& "O3" 'single :default 0))
	 :input (list
		   (event2& "I1" 'single :default 0)
		   (event2& "I2" 'single :default 0)
		   (event2& "I3" 'single :default 0))
	 (par&
	    (run& :bind "I" "I1" :bind "O" "O1" (part 1))
	    (run& :bind "I" "I2" :bind "O" "O2" (part 2))
	    (run& :bind "I" "I3" :bind "O" "O3" (part 3))
	    (seq&
	       (pause&)
	       (emit& "O" (val& "I1"))
	       (emit& "O" (val& "I2"))
	       (emit& "O" (val& "I3"))))))
   :outvec '#((("O1" 1) ("O2" 2) ("O3" 3))
	      (("O" 3 6 9) ("O1" 1) ("O2" 2) ("O3" 3))
	      (("O1" 1) ("O2" 2) ("O3" 3))))

(make-test "trap T in exit T; pause end trap; emit O" :program
   (hiphop& :output (event2& "O" 'single :default 0)
      (local& ((event2& "I" 'single :default 0))
	 (par&
	    (emit& "O" (val& "I"))
	    (seq&
	       (trap& T
		  (exit& T)
		  (pause&))
	       (emit& "I" 1)))))
   :outvec '#((("O" 1))))

(make-tests "atom present" :program
   (hiphop& :input (event2& "I" 'pure)
      (atom& (if (now& "I") (display "OK") (display "NOK"))))
   :events (vector
	      (list :expected-output-string "NOK")
	      (list :invec '#(("I")) :expected-output-string "OK")))

(make-tests "emit present" :program
   (hiphop&
      :input (event2& "I" 'combined :combine (lambda (a b) (format "~a~a" a b)) :default "")
      :output (event2& "O" 'single :default 0)
      (emit& "O" (begin (if (now& "I") (display (val& "I")) (display "0")) 1)))
   :events (vector
	      (list :outvec '#((("O" 1))) :expected-output-string "0")
	      (list :outvec '#((("O" 1))) :invec '#((("I" #f))) :expected-output-string "#f")
	      (list :outvec '#((("O" 1))) :invec '#((("I" "a"))) :expected-output-string "a")))

(make-tests "event-combine+" :program
   (hiphop&
      :input (event2& "I" 'single :default 0)
      :output (event2& "O" 'combined :combine + :default 0)
      (seq&
	 (emit& "O" (val& "I"))
	 (emit& "O" 5)
	 (pause&)
	 (emit& "O" (preval& "O"))
	 (emit& "O" (preval& "I"))))
   :events (vector
	      (list :invec '#((("I" 2)) ()) :outvec '#((("O" 7)) (("O" 9))))))

(make-test "event-combine-local" :program
   (hiphop&
      :input (event2& "I" 'single :default 0)
      :output (event2& "O" 'single :default 0)
      (local& ((event2& "L" 'combined :combine + :default 0))
	 (emit& "L" (val& "I"))
	 (emit& "L" (val& "I"))
	 (emit& "O" (val& "L"))))
   :invec '#((("I" 2)))
   :outvec '#((("O" 4))))

(make-test "event-sync-atom" :program
   (hiphop& :output (event2& "O" 'single :default 0)
      (local& ((event2& "L" 'single :default 0))
	 (par&
	    (emit& "O" (val& "L"))
	    (atom& (print (val& "L")))
	    (emit& "L" 1))))
   :expected-output-string "1\n"
   :outvec '#((("O" 1))))

(make-test "event-combine-bind" :program
   (hiphop& :output (event2& "O" 'combined :combine + :default 0)
      (seq&
	 (emit& "O" 1)
	 (run& :bind "O1" "O"
	    (hiphop& :output (event2& "O1" 'combined :combine + :default 0)
	       (emit& "O" 1)))
	 (emit& "O" 1)))
   :outvec '#((("O" 3))))

(make-test "event-combine-bind-badfunc" :program
   (hiphop& :output (event2& "O" 'combined :combine + :default 0)
      (seq&
	 (emit& "O" 1)
	 (run& :bind "O1" "O"
	    (hiphop& :output (event2& "O1" 'combined :combine * :default 0)
	       (emit& "O" 1)))
	 (emit& "O" 1)))
   :compile-error "Illegal binding. Event declarations do not match")

(make-test "event-combine-bind-baddef" :program
   (hiphop& :output (event2& "O" 'combined :combine + :default 0)
      (seq&
	 (emit& "O" 1)
	 (run& :bind "O1" "O"
	    (hiphop& :output (event2& "O1" 'combined :combine + :default 1)
	       (emit& "O" 1)))
	 (emit& "O" 1)))
   :compile-error "Illegal binding. Event declarations do not match")

(make-test "event-hide-bind" :program
   (hiphop& :output (event2& "O" 'combined :combine + :default 0)
      (seq&
	 (emit& "O" 1)
	 (run& :bind "O" ""
	    (hiphop& :output (event2& "O" 'combined :combine + :default 1)
	       (emit& "O" 1)))
	 (emit& "O" 1)))
   :outvec '#((("O" 2))))

(make-test "suspend-sustain" :program
   (hiphop& :input (event2& "I1" 'pure) (event2& "I2" 'pure) :output (event2& "O1" 'pure) (event2& "O2" 'pure)
      (par&
	 (suspend& (or (now& "I1") (now& "I2"))
	    (sustain& "O1"))
	 (loop&
	    (if& (now& "O1")
	       (emit& "O2"))
	    (pause&))))
   :invec '#(() ("I1") ("I2") ("I1" "I2"))
   :outvec '#(("O1" "O2") () () ()))

(let ((sustainer&
	 (lambda (i)
	    (hiphop&
	       :output (event2& "O" 'multi)
	       (sustain& "O" i)))))

   (make-test "dynamic-simple" :program
      (hiphop& :output (event2& "O" 'multi)
	 (seq&
	    (atom& (hiphop-append! (current-machine) (sustainer& 1)))
	    (pause&)
	    (atom& (hiphop-append! (current-machine) (sustainer& 2)))
	    (pause&)
	    (atom& (hiphop-append! (current-machine) (sustainer& 3)))
	    (pause&)))
      :dynamic? #t
      :outvec (vector
		 '()
		 '(("O" 1))
		 '(("O" 2 1))
		 '(("O" 3 2 1)))))

(make-test "genpar-empty" :program
   (hiphop& :input (event2& "I" 'multi) :output (event2& "O" 'multi)
      (seq&
	 (await& (now& "I"))
	 (genpar& ((i (val& "I")))
	    (emit& "O" i))))
   :invec '#(() ("I")))

(make-test "genpar-static" :program
   (hiphop& :output (event2& "O" 'combined :combine + :default 0)
      (genpar& ((i '(1 2 3)))
	 (emit& "O" i)))
   :outvec '#((("O" 6))))

(make-test "genpar-static/pause" :program
   (hiphop& :output (event2& "O" 'combined :combine + :default 0)
      (genpar& ((i '(1 2 3)))
	 (pause&)
	 (emit& "O" i)))
   :outvec '#(() (("O" 6))))

(make-test "genpar-static-2" :program
   (hiphop& :output (event2& "O" 'combined :combine + :default 0)
      (genpar& ((i '(1 2 3)) (j '(8 9)))
	 (emit& "O" (* i j))))
   :outvec '#((("O" 102))))

(make-test "genpar-input" :program
   (hiphop& :input (event2& "I" 'multi) :output (event2& "O" 'combined :combine + :default 0)
      (seq&
	 (await& (now& "I"))
	 (genpar& ((i (val& "I")))
	    (if& (= i 2)
	       (emit& "O" i)))))
   :invec '#(() (("I" 1 2)))
   :outvec '#(() (("O" 2))))

(make-tests "genpar-exit" :program
   (hiphop& :input (event2& "I" 'multi) (event2& "R" 'multi) :output (event2& "O" 'combined :combine + :default 0)
      (loop&
	 (abort& (now& "R")
	    (seq&
	       (await& (now& "I"))
	       (genpar& ((i (val& "I")))
		  (if& (zerofx? (modulo i 2))
		     (emit& "O" i)))
	       (pause&)))))
   :events `#(
	      (:invec #(()))
	      (:invec #(() (("I" 1))))
	      (:invec #(() (("I" 4))) :outvec #(() (("O" 4))))
	      (:invec #(() (("I" 4 6)) (("R" 0) ("I" 2))) :outvec #(() (("O" 10)) ()))
	      (:invec #(() (("I" 4)) (("R" 0)) (("I" 2))) :outvec #(() (("O" 4)) () (("O" 2))))))

(make-test "suspend-2" :program
   (hiphop& :input (event2& "I" 'pure) :output (event2& "O" 'pure)
      (suspend& :immediate #t (now& "I")
	 (emit& "O")))
   :invec '#(("I") ("I") ("I")))

(make-test "dyngenpar-simple" :program
   (hiphop& :input (event2& "I" 'multi) :output (event2& "O" 'multi)
      (dyngenpar& ((i (val& "I"))) :when (now& "I")
	 (if& (oddfx? i)
	    (emit& "O" i))))
   :invec '#(() (("I" 0 1 2 3)) (("I" 1)) () (("I" 3 5)))
   :outvec '#(() (("O" 1 3)) (("O" 1)) () (("O" 3 5))))

(make-test "dyngenpar-exit" :program
   (hiphop& :input (event2& "I" 'multi) :output (event2& "O" 'multi)
      (loop&
	 (trap& T
	    (dyngenpar& ((i (val& "I"))) :when (now& "I")
	       (if& (=fx i 5)
		  (exit& T)
		  (if& (oddfx? i)
		     (emit& "O" i)))))
	 (pause&)))
   :invec '#(() (("I" 0 1 2 3)) (("I" 1)) () (("I" 3 5)) (("I" 1)) (("I" 2 3)))
   :outvec '#(() (("O" 1 3)) (("O" 1)) () (("O" 3)) () (("O" 3))))

(define (run-tests #!key ignore-failures? port)
   (let ((p (or port (current-output-port)))
	 (failed 0)
	 (ran 0))
      (bind-exit (exit)
	 (map (lambda (test)
		 (let ((test-name (car test))
		       (test-thunk (cdr test)))
		    (display (format "=== Running test #~a: ~a: " (+ ran 1) test-name) p)
		    (flush-output-port p)
		    (with-handler
		       (lambda (e)
			  (fprint p "FAILED")
			  (fprint p "*** Restarting test '" test-name "' in debug mode")
			  (let ((old-debug (bigloo-debug)))
			     (bigloo-debug-set! 50)
			     (let ((res (test-thunk 'debug)))
				(bigloo-debug-set! old-debug)
				(set! failed (+fx failed 1))
				(fprint p "--- end of debug trace of " test-name)
				(unless ignore-failures? (exit res)))))
		       (let ((res (test-thunk)))
			  (set! ran (+ ran 1))
			  (if res
			      (fprint p "OK")
			      (error test-name "Test-failure" test-thunk))))))
	    *tests*))
      (if (zero? failed)
	  (fprintf p "\nAll ~a tests succeeded\n" ran)
	  (fprintf p "\n*** ~a of ~a tests failed\n" failed ran))
      (>fx failed 0)))

(define (main args)
   (let ((keep-going? #f))
      (when (and (pair? args) (pair? (cdr args)))
	 (when (or (string=? (cadr args) "-C") (string=? (cadr args) "-kC"))
	    (bigloo-trace-color-set! #f))
	 (when (or (string=? (cadr args) "-k") (string=? (cadr args) "-kC"))
	    (set! keep-going? #t)))
      (run-tests :ignore-failures? (not keep-going?))))
