(module __hiphop_analysis

   (import __hiphop_types
           __hiphop_debug
	   __hiphop_utils)

   (export (mark! ::hiphop-program)

	   (generic gamma-tardieu::bint statement::hiphop-ast%)
	   (generic omega-tardieu::bint statement::hiphop-ast%)))

(define (mark! prgm::hiphop-program)
   (ast-mark! prgm.reactive-code))

(define-generic (ast-mark! node::hiphop-ast%))

(define-method (ast-mark! nothing::hiphop-nothing)
   (with-trace *hiphop-debug-mark!* 'mark!/nop
      (bint->bv 0)))
(define-method (ast-mark! emit::hiphop-emit)
   (with-trace *hiphop-debug-mark!* 'mark!/emit
      (trace-item "Emit: " emit.id ", event: " emit.event)
      (set! emit.reachable? #t)
      (set! emit.event.event%.reachable-count (+fx emit.event.event%.reachable-count 1))
      (bint->bv 0)))
(define-method (ast-mark! pause::hiphop-pause)
   (with-trace *hiphop-debug-mark!* 'mark!/pause
      (trace-item "pause: " pause.id)
      (if pause.selected?
	  (bint->bv 0)
	  (bint->bv 1))))
(define-method (ast-mark! pause::hiphop-goto-pause)
   (with-trace *hiphop-debug-mark!* 'mark!/goto-pause
      (trace-item "pause: " pause.id)
      (bint->bv 1)))
(define-method (ast-mark! if%::hiphop-if)
   (with-trace *hiphop-debug-mark!* 'mark!/if
      (case if%.branch
	 ((then) (ast-mark! if%.then))
	 ((else) (ast-mark! if%.else))
	 ((none) (bv-union (ast-mark! if%.then) (ast-mark! if%.else)))
	 (else (raise (instantiate::&hiphop-internal-error
			 (proc "ast-mark!/if")
			 (msg "Illegal selected branch")
			 (obj if%)))))))
(define-method (ast-mark! suspend::hiphop-suspend)
   (with-trace *hiphop-debug-mark!* 'mark!/suspend
      (let ((body-mark (ast-mark! suspend.body)))
      (bv-union
	 (bint->bv 1) ; might pause
	 body-mark))))
(define-method (ast-mark! seq::hiphop-sequence)
   (with-trace *hiphop-debug-mark!* 'mark!/seq
      (let ((len (vector-length seq.branches))
	    (current seq.current-branch))
	 (when (=fx current -1) (set! current 0))
	 (let loop ((res (ast-mark! (vector-ref seq.branches current)))
		    (next (+fx current 1)))
	    (if (<fx next len)
		(if (bv-has-zero? res)
		    (loop (bv-union/!zero res (ast-mark! (vector-ref seq.branches next))) (+fx next 1))
		    res)
		res)))))
(define-method (ast-mark! loop::hiphop-loop)
   (with-trace *hiphop-debug-mark!* 'mark!/loop
      (let ((res (ast-mark! loop.body)))
	 (if (bv-has-zero? res)
	     (raise (instantiate::&hiphop-causality-error
		       (proc "loop&")
		       (msg "Potentially unsafe loop")
		       (obj loop)
		       (location loop.source-location)))
	     res))))
(define-method (ast-mark! par::hiphop-parallel)
   (with-trace *hiphop-debug-mark!* 'mark!/par
      (trace-fields par id active-branches)
      (let ((len (vector-length par.branches))
	    (res #f))
	 (let loop ((current 0))
	    (when (<fx current len)
	       (when (vector-ref par.active-branches current)
		  (let ((subres (ast-mark! (vector-ref par.branches current))))
		     (set! res (if res (bv-max res subres) subres))))
	       (loop (+fx current 1))))
	 (cond
	    ((not res) 0)
	    ((zerofx? res)
	     (raise (instantiate::&hiphop-internal-error
		       (proc "ast-mark!")
		       (msg "Parallel has no active branches")
		       (obj par))))
	    (else
	     res)))))
(define-method (ast-mark! genpar::hiphop-genpar)
   (with-trace *hiphop-debug-mark!* 'mark!/genpar
      (if genpar.expanded?
	  (call-next-method)
	  (begin
	     (ast-mark! genpar.can-nil-instance)))))
(define-method (ast-mark! dgp::hiphop-dyngenpar)
   (with-trace *hiphop-debug-mark!* 'mark!/dyngenpar
      ;; Reachable emitters. They exist both in any branches or in the
      ;; can nil instance (reached by inheritance)
      (if dgp.selected-branches?
	  (bv-union
	     (ast-mark! dgp.genpar)  ; it cannot be in the expanded state
	     (call-next-method))
	  (ast-mark! dgp.genpar))))
(define-method (ast-mark! trap::hiphop-trap)
   (with-trace *hiphop-debug-mark!* 'mark!/trap
      (bv-downshift (ast-mark! trap.body))))
(define-method (ast-mark! exit::hiphop-exit)
   (with-trace *hiphop-debug-mark!* 'mark!/exit
      (bint->bv exit.index)))
(define-method (ast-mark! local::hiphop-local)
   (with-trace *hiphop-debug-mark!* 'mark!/local
      (ast-mark! local.body)))
(define-method (ast-mark! run::hiphop-run)
   (with-trace *hiphop-debug-mark!* 'mark!/run
      (ast-mark! run.program.reactive-code)))
(define-method (ast-mark! atom::hiphop-atom)
   (with-trace *hiphop-debug-mark!* 'mark!/atom
      (bint->bv 0)))


(define-generic (gamma-tardieu::bint statement::hiphop-ast%))

(define-method (gamma-tardieu s::hiphop-nothing)
   (bint->bv 0))
(define-method (gamma-tardieu s::hiphop-emit)
   (bint->bv 0))
(define-method (gamma-tardieu s::hiphop-pause)
   (bint->bv 1))
(define-method (gamma-tardieu s::hiphop-goto-pause)
   (bint->bv 1))
(define-method (gamma-tardieu s::hiphop-if)
   (bv-union (gamma-tardieu s.then) (gamma-tardieu s.else)))
(define-method (gamma-tardieu s::hiphop-suspend)
   (gamma-tardieu s.body))
(define-method (gamma-tardieu seq::hiphop-sequence)
   (let ((len (vector-length seq.branches))
	 (gamma-p (gamma-tardieu (vector-ref seq.branches 0))))
      (let loop ((i 1))
	 (when (< i len)
	    (when (bv-has-zero? gamma-p)
	       (let ((gamma-q (gamma-tardieu (vector-ref seq.branches i))))
		  (set! gamma-p (bv-union/!zero gamma-p gamma-q)))
	       (loop (+fx i 1)))))
      gamma-p))
(define-method (gamma-tardieu loop::hiphop-loop)
   (bv-union/!zero (gamma-tardieu loop.body) 0))
(define-method (gamma-tardieu par::hiphop-parallel)
   (let ((len (vector-length par.branches))
	 (gamma-p (gamma-tardieu (vector-ref par.branches 0))))
      (let loop ((i 1))
	 (when (< i len)
	    (let ((gamma-q (gamma-tardieu (vector-ref par.branches i))))
	       (set! gamma-p (bv-max gamma-p gamma-q)))
	    (loop (+fx i 1))))
      gamma-p))
(define-method (gamma-tardieu trap::hiphop-trap)
   (bv-downshift (gamma-tardieu trap.body)))
(define-method (gamma-tardieu exit::hiphop-exit)
   (bint->bv exit.index))
(define-method (gamma-tardieu local::hiphop-local)
   (gamma-tardieu local.body))
(define-method (gamma-tardieu run::hiphop-run)
   (gamma-tardieu run.program.reactive-code))
(define-method (gamma-tardieu atom::hiphop-atom)
   (bint->bv 0))

(define-generic (omega-tardieu::bint statement::hiphop-ast%)
   (gamma-tardieu statement))
(define-method (omega-tardieu pause::hiphop-pause)
   (bv-union
      (bint->bv 0)
      (bint->bv 1)))
(define-method (omega-tardieu pause::hiphop-goto-pause)
   (bint->bv 1))
