(module __hiphop_ast

   (import  __hiphop_types
            __hiphop_utils)

   (export  ; Programs
            (make-hiphop&::hiphop-program loc code::hiphop-ast% name init inputs outputs)

	    ; AST / instructions
	    (make-nothing&::hiphop-nothing loc)
	    (make-emit&::hiphop-emit loc event::bstring valued?::bool value)
	    (make-pause&::hiphop-pause loc #!key id)
	    (make-if&::hiphop-if loc expr::hiphop-host-expr then::hiphop-ast% else::hiphop-ast%)
	    (make-suspend&::hiphop-suspend loc id test::hiphop-host-expr body::hiphop-ast%)
	    (make-sequence&::hiphop-sequence loc id children)
	    (make-loop&::hiphop-loop loc body::hiphop-ast% #!key id)
	    (make-parallel&::hiphop-parallel #!key location id children)
	    (make-trap&::hiphop-trap loc name body::hiphop-ast%)
	    (make-exit&::hiphop-exit loc trap-name)
	    (make-local&::hiphop-local loc user-event body::hiphop-ast%)
	    (make-run&::hiphop-run loc name bindings program::hiphop-program)
	    (make-atom&::hiphop-atom loc expr::hiphop-host-expr)
	    (make-genpar&::hiphop-genpar #!key id location host-identifiers host-expressions generator)
	    (make-dyngenpar&::hiphop-dyngenpar #!key id location when-expression genpar)
	    (make-host-expression&::hiphop-host-expr loc expr::procedure
	       #!key (now-depends '()) (val-depends '()) (pre-depends '()))

	    ; User events
	    (event&::hiphop-user-event ::bstring #!key combine default (multi? #t))
	    (declare-event&::hiphop-user-event2 #!key loc name type (default 'hiphop-nil-default) combine)

	    (generic hiphop-copy::hiphop-program  ::hiphop-program)
	    (generic hiphop-copy-ast::hiphop-ast% ::hiphop-ast%)
	    (generic hiphop-copy-surface::hiphop-ast% ::hiphop-ast%)))

(define (any->hiphop-user-event proc obj)
   (cond
      ((isa? obj hiphop-user-event) obj)
      ((isa? obj hiphop-user-event2) obj)
      ((string? obj) (declare-event& :name obj :type 'multi))
      (else
       (raise (hiphop-type-error proc "event& or string" obj)))))

(define (make-hiphop& loc code name init inputs outputs)
   (let ((flatten-inputs '())
	 (flatten-outputs '()))
      (let loop-in ((t inputs))
	 (when (pair? t)
	    (let ((hd (car t)))
	       (cond
		  ((pair? hd)
		   (loop-in (append hd (cdr t))))
		  ((string? hd)
		   (hiphop-warn loc "hiphop&" "Deprecated use of a plain string used for event declaration -- " hd)
		   (let ((user (any->hiphop-user-event "hiphop&" hd)))
		      (set! flatten-inputs (cons hd flatten-inputs)))
		   (loop-in (cdr t)))
		  ((isa? hd hiphop-user-event)
		   (set! flatten-inputs (cons hd flatten-inputs))
		   (loop-in (cdr t)))
		  ((isa? hd hiphop-user-event2)
		   (set! flatten-inputs (cons hd flatten-inputs))
		   (loop-in (cdr t)))
		  ((isa? hd hiphop-ast%)
		   (raise (hiphop-compile-error
			     (let ((ast::hiphop-ast% hd)) ast.source-location)
			     "hiphop&"
			     "Only one statement is allowed within hiphop& form"
			     (list hd code))))
		  (else
		   (raise (hiphop-type-error "make-hiphop&" "event& or string" hd loc code)))))))
      (let loop-out ((t outputs))
	 (when (pair? t)
	    (let ((hd (car t)))
	       (cond
		  ((pair? hd)
		   (loop-out (append hd (cdr t))))
		  ((string? hd)
		   (hiphop-warn loc "hiphop&" "Deprecated use of a plain string used for event declaration -- " hd)
		   (let ((user (any->hiphop-user-event "hiphop&" hd)))
		      (set! flatten-outputs (cons hd flatten-outputs)))
		   (loop-out (cdr t)))
		  ((isa? hd hiphop-user-event)
		   (set! flatten-outputs (cons hd flatten-outputs))
		   (loop-out (cdr t)))
		  ((isa? hd hiphop-user-event2)
		   (set! flatten-outputs (cons hd flatten-outputs))
		   (loop-out (cdr t)))
		  ((isa? hd hiphop-ast%)
		   (raise (hiphop-compile-error
			     (let ((ast::hiphop-ast% hd)) ast.source-location)
			     "hiphop&"
			     "Only one statement is allowed within hiphop& form"
			     (list hd code))))
		  (else
		   (raise (hiphop-type-error "make-hiphop&" "event& or string" hd loc code)))))))
   (instantiate::hiphop-program
      (source-location loc)
      (reactive-code code)
      (name (or name (symbol->string (gensym 'hiphop))))
      (initializer init)
      (formal-inputs  flatten-inputs)
      (formal-outputs flatten-outputs))))

(define (make-nothing& loc)
   (instantiate::hiphop-nothing
      (source-location loc)))

(define (make-emit& loc event valued? expr)
   (instantiate::hiphop-emit
      (source-location loc)
      (name event)
      (valued? valued?)
      (expr expr)))

(define (make-pause& loc #!key id)
   (instantiate::hiphop-pause
      (source-location loc)
      (id (or id (gensym 'pause)))))

(define (make-if& loc test then else)
   (instantiate::hiphop-if
      (source-location loc)
      (test test)
      (then then)
      (else else)))

(define (make-suspend& loc id test body)
   (instantiate::hiphop-suspend
      (source-location loc)
      (id (or id (gensym 'suspend&-)))
      (event-test test)
      (body body)))

(define (make-sequence& loc id children)
   (let ((branches '()))
      (let loop ((t children))
	 (when (pair? t)
	    (let ((hd (car t)))
	       (cond
		  ((pair? hd)
		   (loop (append hd (cdr t))))
		  ((isa? hd hiphop-ast%)
		   (set! branches (cons hd branches))
		   (loop (cdr t)))
		  (else
		   (raise (hiphop-type-error "seq&"
			     "pair or hiphop-ast%" hd loc children)))))))
      (instantiate::hiphop-sequence
	 (source-location loc)
	 (branches (list->vector (reverse branches)))
	 (id (or id (gensym 'seq))))))

(define (make-loop& loc body #!key id)
   (instantiate::hiphop-loop
      (source-location loc)
      (body body)
      (id (or id (gensym 'loop)))))

(define (make-parallel& #!key location id children)
   (let ((branches '()))
      (let loop ((t children))
	 (when (pair? t)
	    (let ((hd (car t)))
	       (cond
		  ((pair? hd)
		   (loop (append hd (cdr t))))
		  ((isa? hd hiphop-ast%)
		   (set! branches (cons hd branches))
		   (loop (cdr t)))
		  (else
		   (raise (hiphop-type-error "par&" "pair or hiphop-ast%" hd location children)))))))
      (instantiate::hiphop-parallel
	 (source-location location)
	 (id (or id (gensym 'par)))
	 (branches (list->vector (reverse branches)))
	 (active-branches (make-vector (length branches) #t)))))

(define (make-trap& loc name body)
   (instantiate::hiphop-trap
      (source-location loc)
      (name name)
      (body body)))

(define (make-exit& loc trap-name)
   (instantiate::hiphop-exit
      (source-location loc)
      (trap-name trap-name)))

(define (make-local& loc event body)
   (instantiate::hiphop-local
      (source-location loc)
      (id (gensym 'loc))
      (user-event (any->hiphop-user-event "local&" event))
      (body body)))

(define (make-run& loc name bindings program)
   (instantiate::hiphop-run
      (source-location loc)
      (name name)
      (bindings bindings)
      (program program)))

(define (make-atom& loc expr)
   (instantiate::hiphop-atom
      (source-location loc)
      (expr expr)))

(define (make-genpar& #!key id location host-identifiers host-expressions generator)
   (unless (or (pair? host-identifiers) (pair? host-expressions) (procedure? generator))
      (raise (instantiate::&hiphop-compile-error
		(proc "genpar&")
		(msg "Bad syntax, missing parameters")
		(obj id)
		(location location))))
   (instantiate::hiphop-genpar
      ;; inherited from ::hiphop-ast%
      (source-location location)
      (id (or id (gensym 'genpar)))
      ;; inherited from ::hiphop-parallel
      (branches (vector))
      (active-branches (vector))
      ;; own fields
      (host-identifiers host-identifiers)
      (host-expressions host-expressions)
      (generator generator)))

(define (make-dyngenpar& #!key id location when-expression genpar)
   (instantiate::hiphop-dyngenpar
      ;; inherited from ::hiphop-ast%
      (source-location location)
      (id (or id (gensym 'dyngenpar)))
      ;; inherited from ::hiphop-parallel
      (branches (vector))
      (active-branches (vector))
      ;; (nothing inherited from ::hiphop-parallel-dynamic)
      ;; own fields
      (when when-expression)
      (genpar genpar)))

(define (make-host-expression& loc expr #!key (now-depends '()) (val-depends '()) (pre-depends '()))
   (instantiate::hiphop-host-expr
      (source-location loc)
      (expression expr)
      (%now-depends now-depends)
      (%val-depends val-depends)
      (%pre-depends pre-depends)))

(define (hiphop-copy-host::hiphop-host-expr orig::hiphop-host-expr)
   (instantiate::hiphop-host-expr
      (source-location orig.source-location)
      (expression orig.expression) ; procedures are not duplicatable
      (%now-depends orig.%now-depends)
      (%val-depends orig.%val-depends)
      (%pre-depends orig.%pre-depends)))

(define (event& name::bstring #!key combine default (multi? #t))
   (when combine
      (unless (procedure? combine)
	 (raise (hiphop-type-error "event&" "procedure" combine))))
   (instantiate::hiphop-user-event
      (name name)
      (combine combine)
      (default default)
      (multi? (and (not combine) multi?))))

(define (declare-event& #!key loc name type (default 'hiphop-nil-default) combine)
   (unless (string? name)
      (raise (hiphop-type-error "event&" "string" name loc)))
   (unless (symbol? type)
      (raise (hiphop-type-error "event&" "symbol" type loc)))
   (case type
      ((pure)
       (when (or combine (not (eq? default 'hiphop-nil-default)))
	  (hiphop-warn loc "event&: " "unused :combine and/or :default argument for 'pure' event " name "; ignored"))
       (instantiate::hiphop-user-event-pure
	  (source-location loc)
	  (name name)))
      ((single)
       (when combine
	  (hiphop-warn loc "event&: " "unused :combine argument for 'single' event " name "; ignored"))
       (when (eq? default 'hiphop-nil-default)
	  (hiphop-warn loc "event&: " "missing :default argument for 'single' event " name "; assuming #f"))
       (instantiate::hiphop-user-event-single
	  (source-location loc)
	  (name name)
	  (default default)))
      ((combined)
       (if combine
	   (unless (procedure? combine)
	      (raise (hiphop-type-error "event&" "procedure" combine loc)))
	   (raise (hiphop-compile-error loc "event&" "Missing :combine procedure argument for 'combined' event" name)))
       (when (eq? default 'hiphop-nil-default)
	  (let ((def (with-handler
			(lambda (e)
			   (raise
			      (hiphop-compile-error loc "event&" "Missing :default value argument for 'combined' event" name)))
			;; At that point, we checked that 'combine' is a procedure. We known nothing about its arity
			(combine))))
	     (hiphop-warn loc "event&: " "missing :default value argument for 'combined' event " name (format "; assuming ~a" def))
	     (set! default def)))
       (instantiate::hiphop-user-event-combined
	  (source-location loc)
	  (name name)
	  (combine combine)
	  (default default)))
      ((multi)
       (when (or combine (not (eq? default 'hiphop-nil-default)))
	  (hiphop-warn loc "event&: " "unused :combine and/or :default argument for 'multi' event " name "; ignored"))
       (instantiate::hiphop-user-event-multi
	  (source-location loc)
	  (name name)))
      (else
       (raise (hiphop-compile-error loc "event&" "Illegal event type. Must be one of (pure single combined multi)" type)))))

;;; Copy
;; Deep copy of a program
(define-generic (hiphop-copy::hiphop-program p::hiphop-program)
   (make-hiphop&
      p.source-location
      (hiphop-copy-ast p.reactive-code)
      p.name
      p.initializer
      p.formal-inputs
      p.formal-outputs))

;; Deep copy of a node of the ast
(define-generic (hiphop-copy-ast::hiphop-ast% s::hiphop-ast%))

(define-method (hiphop-copy-ast s::hiphop-nothing)
   (with-access::hiphop-nothing s (source-location)
      (make-nothing&
	 source-location)))

(define-method (hiphop-copy-ast s::hiphop-emit)
   (with-access::hiphop-emit s (source-location name valued? expr)
      (make-emit&
	 source-location
	 name
	 valued?
	 (when valued? (hiphop-copy-host expr)))))

(define-method (hiphop-copy-ast pause::hiphop-pause)
   (make-pause& pause.source-location :id pause.id))

(define-method (hiphop-copy-ast goto-pause::hiphop-goto-pause)
   (instantiate::hiphop-goto-pause
      (source-location goto-pause.source-location)
      (id goto-pause.id)))

; if
(define-method (hiphop-copy-ast s::hiphop-if)
   (with-access::hiphop-if s (source-location test then else)
      (make-if&
	 source-location
	 (hiphop-copy-host test)
	 (hiphop-copy-ast then)
	 (hiphop-copy-ast else))))

(define-method (hiphop-copy-ast s::hiphop-suspend)
   (make-suspend&
      s.source-location
      s.id
      (hiphop-copy-host s.event-test)
      (hiphop-copy-ast s.body)))

(define-method (hiphop-copy-ast s::hiphop-sequence)
   (make-sequence&
      s.source-location
      s.id
      (map hiphop-copy-ast (vector->list s.branches))))

(define-method (hiphop-copy-ast s::hiphop-loop)
   (with-access::hiphop-loop s (source-location body)
      (make-loop&
	 source-location
	 (hiphop-copy-ast body))))

(define-method (hiphop-copy-ast s::hiphop-parallel)
   (make-parallel&
      :location s.source-location
      :id (string->symbol (format "copy-of-~a" s.id))
      :children (map hiphop-copy-ast (vector->list s.branches))))

;; Overriden as genpar must not copy its branches
(define-method (hiphop-copy-ast s::hiphop-genpar)
   (make-genpar&
      :location s.source-location
      :id s.id
      :host-identifiers s.host-identifiers
      :host-expressions (map hiphop-copy-host s.host-expressions)
      :generator s.generator))

(define-method (hiphop-copy-ast s::hiphop-dyngenpar)
   (make-dyngenpar&
      :location s.source-location
      :id s.id
      :when-expression (hiphop-copy-host s.when)
      :genpar (hiphop-copy-ast s.genpar)))

(define-method (hiphop-copy-ast trap::hiphop-trap)
   (make-trap&
      trap.source-location
      trap.name
      (hiphop-copy-ast trap.body)))

(define-method (hiphop-copy-ast exit::hiphop-exit)
   (make-exit&
      exit.source-location
      exit.trap-name))

(define-method (hiphop-copy-ast s::hiphop-local)
   (with-access::hiphop-local s (source-location user-event body)
      (make-local&
	 source-location
	 user-event
	 (hiphop-copy-ast body))))

(define-method (hiphop-copy-ast run::hiphop-run)
   (make-run&
      run.source-location
      run.name
      run.bindings
      (hiphop-copy run.program)))

(define-method (hiphop-copy-ast s::hiphop-atom)
   (with-access::hiphop-atom s (source-location expr)
      (make-atom& source-location (hiphop-copy-host expr))))

;;; Copy surface
;; Prototype: hiphop-ast% -> (values ast-copy has-pause?)
; where has-pause? is a boolean meaning that the copied ast contains gotopause
; A #t value for has-pause? means that we reach the end of the surface
; The intended use is for sequence, to not got further if a copied branch pauses
(define-generic (hiphop-copy-surface ast::hiphop-ast%))

(define-method (hiphop-copy-surface nothing::hiphop-nothing)
   (values (hiphop-copy-ast nothing) #f))

(define-method (hiphop-copy-surface emit::hiphop-emit)
   (values (hiphop-copy-ast emit) #f))

;; The surface copy of a pause is a goto-pause
(define-method (hiphop-copy-surface pause::hiphop-pause)
   (values
      (instantiate::hiphop-goto-pause
	 (source-location pause.source-location)
	 (id pause.id))
      #t))

(define-method (hiphop-copy-surface goto-pause::hiphop-goto-pause)
   (values (hiphop-copy-ast goto-pause) #t))

(define-method (hiphop-copy-surface if%::hiphop-if)
   (multiple-value-bind (then-copy then-has-pause?)
      (hiphop-copy-surface if%.then)
      (multiple-value-bind (else-copy else-has-pause?)
	 (hiphop-copy-surface if%.else)
	 (values
	    (make-if&
	       if%.source-location
	       (hiphop-copy-host if%.test)
	       then-copy
	       else-copy)
	    (and then-has-pause? else-has-pause?)))))

(define-method (hiphop-copy-surface s::hiphop-suspend)
   (multiple-value-bind (body-copy body-has-pause?)
      (hiphop-copy-surface s.body)
      (values
	 (make-suspend&
	    s.source-location
	    s.id
	    (hiphop-copy-host s.event-test)
	    body-copy)
	 body-has-pause?)))

(define-method (hiphop-copy-surface seq::hiphop-sequence)
   (let ((length (vector-length seq.branches))
	 (branch-copies '())
	 (paused? #f))
      (let loop ((i 0))
	 (if (and (<fx i length) #t);(not paused?))
	     (multiple-value-bind (copy has-pause?)
		(hiphop-copy-surface (vector-ref seq.branches i))
		(set! branch-copies (cons copy branch-copies))
		(set! paused? has-pause?)
		(loop (+fx i 1)))
	     (values
		(make-sequence&
		   seq.source-location
		   seq.id
		   (reverse branch-copies))
		paused?)))))

(define-method (hiphop-copy-surface loop::hiphop-loop)
   (multiple-value-bind (body-copy body-has-pause?)
      (hiphop-copy-surface loop.body)
      (values
	 body-copy
	 body-has-pause?)))

;; The surface ends if all branches pauses
(define-method (hiphop-copy-surface par::hiphop-parallel)
   (let ((length (vector-length par.branches))
	 (branch-copies '())
	 (branch-pauses '()))
      (let loop ((i 0))
	 (if (<fx i length)
	     (multiple-value-bind (branch-copy branch-has-pause?)
		(hiphop-copy-surface (vector-ref par.branches i))
		(set! branch-copies (cons branch-copy branch-copies))
		(set! branch-pauses (cons branch-has-pause? branch-pauses))
		(loop (+fx i 1)))
	     (values
		(make-parallel&
		   :location par.source-location
		   :id (string->symbol (format "surf-of-~a" par.id))
		   :children (reverse branch-copies))
		(every? (lambda (b) b) branch-pauses))))))

;; There is no way to ensure that the surface will pause
(define-method (hiphop-copy-surface genpar::hiphop-genpar)
   (values (hiphop-copy-ast genpar) #f))

;; FIXME: Seems useless, as this would be done by the parent method
(define-method (hiphop-copy-surface dyngenpar::hiphop-dyngenpar)
   (values (hiphop-copy-ast dyngenpar) #f))

(define-method (hiphop-copy-surface trap::hiphop-trap)
   (multiple-value-bind (body-copy body-has-pause?)
      (hiphop-copy-surface trap.body)
      (values
	 (make-trap&
	    trap.source-location
	    trap.name
	    body-copy)
	 body-has-pause?)))

(define-method (hiphop-copy-surface exit::hiphop-exit)
   (values (hiphop-copy-ast exit) #f))

(define-method (hiphop-copy-surface local::hiphop-local)
   (multiple-value-bind (body-copy body-has-pause?)
      (hiphop-copy-surface local.body)
      (values
	 (make-local&
	    local.source-location
	    local.user-event
	    body-copy)
	 body-has-pause?)))

;; Note that we directly copy the surface of the submodule
; not the program itself
(define-method (hiphop-copy-surface run::hiphop-run)
   (multiple-value-bind (program-copy program-has-pause?)
      (hiphop-copy-surface run.program.reactive-code)
      (values
   	 (make-run&
   	    run.source-location
	    run.name
	    run.bindings
   	    (make-hiphop&
   	       run.program.source-location
   	       program-copy
	       run.program.name
	       run.program.initializer
   	       run.program.formal-inputs
   	       run.program.formal-outputs))
   	 program-has-pause?)))

(define-method (hiphop-copy-surface atom::hiphop-atom)
   (values (hiphop-copy-ast atom) #f))
