;; DOM-like API functions for HipHop AST.
(module __hiphop_dom

   (import __hiphop_types
           __hiphop_debug)

   (export (generic hiphop-get-element-by-id ::hiphop-ast% ::symbol)
           (generic hiphop-append-child! ::hiphop-ast% ::hiphop-ast%)))

;; Deep-first search
(define-generic (hiphop-get-element-by-id leaf-node::hiphop-ast% id)
   (with-trace *hiphop-debug-dom* 'getElementById/leaf
      (trace-item "Leaf.id: " leaf-node.id)
      (when (eq? leaf-node.id id)
	 leaf-node)))
(define-method (hiphop-get-element-by-id node::hiphop-if id)
   (with-trace *hiphop-debug-dom* 'getElementById/if
      (or (call-next-method)
	  (hiphop-get-element-by-id node.then id)
	  (hiphop-get-element-by-id node.else id))))
(define-method (hiphop-get-element-by-id node::hiphop-suspend id)
   (with-trace *hiphop-debug-dom* 'getElementById/suspend
      (or (call-next-method)
	  (hiphop-get-element-by-id node.body id))))
(define-method (hiphop-get-element-by-id node::hiphop-sequence id)
   (with-trace *hiphop-debug-dom* 'getElementById/seq
      (or (call-next-method)
	  (let ((len (vector-length node.branches)))
	     (let loop ((i 0))
		(when (<fx i len)
		   (or (hiphop-get-element-by-id (vector-ref node.branches i) id)
		       (loop (+fx i 1)))))))))
(define-method (hiphop-get-element-by-id node::hiphop-loop id)
   (with-trace *hiphop-debug-dom* 'getElementById/loop
      (or (call-next-method)
	  (hiphop-get-element-by-id node.body id))))
(define-method (hiphop-get-element-by-id node::hiphop-parallel id)
   (with-trace *hiphop-debug-dom* 'getElementById/par
      (or (call-next-method)
	  (let ((len (vector-length node.branches)))
	     (let loop ((i 0))
		(when (<fx i len)
		   (or (hiphop-get-element-by-id (vector-ref node.branches i) id)
		       (loop (+fx i 1)))))))))
(define-method (hiphop-get-element-by-id node::hiphop-trap id)
   (with-trace *hiphop-debug-dom* 'getElementById/trap
      (or (call-next-method)
	  (hiphop-get-element-by-id node.body id))))
(define-method (hiphop-get-element-by-id node::hiphop-local id)
   (with-trace *hiphop-debug-dom* 'getElementById/local
      (or (call-next-method)
	  (hiphop-get-element-by-id node.body id))))
(define-method (hiphop-get-element-by-id node::hiphop-run id)
   (with-trace *hiphop-debug-dom* 'getElementById/run
      (or (call-next-method)
	  (hiphop-get-element-by-id node.program.reactive-code id))))

;; append-child! is only authorized on hiphop-parallel-dynamic nodes
(define-generic (hiphop-append-child! parent::hiphop-ast% child::hiphop-ast%)
   (raise (instantiate::&hiphop-internal-error
	     (proc "hiphop-append-child!")
	     (msg  "This method is not supported for this element")
	     (obj  parent))))
(define-method (hiphop-append-child! parent::hiphop-parallel-dynamic child::hiphop-ast%)
   (let ((new-active-branches (vector-append parent.active-branches '#(#f)))
	 (new-branches (vector-append parent.branches (vector child))))
      (set! parent.active-branches new-active-branches)
      (set! parent.branches new-branches)))
