(module __hiphop_events
   ; events management
   (import __hiphop_types
           __hiphop_debug)

   (export ;;; Event life
      ;; Creation
      (make-event&::hiphop-event ::bstring ::bint)
      (make-uevent&::hiphop-event ::hiphop-user-event ::bint)
      (generic make-uevent2&::hiphop-event ::hiphop-user-event2 ::bint)
      ;; Comparison (but name)
      (generic uevent-equal?::bool ::hiphop-user-event2 ::obj)

      ;; Emission with no value
      (event-emit! ::hiphop-event%)
      ;; Emission with a value
      (event-emit!/value ::hiphop-event% ::obj)
      ;; Multiple emission with values
      (event-emit!/values ::hiphop-event% ::pair)
      ;; Presence access
      (event-present?::bool ::hiphop-event%)
      ;; Value access when emitted
      (generic event-value::obj ::hiphop-event%)
      ;; Value access when emitted with no value
      ;; Value access when not emitted
      ;; Previous presence access
      (event-previously-present?::bool ::hiphop-event%)
      ;; Previous value access
      (generic event-previous-value::obj ::hiphop-event%)

      ;; Status update
      (event-update-status! ::hiphop-event%)

      ;; Resetting
      (event-reset! ::hiphop-event%)

      ;; Name resolution
      (get-event-by-name::hiphop-event ::hiphop-machine% ::bstring ::bstring))

   (static ;;; Private functions prototypes
      ;; Move value into pre-value when needed
      (generic event-reset-values! ::hiphop-event%)
      ;; Add a single value to an event
      (generic event-add-value! ::hiphop-event% ::obj)
      ;; Add several values to an event
      (event-add-values! ::hiphop-event% ::pair)))


(define (make-event& name::bstring index::bint)
   (make-uevent&
      (instantiate::hiphop-user-event (name name) (combine #f) (default #f))
      index))

(define (make-uevent& user::hiphop-user-event index::bint)
   (instantiate::hiphop-event
      (name user.name)
      (user user)
      (event%
	 (cond
	    (user.multi?
	       (instantiate::hiphop-multi-event
		  (index index)))
	    (user.combine
	       (instantiate::hiphop-combined-event
		  (index index)
		  (combine user.combine)
		  (default user.default)
		  (value user.default)
		  (pre-value user.default)))
	    (user.default
	       (instantiate::hiphop-valued-event
		  (index index)
		  (default user.default)
		  (value user.default)
		  (pre-value user.default)))
	    (else
	     (instantiate::hiphop-pure-event
		(index index)))))))

(define-generic (make-uevent2& user::hiphop-user-event2 index::bint))
;; Pure events
(define-method (make-uevent2&  user::hiphop-user-event-pure index)
   (instantiate::hiphop-event
      (name user.name)
      (user user)
      (event%
	 (instantiate::hiphop-pure-event
	    (index index)))))
;; Single-valued events
(define-method (make-uevent2&  user::hiphop-user-event-single index)
   (instantiate::hiphop-event
      (name user.name)
      (user user)
      (event%
	 (instantiate::hiphop-valued-event
	    (index     index)
	    (default   user.default)
	    (value     user.default)
	    (pre-value user.default)))))
;; Combined events
(define-method (make-uevent2&  user::hiphop-user-event-combined index)
   (instantiate::hiphop-event
      (name user.name)
      (user user)
      (event%
	 (instantiate::hiphop-combined-event
	    (index     index)
	    (default   user.default)
	    (value     user.default)
	    (pre-value user.default)
	    (combine   user.combine)))))
;; Multi-valued events
(define-method (make-uevent2&  user::hiphop-user-event-multi index)
   (instantiate::hiphop-event
      (name user.name)
      (user user)
      (event%
	 (instantiate::hiphop-multi-event
	    (index index)))))

(define-generic (uevent-equal? user::hiphop-user-event2 o::obj))
(define-method  (uevent-equal? user::hiphop-user-event-pure o)
   (isa? o hiphop-user-event-pure))
(define-method  (uevent-equal? user::hiphop-user-event-single o)
   (when (isa? o hiphop-user-event-single)
      (let ((ouser::hiphop-user-event-single o))
	 (equal? user.default ouser.default))))
(define-method  (uevent-equal? user::hiphop-user-event-combined o)
   (when (isa? o hiphop-user-event-combined)
      (let ((ouser::hiphop-user-event-combined o))
	 (and (equal? user.default ouser.default)
	      (equal? user.combine ouser.combine)))))
(define-method  (uevent-equal? user::hiphop-user-event-multi o)
   (isa? o hiphop-user-event-multi))


(define (event-emit! event%::hiphop-event%)
   (with-trace *hiphop-debug-events* 'event-emit!
      (trace-item "Emitting event #" event%.index)
      ; Set the status to present
      (when (eq? event%.status 'absent)
	 (raise (instantiate::&hiphop-internal-error
		   (proc "event-emit!")
		   (msg "Event is absent")
		   (obj event%))))
      (set! event%.status 'present)))

(define (event-emit!/value event%::hiphop-event% value::obj)
   (with-trace *hiphop-debug-events* 'event-emit!/value
      (trace-item "event%: "  event%)
      (trace-item "value:  " value)
      (%event-emit!/val event% value event-add-value!)))

(define (event-emit!/values event%::hiphop-event% values::pair)
   (with-trace *hiphop-debug-events* 'event-emit!/values
      (trace-item "event%: "  event%)
      (trace-item "values: " values)
      (%event-emit!/val event% values event-add-values!)))

(define (%event-emit!/val event%::hiphop-event% val::obj adder!::procedure)
   ; Mark as present
   (unless (eq? event%.status 'present)
      (event-emit! event%))
   ; Move current values to pre-value if needed
   (event-reset-values! event%)
   ; Add the value
   (adder! event% val))


(define (event-present? event%::hiphop-event%)
   (case event%.status
      ((present) #t)
      ((absent ) #f)
      (else
       (raise (instantiate::&hiphop-internal-error
		 (proc "event-present?")
		 (msg "Unknown event status")
		 (obj event%))))))

(define-generic (event-value event%::hiphop-event%)
   #f)

(define-method (event-value event%::hiphop-valued-event)
   event%.value)

(define-method (event-value event%::hiphop-multi-event)
   event%.value)

(define (event-previously-present? event%::hiphop-event%)
   (case event%.pre-status
      ((present) #t)
      ((absent ) #f)
      (else
       (raise (instantiate::&hiphop-internal-error
		 (proc "event-previously-present?")
		 (msg "Illegal value for pre-status field")
		 (obj event%))))))

(define-generic (event-previous-value event%::hiphop-event%))
(define-method (event-previous-value event%::hiphop-valued-event)
   (when event%.valued?
      (if event%.emitted/value?
	  event%.pre-value
	  event%.value)))
(define-method (event-previous-value event%::hiphop-multi-event)
   (when event%.valued?
      (if event%.emitted/value?
	  event%.pre-value
	  event%.value)))

(define (event-update-status! event%::hiphop-event%)
   (with-trace *hiphop-debug-events* 'event-update-status!
      (when (eq? event%.status 'unknown)
	 (when (>fx event%.emitted-count 0)
	    (trace-item "Marking event as present: " event%)
	    (set! event%.status 'present))
	 (when (zerofx? event%.reachable-count)
	    (trace-item "Marking event as absent: " event%)
	    (set! event%.status 'absent)))))

;; We don't set the status here, it will be set by
;; environment-prepare! according to machine inputs
(define (event-reset! event%::hiphop-event%)
   ;; Reset emitters and emissions
   (set! event%.reachable-emitters '())
   (set! event%.emitted '())
   ;; Reset emissions counters
   (set! event%.reachable-count 0)
   (set! event%.emitted-count 0)
   ;; Reset the fact the event has been emitted with a value
   (set! event%.emitted/value? #f)
   ;; Save the status into pre-status
   (unless (eq? event%.status 'unknown)
      (set! event%.pre-status event%.status))
   ;; Set new status to unknown
   (set! event%.status 'unknown))

;;; Private

;; Move value into pre-value when needed
(define-generic (event-reset-values! event%::hiphop-event%))

(define-method (event-reset-values! event%::hiphop-valued-event)
   ; Warning, multiple calls of this function may loose data
   (unless event%.emitted/value?
      (set! event%.pre-value (event-value event%))
      (set! event%.value event%.default)))

(define-method (event-reset-values! event%::hiphop-multi-event)
   ; Warning, multiple calls of this function may loose data
   (unless event%.emitted/value?
      (set! event%.pre-value (event-value event%))
      (set! event%.value '())))

;; Add a single value to an event
(define-generic (event-add-value! event%::hiphop-event% value::obj))

(define-method (event-add-value! event%::hiphop-valued-event value::obj)
   (when event%.emitted/value?
      (raise (instantiate::&hiphop-causality-error
		(proc "emit&")
		(msg "event is already emitted")
		(obj event%))))
   (set! event%.emitted/value? #t)
   (set! event%.valued? #t)
   (set! event%.value value))

(define-method (event-add-value! event%::hiphop-combined-event value::obj)
   (with-trace *hiphop-debug-events* 'event-add-value!/combined
      (trace-item "Event #" event%.index ", adding value: " value "; " event%)
      (unless event%.emitted/value?
	 (set! event%.emitted/value? #t))
      (unless event%.valued? (set! event%.valued? #t))
      (set! event%.value (event%.combine value event%.value))
      (trace-item "New value: " event%.value)))

; Multi-event cons their values
(define-method (event-add-value! event%::hiphop-multi-event value::obj)
   (with-trace *hiphop-debug-events* 'event-add-value!/multi
      (trace-item "Event #" event%.index ", adding value: " value "; " event%)
      (if event%.emitted/value?
	  (set! event%.mvalued? #t)
	  (set! event%.emitted/value? #t))
      (unless event%.valued? (set! event%.valued? #t))
      (set! event%.value (cons value event%.value))
      (trace-item "New value: " event%.value)))


;; Add several values to an event
(define (event-add-values! event%::hiphop-event% values::pair)
   (for-each (lambda (val) (event-add-value! event% val)) values))

(define (convert-error e::&hiphop-error)
   (cond-expand
      (scheme2js
	 (let ((err (js-new (@ Error js) e.msg)))
	    (set! err.hopLocation (symbol->string e.constructor.sc_name))
	    (set! err.name e.proc)
	    (set! err.scObject e.obj)
	    (when (pair? e.location)
	       (let ((file-char (cdr e.location)))
		  (set! err.fileName (car file-char))
		  (set! err.charNumber (cadr file-char))))
	    (set! err.cause e)
	    err))
      (else
       e)))

;; Event resolution from the interface, used by now, val, pre and preval
(define (get-event-by-name::hiphop-event machine::hiphop-machine% event-name::bstring proc-name::bstring)
   (with-trace *hiphop-debug-events* proc-name
      (trace-item "event-name: " event-name)
      (with-handler
	 (lambda (e)
	    (if (and (isa? e &hiphop-error) (not machine.raw-errors?))
		(raise (convert-error e))
		(raise e)))
	 (let ((event-index (hashtable-get machine.environment.%events event-name)))
	    (if (and (number? event-index) (hashtable-contains? machine.environment.%gevents event-index))
		(hashtable-get machine.environment.%gevents event-index)
		(begin
		   (trace-item "%events: " machine.environment.%events)
		   (trace-item "%gevents: " machine.environment.%gevents)
		   (raise (instantiate::&hiphop-internal-error
			     (proc proc-name)
			     (msg "Unknown event")
			     (obj event-name)))))))))
