;*---------------------------------------------------------------------*/
;*    hiphop-hiphop&-expander ...                                      */
;*---------------------------------------------------------------------*/
(define (hiphop-hiphop&-expander x e)
   (match-case x
      ((?- . ?args)
       ; First, we expand the arguments
       (let ((expanded-args (e args e))
	     (no-keywords #f)
	     (name #f)
	     (initializer #f)
	     (program #f))

	  ; Only one program can be specified, it has to be the car of the last pair
	  (when (pair? expanded-args)
	     (let ((last (last-pair expanded-args)))
		(set! program last)))

	  ; We recurse in the list to replace the last-pair with a specific marker
	  (let find-and-replace-last-pair-penultimate ((l expanded-args))
	     (cond
		((eq? l program) ; no keywords
		 (set! no-keywords #t))
		((pair? l)
		 (if (eq? (cdr l) program) ; penultimate element
		     (set-cdr! l (list :end)) ; we add the specific marker here
		     (find-and-replace-last-pair-penultimate (cdr l))))
		(else
		 (print (typeof program) ": " program)
		 (print (typeof l) ": " l)
		 (error "hiphop&" "Illegal form/penultimate" x))))

	  ; We should have found a program at that point
	  (if (pair? program)
	      (set! program (car program))
	      (error "hiphop&" "No program found" x))

	  (if no-keywords

	      (e `(make-hiphop&
		     (quote ,(and (epair? x) (cer x)))
		     ,program
		     #f
		     #f
		     '()
		     '()) e)

	      ; At that point, expanded-args should only contains keywords and no-keyword element.
	      ; We expect :input and :output keywords here. We stop when we encounter a :end keyword
	      (let loop ((l expanded-args)
			 (in '())   ; place holder for input events
			 (out '())) ; place holder for output events
		 ; we could add some more

		 (if (pair? l) ; the first element must be a keyword, followed by anything
		     (let ((first (car l)))
			(if (keyword? first)
			    (cond
			       ((eq? first :name) ; Syntax check for the name argument
				(cond
				   (name
				      (error "hiphop&" "Name is already defined" name))
				   ((null? (cdr l))
				    (error "hiphop&" "Missing value for name: parameter" l))
				   ((keyword? (cadr l))
				    (error "hiphop&" "Illegal value for name: parameter" l))
				   (else
				    (set! name (cadr l))
				    (loop (cddr l) in out))))

			       ((eq? first :input) ; We greedly add anything that is not a keyword
				(let leep ((l2 (cdr l)) (acc '()))
				   (if (pair? l2)
				       (if (keyword? (car l2))
					   (loop l2  (append in (reverse acc)) out)
					   (leep (cdr l2) (cons (car l2) acc)))
				       (error "hiphop&" "Illegal form" l2))))

			       ((eq? first :output) ; We greedly add anything that is not a keyword
				(let leep ((l2 (cdr l)) (acc '()))
				   (if (pair? l2)
				       (if (keyword? (car l2))
					   (loop l2 in (append out (reverse acc)))
					   (leep (cdr l2) (cons (car l2) acc)))
				       (error "hiphop&" "Illegal form" l2))))

			       ((eq? first :initialize)
				(cond
				   (initializer
				    (error "hiphop&" "Only one initilizer allowed" initializer))
				   ((null? (cdr l))
				    (error "hiphop&" "Missing value for initialize: parameter" l))
				   ((keyword? (cadr l))
				    (error "hiphop&" "Illegal value for initialize: parameter" l))
				   (else
				    (set! initializer (cadr l))
				    (loop (cddr l) in out))))

			       ((eq? first :end)
				(e `(make-hiphop&
				       (quote ,(and (epair? x) (cer x)))
				       ,program
				       ,name
				       ,initializer
				       (list ,@in)
				       (list ,@out)) e))

			       (else
				(error "hiphop&" "Illegal keyword" first)))
			    (error "hiphop&" "Missing keyword" first)))
		     (error "hiphop&" "Illegal form" l))))))
      (else
       (error "hiphop&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-nothing&-expander ...                                     */
;*---------------------------------------------------------------------*/
(define (hiphop-nothing&-expander x e)
   (e `(make-nothing& (quote ,(and (epair? x) (cer x)))) e))

;*---------------------------------------------------------------------*/
;*    hiphop-emit&-expander ...                                        */
;*---------------------------------------------------------------------*/
(define (hiphop-emit&-expander x e)
   (match-case x
      ;; Two distinct cases, depending on the emission of pure or valued event
      ((?- ?event ?value)
       (e `(make-emit&
	      (quote ,(and (epair? x) (cer x))) ; location
	      ,event
	      #t ; Is there a value to be emitted along with the event?
	      ,(hiphop-expressions-expander value e))
	  e))
      ((?- ?event)
       (e `(make-emit&
	      (quote ,(and (epair? x) (cer x))) ; location
	      ,event
	      #f ; Is there a value to be emitted along with the event?
	      #f) e))
      (else
       (error "emit&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-pause&-expander ...                                       */
;*---------------------------------------------------------------------*/
(define (hiphop-pause&-expander x e)
   (match-case x
      ((?- :id ?symb)
       (e `(make-pause&
	      (quote ,(and (epair? x) (cer x)))
	      :id ,symb) e))
      ((?-)
       (e `(make-pause& (quote ,(and (epair? x) (cer x)))) e))
      (else
       (error "pause&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-if&-expander ...                                          */
;*---------------------------------------------------------------------*/
(define (hiphop-if&-expander x e)
   (match-case x
      ((?- ?test ?then ?else)
       (let ((loc (and (epair? x) (cer x))))
	  (when (string? test)
	     (if loc
		 (warning/location (cadr loc) (caddr loc)
		    "host-expression: " "testing for a string always returns true -- " test)
		 (warning "host-expression: " "testing for a string always returns true -- " test)))
	  (e `(make-if&
		 (quote ,loc) ; location
		 ,(hiphop-expressions-expander test e)
		 ,then
		 ,else)
	     e)))
      ((?- ?test ?then)
       (hiphop-if&-expander
	  (evepairify-deep
	     `(if& ,test ,then (nothing&))
	     x)
	  e))
      (else
       (error "if&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-suspend&-expander ...                                     */
;*---------------------------------------------------------------------*/
(define (hiphop-suspend&-expander x e)
   (match-case x
      ((?- :id ?id :immediate ?imm? ?test ?head . ?tail)
       (e (evepairify-deep
	     `(make-suspend&
		 (quote ,(and (epair? x) (cer x)))
		 ,id
		 ,(hiphop-expressions-expander test e)
		 (make-sequence&
		    (quote ,(and (epair? x) (cer x)))
		    (gensym 'suspend&/seq&-)
		    (list
		       (if ,imm?
			   (if& ,test
			      (pause&))
			   (nothing&))
		       ,head
		       ,@tail)))
	     x)
	  e))
      ((?- :id ?id (and (not (? keyword?)) ?event-test) ?head . ?tail)
       (hiphop-suspend&-expander
	  (evepairify-deep
	     `(suspend& :id ,id :immediate #f ,event-test ,head ,@tail)
	     x)
	  e))
      ((?- :immediate ?imm? (and (not (? keyword?)) ?event-test) ?head . ?tail)
       (hiphop-suspend&-expander
	  (evepairify-deep
	     `(suspend& :id #f :immediate ,imm? ,event-test ,head ,@tail)
	     x)
	  e))
      ((?- (and (not (? keyword?)) ?event-test) ?head . ?tail)
       (hiphop-suspend&-expander
	  (evepairify-deep
	     `(suspend& :id #f :immediate #f ,event-test ,head ,@tail)
	     x)
	  e))
      (else
       (error "suspend&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-sequence&-expander ...                                    */
;*---------------------------------------------------------------------*/
(define (hiphop-sequence&-expander x e)
   (match-case x
      ((?- :id ?id . ?rest)
       (e (evepairify-deep
	     `(make-sequence&
		 (quote ,(and (epair? x) (cer x)))
		 ,id
		 (list ,@rest))
	     x)
	  e))
      ((?- ?head . ?rest)
       (e (evepairify-deep
	     `(make-sequence&
		 (quote ,(and (epair? x) (cer x)))
		 #f
		 (list ,head ,@rest))
	     x)
	  e))
      (else
       (error "seq&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-loop&-expander ...                                        */
;*---------------------------------------------------------------------*/
(define (hiphop-loop&-expander x e)
   (match-case x
      ((?- ?body-head . ?body-tail)
       (e `(make-loop&
	      (quote ,(and (epair? x) (cer x)))
	      ,(evepairify `(seq& ,body-head ,@body-tail) x)) e))
      (else
       (error "loop&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-parallel&-expander ...                                    */
;*---------------------------------------------------------------------*/
(define (hiphop-parallel&-expander x e)
   (match-case x
      ((?- :id ?id ?head . ?tail)
       (e (evepairify-deep
	     `(make-parallel&
		 :location (quote ,(and (epair? x) (cer x)))
		 :id ,id
		 :children (list ,head ,@tail))
	     x)
	  e))
      ((?- ?head . ?tail)
       (e (evepairify-deep
	     `(make-parallel&
		 :location (quote ,(and (epair? x) (cer x)))
		 :id #f
		 :children (list ,head ,@tail))
	     x)
	  e))
      (else
       (error "par&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-trap&-expander ...                                        */
;*---------------------------------------------------------------------*/
(define (hiphop-trap&-expander x e)
   (match-case x
      ((?- ?name ?body)
       (e `(make-trap&
	      (quote ,(and (epair? x) (cer x)))
	      (quote ,name)
	      ,body) e))
      ((?- ?name ?body-first . ?body-rest)
       (hiphop-trap&-expander
	  (evepairify-deep `(trap& ,name (seq& ,body-first ,@body-rest)) x) e))
      (else
       (error "trap&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-exit&-expander ...                                        */
;*---------------------------------------------------------------------*/
(define (hiphop-exit&-expander x e)
   (match-case x
      ((?- ?trap-name)
       (e `(make-exit&
	      (quote ,(and (epair? x) (cer x)))
	      (quote ,trap-name)) e))
      (else
       (error "exit&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-local&-expander ...                                       */
;*---------------------------------------------------------------------*/
(define (hiphop-local&-expander x e)
   (match-case x
      ;; Get rid of local in case (local& () first . rest)
      ((?- () ?first . ?rest)
       (e (evepairify `(seq& ,first ,@rest) x) e))
      ;; Normal case
      ((?- (?head . ?tail) . ?expr)
       (e `(make-local&
	      (quote ,(and (epair? x) (cer x)))
	      ,head
	      ,(evepairify `(local& ,tail ,@expr) x)) e))
      (else
       (error "local&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-run&-expander ...                                         */
;*---------------------------------------------------------------------*/
(define (hiphop-run&-expander x e)
   (match-case x
      ((?- . ?args)
       (let ((expanded-args (e args e)) ; expands the arguments
	     (bindings      '())   ; Bindings, list of pairs
	     (rename        #f)    ; The renaming of the program, must be unique
	     (program       #f))   ; The actual program to run, must be unique
	  (let loop ((l expanded-args))
	     (match-case l
		; Bindings
		((:bind ?local-name ?global-name . ?rest)
		 (set! bindings (cons `(cons ,local-name ,global-name) bindings))
		 (loop rest))
		; Renaming
		((:rename ?renaming . ?rest)
		 (if rename
		     (error "run&" "Only one :rename form is allowed" l)
		     (set! rename renaming))
		 (loop rest))
		; Other keywords
		(((? keyword?) ???-)
		 (error "run&" "Illegal keyword" (car l)))
		((?- . ?rest) ;(or ((make-hiphop& ???-) . ?rest) ((? symbol?) . ?rest))
		 (if program
		     (error "run&" "Illegal form. Only one program can be specified" x)
		     (set! program (car l)))
		 (loop rest))
		(() ; No arguments left
		 #unspecified)
		(else
		 (error "run&" "Illegal value" (car l)))))
	  (unless program
	     (error "run&" "Illegal form. Missing hiphop& program" x))
	  (e `(make-run&
		 (quote ,(and (epair? x) (cer x)))
		 ,rename
		 (list ,@(reverse bindings))
		 ,program) e)))
      (else
       (error "run&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-atom&-expander ...                                        */
;*---------------------------------------------------------------------*/
(define (hiphop-atom&-expander x e)
   (match-case x
      ((?- ?sexpr)
       (e `(make-atom&
	      (quote ,(and (epair? x) (cer x))) ; location
	      ,(hiphop-expressions-expander sexpr e))
	  e))
      (else
       (error "atom&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-genpar&-expander ...                                      */
;*---------------------------------------------------------------------*/
(define (hiphop-genpar&-expander x e)
   (match-case x
      ((?- ?bindings ?body-head . ?body-tail)
       (when (null? bindings)
	  (error "genpar&" "Illegal form: no bindings" x))
       (unless (and (every? pair? bindings)
		    (every? pair? (map cdr bindings))
		    (every? null? (map cddr bindings)))
	  (error "genpar&" "Malformed bindings" x))
       (let ((idents (map car bindings))
	     (hexprs (map cadr bindings))
	     (hiphop (if (pair? body-tail)
			 (evepairify-deep
			    `(seq& ,body-head ,@body-tail) x)
			 body-head)))
	  (e `(make-genpar&
		 :location (quote ,(and (epair? x) (cer x)))
		 :host-identifiers (quote ,idents)
		 :host-expressions (list ,@(map (lambda (sexpr) (hiphop-expressions-expander sexpr e)) hexprs))
		 :generator (lambda ,idents ,hiphop))
	     e)))
      (else
       (error "genpar&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-dyngenpar&-expander ...                                   */
;*---------------------------------------------------------------------*/
(define (hiphop-dyngenpar&-expander x e)
   (match-case x
      ((?- ?bindings :when ?host-expr ?body-head . ?body-tail)
       (when (null? bindings)
	  (error "dyngenpar&" "Illegal form: no bindings" x))
       (unless (and (every? pair? bindings)
		    (every? pair? (map cdr bindings))
		    (every? null? (map cddr bindings)))
	  (error "dyngenpar&" "Malformed bindings" x))

       (let ((idents (map car bindings))
	     (hexprs (map cadr bindings))
	     (hiphop (if (pair? body-tail)
			 (evepairify-deep
			    `(seq& ,body-head ,@body-tail) x)
			 body-head)))
	  (e `(make-dyngenpar&
		 :location (quote ,(and (epair? x) (cer x)))
		 :when-expression ,(hiphop-expressions-expander host-expr e)
		 :genpar ,(evepairify `(genpar& ,bindings ,body-head ,@body-tail) x))
	     e)))
      (else
       (error "dyngenpar&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-sustain&-expander ...                                     */
;*---------------------------------------------------------------------*/
(define (hiphop-sustain&-expander x e)
   (match-case x
      ((?- ?event . ?value)
       (e (evepairify-deep `(loop& (emit& ,event ,@value) (pause&)) x) e))
      (else
       (error "sustain&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-halt&-expander ...                                        */
;*---------------------------------------------------------------------*/
(define (hiphop-halt&-expander x e)
   (e (evepairify-deep `(loop& (pause&)) x) e))

;*---------------------------------------------------------------------*/
;*    hiphop-await&-expander ...                                       */
;*---------------------------------------------------------------------*/
(define (hiphop-await&-expander x e)
   (match-case x
      ((?- :immediate ?immediate? ?test)
       (let ((break-loop (gensym (string->symbol "await&/break/"))))
	  (e (evepairify-deep
		`(trap& ,break-loop
		    (if ,immediate? ; immediate? will be resolved at ast's build-time
			(loop&
			   (if& ,test (exit& ,break-loop))
			   (pause&))
			(loop&
			   (pause&)
			   (if& ,test (exit& ,break-loop)))))
		x) e)))
      ((?- ?test)
       (hiphop-await&-expander (evepairify `(await& :immediate #f ,test) x) e))
      (else
       (error "await&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-until&-expander ...                                       */
;*---------------------------------------------------------------------*/
(define (hiphop-until&-expander x e)
   (match-case x
      ;; Canonic usage
      ; we try to find a handler& in either body-head or in the last pair of body-tail
      ((?- :immediate ?immediate? ?test ?body-head . ?body-tail)
       (let ((handlers '())
	     (end (gensym (string->symbol "until&/end")))
	     (handle (gensym (string->symbol "until&/handle"))))
	  (let ((body-list ; parses body to extract handler& forms
		   (filter-map
		      (lambda (x)
			 (match-case x
			    ((handler& ?body) (set! handlers (cons body handlers)) #f)
			    ((handler& ?- ???-)
			     (error "until&" "Illegal handler& form" x))
			    (else x)))
		      (cons body-head body-tail))))
	     (e
		(evepairify-deep
		   (if (pair? handlers)
		       `(trap& ,end
			   (seq&
			      (trap& ,handle
				 (par&
				    (seq&
				       ,@body-list
				       (exit& ,end))
				    (seq&
				       (await& :immediate ,immediate? ,test)
				       (exit& ,handle))))
			      (seq& ,@(reverse handlers))))
		       `(trap& ,end
			   (par&
			      (seq&
				 ,@body-list
				 (exit& ,end))
			      (seq&
				 (await& :immediate ,immediate? ,test)
				 (exit& ,end)))))
		   x)
		e))))
      ;; Add the :immediate keyword, to get a canonic use of until&
      ((?- ?test ?body-head . ?body-tail)
       (hiphop-until&-expander
	  (evepairify `(until& :immediate #f ,test ,body-head ,@body-tail) x) e))
      (else
       (error "until&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-abort&-expander ...                                       */
;*---------------------------------------------------------------------*/
(define (hiphop-abort&-expander x e)
   (match-case x
      ((?- :immediate ?immediate? ?event-test (handler& ?handler) ?body-head . ?body-tail)
       (let ((symb (gensym 'abort-)))
	  (let ((trap-end (string->symbol (format "~a/end" symb)))
		(trap-handle (string->symbol (format "~a/handle" symb)))
		(suspend-seq (string->symbol (format "~a/suspend-exit/end" symb)))
		(await-seq   (string->symbol (format "~a/await-exit/handle" symb))))
	     (e (evepairify-deep
		   `(trap& ,trap-end
		       (seq& :id (quote ,trap-end)
			  (trap& ,trap-handle
			     (par& :id (quote ,trap-handle)
				(seq& :id (quote ,suspend-seq)
				   (suspend& :immediate ,immediate? ,event-test ,body-head ,@body-tail)
				   (exit& ,trap-end))
				(seq& :id (quote ,await-seq)
				   (await& :immediate ,immediate? ,event-test)
				   (exit& ,trap-handle))))
			  ,handler))
		   x)
		e))))
      ((?- :immediate ?immediate? ?event-test ?body-head . ?body-tail)
       (hiphop-abort&-expander
	  (evepairify-deep
	     `(abort& :immediate ,immediate?
		 ,event-test
		 (handler& (nothing&))
		 ,body-head
		 ,@body-tail)
	     x)
	  e))
      ((?- ?event-test (handler& ?handler) ?body-head . ?body-tail)
       (hiphop-abort&-expander
	  (evepairify-deep
	     `(abort& :immediate #f
		 ,event-test
		 (handler& ,handler)
		 ,body-head
		 ,@body-tail)
	     x)
	  e))
      ((?- ?event-test ?body-head . ?body-tail)
       (hiphop-abort&-expander
	  (evepairify-deep
	     `(abort& :immediate #f
		 ,event-test
		 (handler& (nothing&))
		 ,body-head
		 ,@body-tail)
	     x)
	  e))
      (else
       (error "abort&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-every&-expander ...                                       */
;*---------------------------------------------------------------------*/
(define (hiphop-every&-expander x e)
   (match-case x
      ((?- :immediate ?immediate? ?event-test ?body-head . ?body-tail)
       (e (evepairify-deep
	     `(seq& :id (gensym 'every&-)
		 (await& :immediate ,immediate? ,event-test)
		 (loop& (abort& ,event-test ,body-head ,@body-tail (halt&))))
	     x)
	  e))
      ((?- (and (not (? keyword?)) ?first) . ?rest)
       (hiphop-every&-expander
	  (evepairify-deep
	     `(every& :immediate #f ,first ,@rest)
	     x)
	  e))
      (else
       (error "every&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-make-host-expression-expander ...                         */
;*---------------------------------------------------------------------*/
(define (hiphop-host-expression-expander x e)
   (e `(make-host-expression&
	  (quote ,(and (epair? x) (cer x)))
	  ,x)
      e))

;*---------------------------------------------------------------------*/
;*    hiphop-expressions-expander ...                                  */
;*---------------------------------------------------------------------*/
(define (hiphop-expressions-expander x e)
   ;; This expander is in charge of finding any val& or now& forms in tests
   (let* ((now-table (make-hashtable))
	  (val-table (make-hashtable))
	  (pre-table (make-hashtable))
	  (vexpd (lambda (x e2) ; This expander recursively search for val& forms
		    (match-case x
		       ((now& ?evt)
			(hashtable-put! now-table evt (and (epair? x) (cer x)))
			(e `(hiphop-now& (current-machine) ,evt) e))
		       ((val& ?evt)
			(hashtable-put! val-table evt (and (epair? x) (cer x)))
			(e `(hiphop-val& (current-machine) ,evt) e))
		       ((pre& ?evt)
			(hashtable-put! pre-table evt (and (epair? x) (cer x)))
			(e `(hiphop-pre& (current-machine) ,evt) e))
		       ((preval& ?evt)
			(hashtable-put! pre-table evt (and (epair? x) (cer x)))
			(e `(hiphop-preval& (current-machine) ,evt) e))
		       ((?- ???-) ; otherwise, recurse
			(map (lambda (x2) (e2 x2 e2)) x))
		       (else ; x is an atomic form
			(e x e)))))
	  ;; Expansion of the expression
	  (expanded-form (vexpd x vexpd)))
	  ;; Enclose the expansion into (lambda (machine) ...)
      (e `(make-host-expression&
	     (quote ,(and (epair? x) (cer x)))
	     (lambda (the-machine)
		(let ((current-machine (lambda () the-machine)))
		   ,expanded-form))
	     :now-depends (list ,@(hashtable-key-list now-table))
	     :val-depends (list ,@(hashtable-key-list val-table))
	     :pre-depends (list ,@(hashtable-key-list pre-table)))
	 e)))


;*---------------------------------------------------------------------*/
;*    hiphop-event&-expander ...                                       */
;*---------------------------------------------------------------------*/
(define (hiphop-event&-expander x e)
   (match-case x
      ((?- ?name ?type . ?rest)
       (e `(declare-event&
	      :loc (quote ,(and (epair? x) (cer x)))
	      :name ,name
	      :type ,type
	      ,@rest) e))
      (else
       (error "event&" "Illegal form" x))))

;*---------------------------------------------------------------------*/
;*    hiphop-let&-expander ...                                         */
;*---------------------------------------------------------------------*/
(define (hiphop-let&-expander x e)
   (error "let&" "The let& form is deprecated and removed" x))

;*---------------------------------------------------------------------*/
;*    hiphop-handler&-expander ...                                     */
;*---------------------------------------------------------------------*/
(define (hiphop-handler&-expander x e)
   (error "handler&" "Illegal use of handler&. Must be in an until& form" x))

;*---------------------------------------------------------------------*/
;*    hiphop-now&-expander ...                                     */
;*---------------------------------------------------------------------*/
(define (hiphop-now&-expander x e)
   (error "now&" "Illegal use of now& outside an host expression" x))

;*---------------------------------------------------------------------*/
;*    hiphop-val&-expander ...                                         */
;*---------------------------------------------------------------------*/
(define (hiphop-val&-expander x e)
   (error "val&" "Illegal use of val& outside an host expression" x))

;*---------------------------------------------------------------------*/
;*    hiphop-pre&-expander ...                                         */
;*---------------------------------------------------------------------*/
(define (hiphop-pre&-expander x e)
   (error "pre&" "Illegal use of pre& outside an host expression" x))

;*---------------------------------------------------------------------*/
;*    hiphop-preval&-expander ...                                      */
;*---------------------------------------------------------------------*/
(define (hiphop-preval&-expander x e)
   (error "preval&" "Illegal use of preval& outside an host expression" x))
