(module __hiphop_types

   (cond-expand
      (scheme2js
         (export (class &error obj proc msg
			(location (default #f))
			(fname (default #f))))))

   (export  (class hiphop-machine%
	       program::hiphop-program
	       inputs::pair-nil
	       outputs::pair-nil
	       current-inputs::pair-nil
	       state::symbol
	       environment::hiphop-environment%
	       (auto-react?::bool (default #f))
	       (must-react?::bool (default #f))
	       (raw-errors?::bool (default #f))
	       (dynamic?::bool (default #f))
	       (dynprog::pair-nil (default '())))

	    (class hiphop-sub-machine%::hiphop-machine%)

	    (class hiphop-environment%
	       events::vector
	       inputs::pair-nil
	       outputs::pair-nil
	       (current-machine::hiphop-machine% (default (class-nil hiphop-machine%)))
	       ;; Hashtables
	       ; The local event table
	       (%events (default (make-hashtable)))
	       ; The global event table
	       %gevents
	       ; The table of all the pauses
	       %pauses
	       ; The table of namespaces and programs
	       %init-table
	       ; The table of hiphop-emit objects (key are: (list event-index emit-index))
	       %emit-table
	       ; A boolean flag to test wether the environment has been altered by Must or Can
	       (%altered? (default #f)))

	    (class hiphop-program
	       source-location
	       reactive-code::hiphop-ast%
	       name::bstring
	       initializer
	       (initialized?::bool (default #f))
	       formal-inputs
	       formal-outputs
	       (%env::pair-nil (default '())))

	    (abstract-class hiphop-ast%
	       source-location
	       (id::symbol (default (gensym)))
	       ; The following fields are bit-vectors describing sets. 0 means the empty set.
	       (%gamma::bint (default 0))
	       (%omega::bint (default 0))
	       (%risk::bint (default 0)))

	    ; Nothing
	    (class hiphop-nothing::hiphop-ast%)

	    ; Emit
	    (class hiphop-emit::hiphop-ast%
	       name::bstring
	       valued?::bool
	       expr ; hiphop-host-expr that compute the emitted value or #f
	       (event::hiphop-event (default (class-nil hiphop-event)))
	       (flagged?::bool (default #f)) ; has the event been emitted yet?
	       (index (default -1))
	       (reachable? (default #f)) ; Can the emit be reached?
	       (%last-seen::bint (default -1)))

	    ; Pause
	    (class hiphop-pause::hiphop-ast%
	       (selected?::bool (default #f)))

	    ; Goto Pause
	    (class hiphop-goto-pause::hiphop-ast%)

	    ; If
	    (class hiphop-if::hiphop-ast%
	       test::hiphop-host-expr
	       then::hiphop-ast%
	       else::hiphop-ast%
	       (branch::symbol (default 'none)))

	    ; Suspend
	    (class hiphop-suspend::hiphop-ast%
	       event-test::hiphop-host-expr
	       body::hiphop-ast%
	       (%rego?::bool (default #f)))

	    ; Seq
	    (class hiphop-sequence::hiphop-ast%
	       branches::vector
	       (current-branch::int (default -1))) ; Current selected branch

	    ; Loop
	    (class hiphop-loop::hiphop-ast%
	       body::hiphop-ast%)

	    ; Par
	    (class hiphop-parallel::hiphop-ast%
	       active-branches::vector
	       branches::vector)

	    ;; Dedicated parallel statement for dynamic addition of programs
	    (class hiphop-parallel-dynamic::hiphop-parallel
	       ;; The index of the last initialized branch
	       (%next-initializable-index::bint (default 0))
	       ;; The closure used to update the environment
	       (%env-builder (default #f)))

	    ;; genpar&
	    (class hiphop-genpar::hiphop-parallel
	       ;; The HipHop expression generator
	       generator::procedure
	       ;; The identifiers used
	       host-identifiers::pair
	       ;; The host-expressions
	       host-expressions::pair
	       ;; The state
	       (expanded?::bool (default #f))
	       ;; The Can "nil-instance" of the generator
	       (can-nil-instance (default #f))
	       ;; The closure used to build the environment
	       (%env-builder (default #f)))

	    ;; dyngenpar&
	    (class hiphop-dyngenpar::hiphop-parallel-dynamic
	       ;; The when expression
	       when::hiphop-host-expr
	       ;; The genpar& expression to expand
	       genpar::hiphop-genpar
	       ;; A Boolean to know if the statement as started yet
	       ;; (for Go/Resume distinction)
	       (has-started? (default #f))
	       ;; A Boolean stating if the dygnepar has expanded
	       ;; during the current reaction (for Can). True event
	       ;; for empty expansions
	       (expanded? (default #f))
	       ;; A Boolean stating if the inherited par& has selected branches or not
	       (selected-branches? (default #f))
	       ;; A boolean stating that dyngenpar has a new branch
	       (has-new-branch? (default #f)))

	    ; Trap
	    (class hiphop-trap::hiphop-ast%
	       name::symbol
	       body::hiphop-ast%)

	    ; Exit
	    (class hiphop-exit::hiphop-ast%
	       trap-name::symbol
	       (index::int (default -1))) ; trap-index

	    ; Local
	    (class hiphop-local::hiphop-ast%
	       user-event;::hiphop-user-event
	       body::hiphop-ast%
	       (event::hiphop-event (default (class-nil hiphop-event))))

	    ; Run
	    (class hiphop-run::hiphop-ast%
	       name::obj ; The name of the renaming, if any
	       bindings::pair-nil ; The list of bindings
	       program::hiphop-program
	       (%bindings (default (make-hashtable))))

	    ; Atom
	    (class hiphop-atom::hiphop-ast%
	       expr::hiphop-host-expr
	       (ran?::bool (default #f)))

	    ; Host expressions
	    (class hiphop-host-expr
	       source-location
	       expression::procedure
	       ; Dependencies as lists (given as argument)
	       %now-depends::pair-nil
	       %val-depends::pair-nil
	       %pre-depends::pair-nil
	       ; Dependencies as index (computed at compile-time)
	       (%now-vindexes::vector (default '#()))
	       (%val-vindexes::vector (default '#()))
	       ; Table of all referenced events in the expression
	       (%event-table (default (make-hashtable)))
	       ; Are the dependencies solved?
	       (%resolved?::bool (default #f))
	       ; Computed expression
	       (%computed?::bool (default #f))
	       (%computed-expression::obj (default #unspecified)))

	    ;;; Events
	    ;; user event, only used at compile-time. Will be
	    ;; converted later into hiphop-event%
	    (class hiphop-user-event
	       ; name
	       name::bstring
	       ; combine function, actually, ::procedure or #f
	       combine::obj
	       ; default event value
	       default::obj
	       ; only one emit allowed at a time
	       (multi?::bool (default #t)))

	    (abstract-class hiphop-user-event2
	       ; location
	       source-location
	       ; name
	       name::bstring)
	    (class hiphop-user-event-pure::hiphop-user-event2)
	    (class hiphop-user-event-single::hiphop-user-event2
	       ; default value
	       default::obj)
	    (class hiphop-user-event-combined::hiphop-user-event-single
	       ; combine procedure
	       combine::procedure)
	    (class hiphop-user-event-multi::hiphop-user-event2)

	    ;; Named event
	    (class hiphop-event
	       name
	       user;::hiphop-user-event
	       event%::hiphop-event%)

	    ;; Abstract event
	    (class hiphop-event%
	       index::int
	       (status::symbol (default 'unknown))
	       ; Does the event carry a value?
	       (valued?::bool (default #f))
	       ; Does this value(s) come from the current reaction
	       (emitted/value?::bool (default #f))
	       ; listeners
	       (listeners::pair-nil (default '()))
	       ; emitters
	       (emitters::pair-nil (default '())) ; use counter
	       (reachable-emitters::pair-nil (default '()))
	       (emitted::pair-nil (default '())) ; use counter
	       (emitter-counter::int (default 0))
	       ; counters
	       (reachable-count::int (default 0))
	       (emitted-count::int (default 0))
	       ; previous status
	       (pre-status::symbol (default 'absent)))

	    ;; Pure event
	    (class hiphop-pure-event::hiphop-event%)

	    ;; Single-valued event
	    (class hiphop-valued-event::hiphop-event%
	       ; The event default value
	       (default::obj (default #f))
	       ; the actual event value
	       (value::obj (default #f))
	       ; the previous event value
	       (pre-value::obj (default #f)))

	    ;; Combined event
	    (class hiphop-combined-event::hiphop-valued-event
	       ; the combine procedure
	       combine::procedure)

	    ;; Multi-valued event
	    (class hiphop-multi-event::hiphop-event%
	       ; the actual event value
	       (value::obj (default '()))
	       ; Is this value multi-valued?
	       (mvalued?::bool (default #f))
	       ; the previous event value
	       (pre-value::obj (default '())))


	    ;; Errors ;;
	    ; Super type of HipHop-related errors
	    (abstract-class &hiphop-error::&error)
	    ; Compile-time error (mainly syntax errors)
	    (class &hiphop-compile-error::&hiphop-error)
	    ; Causality errors
	    (class &hiphop-causality-error::&hiphop-error)
	    ; Bad arguments to hiphop-* API functions
	    (class &hiphop-interface-error::&hiphop-error)
	    ; Error that are bugs in the implementation
	    (class &hiphop-internal-error::&hiphop-error)))
